/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  Part of FSLView 1.0
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    LICENCE

FSLView v1.0 sources are released under the terms of the GPL
(http://www.gnu.org/copyleft/gpl.html). FSLView binaries use Qt
(http://www.trolltech.com) which is released under GPL for X11 and
under the "Qt non-commercial licence" for Windows.  The FSLView
binaries are released under the GPL, with the added exception that we
also give permission to link this program with the Qt non-commercial
edition, and distribute the resulting executable, without including
the source code for the Qt non-commercial edition in the source
distribution.

FSLView is distributed "AS IS" under this Licence in the hope that it
will be useful, but in order that the University as a charitable
foundation protects its assets for the benefit of its educational and
research purposes, the University makes clear that no condition is
made or to be implied, nor is any warranty given or to be implied, as
to the accuracy of the Software, or that it will be suitable for any
particular purpose or for use under any specific conditions.
Furthermore, the University disclaims all responsibility for the use
which is made of the Software. It further disclaims any liability for
the outcomes arising from using the Software.
    
 */


#include "image.h"
#include <stdio.h>
#include "error.h"
#include <assert.h>

Image::Image(const std::string& filename)
{
  m_avw = AvwOpen(filename.c_str(), "r");
  m_imageInfo = ImageInfo::init(m_avw,filename); 
  m_cachedVolumes.clear();
  m_cachedVolumes.resize(m_imageInfo->inqNumVolumes());
}

Image::Image(ImageInfo::Handle info)
{
  m_avw = NULL;
  m_imageInfo = info;
  m_cachedVolumes.resize(info->inqNumVolumes());
 
  int x = m_imageInfo->inqX();
  int y = m_imageInfo->inqY();
  int z = m_imageInfo->inqZ();
  int v = m_imageInfo->inqNumVolumes();

  m_imageInfo->setMin(0);
  m_imageInfo->setMax(4);

  for(int n = 0; n < v; ++n)
    {
      switch(m_imageInfo->inqDt()) 
      {
       case DT_UNSIGNED_CHAR: m_cachedVolumes[n] = VolumeB::createBlank(x,y,z); break;
       case DT_SIGNED_SHORT:  m_cachedVolumes[n] = VolumeS::createBlank(x,y,z); break;
       case DT_SIGNED_INT:    m_cachedVolumes[n] = VolumeI::createBlank(x,y,z); break;
       case DT_FLOAT:         m_cachedVolumes[n] = VolumeF::createBlank(x,y,z); break;
       case DT_DOUBLE:        m_cachedVolumes[n] = VolumeD::createBlank(x,y,z); break;
      }
    }
}

Image::~Image()
{  
  if(m_avw != NULL){AvwClose(m_avw);}
}

Image::Handle Image::load(const std::string& filename)
{
  return Image::Handle(new Image(filename));
}

bool Image::save(const std::string& filename)
{
  AVW *avw = AvwOpen(filename.c_str(), "w");

  if( !avw ) return false;

  m_imageInfo->saveAvwHeader(avw);

  int v = m_imageInfo->inqNumVolumes();
  for(int n = 0; n < v; ++n)
    {
      Volume::Handle v = getVolume(n);

      v->saveVolume(avw);
    }

  AvwClose(avw);

  return true;
}

Image::Handle Image::cloneStructure()
{
  Image::Handle h = Image::Handle(new Image(m_imageInfo->clone()));
  h->getInfo()->setLutName("Red to yellow");
  return h;
}

Volume::Handle Image::getVolume(short n) const
{
  assert(n < m_imageInfo->inqNumVolumes());
 
  float min = m_imageInfo->inqMin();
  float max = m_imageInfo->inqMax();

  if(!m_cachedVolumes[n].get())
    {
	  std::string name;
      switch(m_imageInfo->inqDt()) 
      {
      case DT_UNSIGNED_CHAR: name="unsigned char"; m_cachedVolumes[n] = VolumeB::getVolume(m_avw, min, max, n); break;
      case DT_SIGNED_SHORT:  name="signed short"; m_cachedVolumes[n] = VolumeS::getVolume(m_avw, min, max, n); break;
      case DT_SIGNED_INT:    name="signed int"; m_cachedVolumes[n] = VolumeI::getVolume(m_avw, min, max, n); break;
      case DT_FLOAT:         name="float"; m_cachedVolumes[n] = VolumeF::getVolume(m_avw, min, max, n); break;
      case DT_DOUBLE:        name="double"; m_cachedVolumes[n] = VolumeD::getVolume(m_avw, min, max, n); break;
      default:
            throw FileError(m_imageInfo->inqFileName(),"Data type not supported.");
      }
    }

  return m_cachedVolumes[n];
}

TimeSeries::Handle Image::getTimeSeries(short x, short y, short z) const
{
  
  short xDim,yDim,zDim;
  long  voxel;
  TimeSeriesMap::iterator i;
  
  xDim = m_imageInfo->inqX();
  yDim = m_imageInfo->inqY();
  zDim = m_imageInfo->inqZ();

  voxel = xDim*(y + z * yDim) + x;
  i =  m_cachedTimeSeries.find(voxel);

  if(m_cachedTimeSeries.end() == i)
  {      
      switch(m_imageInfo->inqDt()) 
      {
       case DT_UNSIGNED_CHAR: m_cachedTimeSeries.insert(TimeSeriesMap::value_type(voxel, TimeSeriesB::getTimeSeries(m_avw,x,y,z))); break;
       case DT_SIGNED_SHORT:  m_cachedTimeSeries.insert(TimeSeriesMap::value_type(voxel, TimeSeriesS::getTimeSeries(m_avw,x,y,z))); break;
       case DT_SIGNED_INT:    m_cachedTimeSeries.insert(TimeSeriesMap::value_type(voxel, TimeSeriesI::getTimeSeries(m_avw,x,y,z))); break;
       case DT_FLOAT:         m_cachedTimeSeries.insert(TimeSeriesMap::value_type(voxel, TimeSeriesF::getTimeSeries(m_avw,x,y,z))); break;
       case DT_DOUBLE:        m_cachedTimeSeries.insert(TimeSeriesMap::value_type(voxel, TimeSeriesD::getTimeSeries(m_avw,x,y,z))); break;
      }
  }

 return m_cachedTimeSeries[voxel];
}

Volume::Handle Image::blankDraw()
{
  VolumeS::Handle v = VolumeS::create(m_imageInfo->inqX(),
                                      m_imageInfo->inqY(),
                                      m_imageInfo->inqZ(),
                                      new short[m_imageInfo->inqX()*
                                                m_imageInfo->inqY()*
                                                m_imageInfo->inqZ()]);
  v->setMin(0);
  v->setMax(255);
  

  unsigned int width  = v->inqX();
  unsigned int height = v->inqY();
  unsigned int depth  = v->inqZ();

  for(unsigned int z = 0; z< depth; ++z)    
  {
    for(unsigned int y = 0; y < height; ++y)
    {      
      for(unsigned int x = 0; x < width; ++x)
        {
          (*v)(x,y,z) = 0;
        }
    }
  }
  
  return v;
}
