/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  Part of FSLView 1.0
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    LICENCE

FSLView v1.0 sources are released under the terms of the GPL
(http://www.gnu.org/copyleft/gpl.html). FSLView binaries use Qt
(http://www.trolltech.com) which is released under GPL for X11 and
under the "Qt non-commercial licence" for Windows.  The FSLView
binaries are released under the GPL, with the added exception that we
also give permission to link this program with the Qt non-commercial
edition, and distribute the resulting executable, without including
the source code for the Qt non-commercial edition in the source
distribution.

FSLView is distributed "AS IS" under this Licence in the hope that it
will be useful, but in order that the University as a charitable
foundation protects its assets for the benefit of its educational and
research purposes, the University makes clear that no condition is
made or to be implied, nor is any warranty given or to be implied, as
to the accuracy of the Software, or that it will be suitable for any
particular purpose or for use under any specific conditions.
Furthermore, the University disclaims all responsibility for the use
which is made of the Software. It further disclaims any liability for
the outcomes arising from using the Software.
    
 */


#if !defined(IMAGEINFO_H)
#define IMAGEINFO_H

#include <boost/shared_ptr.hpp>
#include <string>

#include "avwio/avwio.h"

class ImageInfo
{
public:
  typedef boost::shared_ptr< ImageInfo > Handle;

  ~ImageInfo();

  short inqX() const;
  short inqY() const;
  short inqZ() const;
  short inqNumVolumes() const;
  float inqXDim() const;
  float inqYDim() const;
  float inqZDim() const;
  short inqXOrig() const;
  short inqYOrig() const;
  short inqZOrig() const;
  void setXOrig(short);
  void setYOrig(short);
  void setZOrig(short);

  float inqTr() const;
  short inqDt() const;
  size_t inqBytesPerVoxel() const;
  size_t inqBitsPerVoxel() const;
  std::string inqImageName() const;
  std::string inqFileName() const;
  std::string inqLutName() const;
  bool inqReadOnly() const;
  bool isCompatible(ImageInfo::Handle) const;
  void setReadOnly(bool);
  void setLutName(std::string);

  /** Image min instensity
   *  @return The stored minimum value or the minimum from the header if not yet set.
   */
  float inqMin() const;
  float inqMax() const;
  void setMin(float);
  void setMax(float);

//   int  inqHeaderMin() const;
//   int  inqHeaderMax() const;
//   void setHeaderMin(int);
//   void setHeaderMax(int);

  static ImageInfo::Handle init(AVW *avw, std::string filename) {
    return ImageInfo::Handle( new ImageInfo(avw,filename) );
  }

  void saveAvwHeader(AVW *avw);

  Handle clone();

private:
  ImageInfo(AVW *avw,std::string filename);
  ImageInfo(short x,short y,short z,short v,short dt,
            float xDim, float yDim, float zDim,float tr,
            size_t bitsPerVoxel,std::string name);
  std::string extractName(std::string filename);

//   int m_headerMin;
//   int m_headerMax;
  bool m_minSet;
  bool m_maxSet;
  float m_min;
  float m_max;

  short m_x;
  short m_y;
  short m_z;
  short m_v;
  short m_dt;
  float m_xDim;
  float m_yDim;
  float m_zDim;
  float m_tr;
  short m_orig[5];
  char  m_auxFile[24];
  std::string m_fileName;
  std::string m_imageName;
 
  size_t m_bitsPerVoxel;
  short m_type;  
  bool  m_readOnly;
};

inline void  ImageInfo::setReadOnly(bool state) {m_readOnly = state;}
inline bool  ImageInfo::inqReadOnly() const {return m_readOnly;}
inline short ImageInfo::inqX() const { return m_x; }
inline short ImageInfo::inqY() const { return m_y; }
inline short ImageInfo::inqZ() const { return m_z; }
inline short ImageInfo::inqNumVolumes() const { return m_v; }
inline float ImageInfo::inqXDim() const { return m_xDim; }
inline float ImageInfo::inqYDim() const { return m_yDim; }
inline float ImageInfo::inqZDim() const { return m_zDim; }
inline short ImageInfo::inqXOrig() const { return m_orig[0];}
inline short ImageInfo::inqYOrig() const { return m_orig[1];}
inline short ImageInfo::inqZOrig() const { return m_orig[2];}
inline void ImageInfo::setXOrig(short v) { m_orig[0] = v;}
inline void ImageInfo::setYOrig(short v) { m_orig[1] = v;}
inline void ImageInfo::setZOrig(short v) { m_orig[2] = v;}
inline float ImageInfo::inqTr() const { return m_tr; }
inline short ImageInfo::inqDt() const { return m_dt; }
// inline int   ImageInfo::inqHeaderMin() const { return m_headerMin; }
// inline int   ImageInfo::inqHeaderMax() const { return m_headerMin; }
// inline void  ImageInfo::setHeaderMin(int min) { m_min = min; }
// inline void  ImageInfo::setHeaderMax(int max) { m_max = max; }
inline std::string ImageInfo::inqImageName() const { return m_imageName; }
inline std::string ImageInfo::inqFileName() const { return m_fileName; }
inline size_t ImageInfo::inqBytesPerVoxel() const { return m_bitsPerVoxel / 8; }
inline size_t ImageInfo::inqBitsPerVoxel() const { return m_bitsPerVoxel; }

#endif
