/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  Part of FSLView 1.0
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    LICENCE

FSLView v1.0 sources are released under the terms of the GPL
(http://www.gnu.org/copyleft/gpl.html). FSLView binaries use Qt
(http://www.trolltech.com) which is released under GPL for X11 and
under the "Qt non-commercial licence" for Windows.  The FSLView
binaries are released under the GPL, with the added exception that we
also give permission to link this program with the Qt non-commercial
edition, and distribute the resulting executable, without including
the source code for the Qt non-commercial edition in the source
distribution.

FSLView is distributed "AS IS" under this Licence in the hope that it
will be useful, but in order that the University as a charitable
foundation protects its assets for the benefit of its educational and
research purposes, the University makes clear that no condition is
made or to be implied, nor is any warranty given or to be implied, as
to the accuracy of the Software, or that it will be suitable for any
particular purpose or for use under any specific conditions.
Furthermore, the University disclaims all responsibility for the use
which is made of the Software. It further disclaims any liability for
the outcomes arising from using the Software.
    
 */


// volume.h: interface for the Volume class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(VOLUME_H)
#define VOLUME_H

#include <avwio/avwio.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <string>

class Volume
{
public:
  typedef boost::shared_ptr< Volume > Handle;

  virtual short inqX() const = 0;
  virtual short inqY() const = 0;
  virtual short inqZ() const = 0;

  virtual float normalized(short x, short y, short z) = 0;
  virtual float value(unsigned int offset) = 0;
  virtual float value(short x, short y, short z) = 0;
  virtual void  setValue(short x, short y, short z, float value) = 0;
  virtual float inqMin() const = 0;
  virtual float inqMax() const = 0;
  virtual void calculateMinMax() = 0;
  virtual bool saveVolume(AVW *avw) = 0;

  Volume() {}
  virtual ~Volume() {}
};

template <typename VoxelType>
class VolumeStore: public Volume
{
  friend class TestVolumeStore;

public:
  typedef boost::shared_ptr< VolumeStore<VoxelType> > Handle;

  virtual ~VolumeStore();
	
  VoxelType& operator()(short x, short y, short z);
  VoxelType& operator()(unsigned int offset);

  virtual float normalized(short x, short y, short z);
  virtual float value(short x, short y, short z);
  virtual float value(unsigned int offset) { return m_buffer[offset]; }
  virtual void  setValue(short x, short y, short z, float value);
  virtual short inqX() const;
  virtual short inqY() const;
  virtual short inqZ() const;

  void setMin(float min){m_min = min;}
  void setMax(float max){m_max = max;}

  virtual float inqMin() const { return m_min; }
  virtual float inqMax() const { return m_max; }

  virtual void calculateMinMax();

  //  static Handle load(const std::string& filename, size_t n = 0);
  static Handle getVolume(AVW* avw, float min, float max, size_t n);
  static Handle clone(const Handle& src);
  static Handle create(short x, short y, short z, VoxelType* buffer);
  static Handle createBlank(short x, short y, short z);

  bool saveVolume(AVW *avw);

private:
  VolumeStore(short x, short y, short z, VoxelType* buffer);

  short m_x, m_y, m_z;
  float m_min, m_max;
  VoxelType* m_buffer;
};

typedef VolumeStore<unsigned char> VolumeB;
typedef VolumeStore<short>         VolumeS;
typedef VolumeStore<int>           VolumeI;
typedef VolumeStore<float>         VolumeF;
typedef VolumeStore<double>        VolumeD;

#include "volume.inc"

#endif
