/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

/****************************************************************************
** $Id: application.cpp,v 1.76 2003/07/10 14:39:35 jim Exp $
**
** Copyright (C) 2002 University of Oxford.  All rights reserved.
**
** FSLView
**
*****************************************************************************/
#include "viewwidget.h"
#include "orthowidget.h"
#include "lightboxwidget.h"
#include "moviewidget.h"
#include "timeserieswidget.h"
#include "multserieswidget.h"
#include "histogramwidget.h"
#include "application.h"
#include <stdio.h>
#include <qworkspace.h>
#include <qpixmap.h>
#include <qlabel.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qkeycode.h>
#include <qfiledialog.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qaccel.h>
#include <qwhatsthis.h>
#include <qtoolbutton.h>
#include <qtooltip.h>

#include "fileopen.xpm"
#include "fsllogo.xpm"

#include "glinfo.h"
#include "tracker.h"
#include "version.h"
#include "storage/error.h"

const char * fileOpenText = "Click this button to open a <em>folder</em>.<br><hr>"
"You can also select the <b>Open command</b> from the File menu.";

ApplicationWindow::ApplicationWindow()
  : QMainWindow( 0, "FSLView", WDestructiveClose ),m_lightBoxEnabled(true)
{  
  TRACKER("ApplicationWindow::ApplicationWindow");
  qApp->setMainWidget(this);
  setIcon( QPixmap(fsllogo_xpm) );
  setupStatusBar(); 

  m_ws = new QWorkspace(this);
  m_ws->setBackgroundMode( PaletteMidlight );

  #if !defined(WIN32)
	m_ws->setScrollBarsEnabled(true); //QT 3 only
  #endif

  setCentralWidget(m_ws);
  buildMenus(); 
  connectControls(); 
  show();
  
  emit message( "Ready", 2000 ); 
  
  GLInfo::Handle gli = GLInfo::create();
  
  if(gli->limitToContexts())
  {
    emit message("Warning: Graphics driver will not support lightbox view", 3000);
    m_lightBoxEnabled = false;
   }

  if(qApp->argc() > 1)
    {
      QString arg(qApp->argv()[1]);
      if(arg == "-drivertest")
        {
          m_lightBoxEnabled = true;
        }
      else
        {
          QFileInfo  fi(qApp->argv()[1]);
          loadFile(fi.absFilePath());
          QDir::setCurrent(fi.dirPath(true));
        }
    }
}



ApplicationWindow::~ApplicationWindow()
{
  TRACKER("ApplicationWindow::~ApplicationWindow");
}


void ApplicationWindow::load()
{
 QString fn = QFileDialog::getOpenFileName( QString::null, "Analyze files (*.hdr)", this );
 if(!fn.isEmpty())loadFile(fn);

 fileMenuAboutToShow();
 viewMenuAboutToShow();
}

void ApplicationWindow::loadFile(const QString & absFilePath)
{
  if(!checkAbsFilePath(absFilePath))
    {    
      emit message(QString("Cannot open...%1").arg(absFilePath),4000);           return;
    }

  QString fn(absFilePath);

  try
  {
    if ( !fn.isEmpty() ) 
    {
      fn.remove( fn.findRev("."), 4 );
      emit message( QString("Loading.... %1").arg(fn), 0);
      
      m_imageGroup = ImageGroup::create( Image::load((const char *)fn) );
      m_cursor = Cursor::create(this);
      ImageInfo::Handle info(m_imageGroup->getMainImage()->getInfo());
      
      
      orthoView();
   
      setCaption( fn );
      emit message( QString("Loaded image %1").arg(fn), 2000 );

      m_cursor->setCursor( info->inqX()/2, info->inqY()/2, info->inqZ()/2 ); 

      QApplication::restoreOverrideCursor();
    }  else {
      emit message( "Loading aborted", 2000 );
    }
  }
  catch(FileError& f)
    {
      catchFileError(f);
    }  

  fileMenuAboutToShow();
  viewMenuAboutToShow();
}

void ApplicationWindow::closeWindow()
{
  QWidget* w = m_ws->activeWindow();
  if(w){w->close();}

  fileMenuAboutToShow();
  viewMenuAboutToShow();
}

bool ApplicationWindow::windowListEmpty()
{
  return m_ws->windowList().isEmpty();
}

void ApplicationWindow::about()
{
  QMessageBox::about( this, "FSLView",
		      QString("<h2>FSLView</h2><hr><b>Version %1 %2</b><br>"
			      "Written by Dave Flitney, James Saunders and Steve Smith<br>"
			      "Copyright(c) 2003 University of Oxford<br>"
			      "<b>html://www.fmrib.ox.ac.uk/fsl</b><hr>"
			      "Please report  bugs to:<br><b>fslview-bugs@fmrib.ox.ac.uk<b>")
		      .arg(Version).arg(Release));
}

void ApplicationWindow::aboutGraphics()
{
  GLInfo::Handle info = GLInfo::create();  
   info->inqVendor();
  info->inqRenderer();
  info->inqVersion();
  QString limitMsg;
  QString blendMsg;
  if(info->limitToContexts())limitMsg = "Lightbox view not supported";
  if(!info->imagingSubset()) blendMsg = "Image blending not supported";

  QMessageBox::about( this, "FSLView",
		      QString("<h2>FSLView</h2><hr>Graphics Drivers<br>"
			      "Vendor:   %1<br>"
            "Renderer: %2<br>"
            "Version:  %3<br>"
            "%4<br>"
            "%5<br>")
  		      .arg(info->inqVendor())
          .arg(info->inqRenderer())
          .arg(info->inqVersion())
          .arg(limitMsg)
          .arg(blendMsg));  
}

void ApplicationWindow::aboutQt()
{
  QMessageBox::aboutQt( this, "fslview" );
}

void ApplicationWindow::windowsMenuAboutToShow()
{ 
  m_windowsMenu->clear();
  int cascadeId = m_windowsMenu->insertItem("&Cascade", m_ws, SLOT(cascade() ) );
  int tileId = m_windowsMenu->insertItem("&Tile", m_ws, SLOT(tile() ) );
  if ( m_ws->windowList().isEmpty() ) {
    m_windowsMenu->setItemEnabled( cascadeId, FALSE );
    m_windowsMenu->setItemEnabled( tileId, FALSE );
  }
  m_windowsMenu->insertSeparator();
  QWidgetList windows = m_ws->windowList();
  for ( int i = 0; i < int(windows.count()); ++i ) {
    int id = m_windowsMenu->insertItem(windows.at(i)->caption(),
				      this, SLOT( windowsMenuActivated( int ) ) );
    m_windowsMenu->setItemParameter( id, i );
    m_windowsMenu->setItemChecked( id, m_ws->activeWindow() == windows.at(i) );
  }
}

void ApplicationWindow::windowsMenuActivated( int id )
{
  QWidget* w = m_ws->windowList().at( id );
  if ( w ) {
    w->showNormal();
    w->setFocus();
  }
}
void ApplicationWindow::fileMenuAboutToShow()
{
  bool overlays(false);
  bool state(windowListEmpty());
  if (!state)
      overlays = m_imageGroup->size();

  m_fileMenu->setItemEnabled(m_fileCreateMaskID,!state);
  m_fileMenu->setItemEnabled(m_fileAddOverlayID,!state);
  m_fileMenu->setItemEnabled(m_fileRemOverlayID,overlays );
  m_fileMenu->setItemEnabled(m_fileSaveOverlayID,!state );
  m_fileMenu->setItemEnabled(m_fileOpenID,state);
  m_fileMenu->setItemEnabled(m_fileCloseID,!state);      
}

void ApplicationWindow::viewMenuAboutToShow()
{
  TRACKER("ApplicationWindow::viewMenuAboutToShow");
  bool state(windowListEmpty());
  m_viewMenu->setItemEnabled(m_viewOrthoID,!state);
  m_viewMenu->setItemEnabled(m_viewLightID,!state);  
  m_viewMenu->setItemEnabled(m_viewMovieID,!state);   
  m_viewMenu->setItemEnabled(m_viewHistID,!state);  
  m_viewMenu->setItemEnabled(m_viewTimeID,!state);  
  m_viewMenu->setItemEnabled(m_viewTimeGridID,!state);
  
  if(!m_lightBoxEnabled) 
    m_viewMenu->setItemEnabled(m_viewLightID,false);    
}

void ApplicationWindow::setupStatusBar()
{
  m_statusBar = statusBar();
  m_statusBar->addWidget(new QLabel(m_statusBar), 1, FALSE);
}

void ApplicationWindow::buildMenus()
{
  m_mb = menuBar();
  
  QPixmap openIcon = QPixmap( fileopen_xpm );

  m_fileMenu = new QPopupMenu( this );
  connect( m_fileMenu, SIGNAL( aboutToShow() ),
	   this, SLOT( fileMenuAboutToShow() ) );
  m_fileOpenID        = m_fileMenu->insertItem( openIcon, "&Open...", this, SLOT(load()), CTRL+SHIFT+Key_O ); 
  m_fileCreateMaskID  = m_fileMenu->insertItem( "Create &mask", this, SLOT(createMask()), CTRL+Key_C );
  m_fileSaveOverlayID = m_fileMenu->insertItem( "&Save as...",   this, SLOT(saveOverlay()), CTRL+Key_S );
  m_fileAddOverlayID  = m_fileMenu->insertItem( "&Add...",    this, SLOT(addOverlay()), CTRL+Key_A );   
  m_fileRemOverlayID  = m_fileMenu->insertItem( "&Remove", this, SLOT(remOverlay()), CTRL+Key_R );
  m_fileMenu->insertSeparator();
  m_fileCloseID       = m_fileMenu->insertItem( "&Close", this, SLOT(closeWindow()), CTRL+Key_W );
  
  m_fileMenu->insertItem( "&Quit", qApp, SLOT( closeAllWindows() ), CTRL+Key_Q );
  
  m_windowsMenu = new QPopupMenu( this );
  connect( m_windowsMenu, SIGNAL( aboutToShow() ),
	   this, SLOT( windowsMenuAboutToShow() ) );
  m_windowsMenu->setCheckable( TRUE );
  
  m_viewMenu = new QPopupMenu( this );
  connect( m_viewMenu, SIGNAL( aboutToShow() ),
	   this, SLOT( viewMenuAboutToShow() ) );
  m_viewMenu->setCheckable(true);
  m_viewOrthoID = m_viewMenu->insertItem( "&Orthographic", this, SLOT(orthoView()), CTRL+Key_O );
  m_viewLightID = m_viewMenu->insertItem( "&Lightbox", this, SLOT(lightboxView()), CTRL+Key_L );
  m_viewMovieID = m_viewMenu->insertItem( "&Movie", this, SLOT(movieView()), CTRL+Key_M );
  m_viewMenu->insertSeparator();
  m_viewHistID = m_viewMenu->insertItem( "Image &histogram", this, SLOT(openHistogram()), CTRL+Key_H );
  m_viewTimeID = m_viewMenu->insertItem( "&Timeseries", this, SLOT(openTimeSeries()), CTRL+Key_T );
  m_viewTimeGridID = m_viewMenu->insertItem( "Timeseries &grid", this, SLOT(openMultTimeSeries()), CTRL+Key_G );
 
  QPopupMenu * help = new QPopupMenu( this );

  help->insertItem( "&About FSLView", this, SLOT(about()), Key_F1);
  help->insertItem( "About &graphics", this, SLOT(aboutGraphics()));
  help->insertSeparator();
  help->insertItem( "About&Qt", this, SLOT(aboutQt()));
  
  m_mb->insertItem( "&File", m_fileMenu );  
  m_mb->insertItem( "&Windows", m_windowsMenu );
  m_mb->insertItem( "&View",m_viewMenu  );
  m_mb->insertSeparator();
  m_mb->insertItem( "&Help", help );
 
  fileMenuAboutToShow();
  viewMenuAboutToShow();
}

void ApplicationWindow::connectControls()
{  
  connect( this, SIGNAL(      message(const QString&, int)),
           this, SLOT( displayMessage(const QString&, int)));
}

void ApplicationWindow::update(const Cursor::Handle& c)
{
}

void ApplicationWindow::saveOverlay()
{
  ViewWidget *view = (ViewWidget*)m_ws->activeWindow();
  OverlayList::Handle ol = view->getOverlayList();
  QString fn;
  bool ok(true);

  

  if(!isValidOverlayList(ol))
    {
      QMessageBox::warning(this, "FSLView",
			   "Active window must be an OrthoView, Lightbox or Movie");
      return;
    }

 
  
  Image::Handle image = ol->inqActiveImage();
    
  if(!isValidImage(image))
        {
          QMessageBox::warning(this,"FSLView",
			       "No image selected in Active window");
          return;
        }
   
          // Okay can go ahead and try to save it then
  fn = QFileDialog::getSaveFileName(QString(image->getInfo()->inqImageName().c_str()) 
                                    + ".hdr", "Analyze files (*.hdr)", this,
                                    "save file dialog",
                                    "Select a filename for saving");
         
        
 
  if(QFile::exists(fn))      
    {
      if(QMessageBox::warning( this,
                               "FSLView",
                               "File already exists. Do you want to overwrite it?",
                               "Cancel","OK","",1,0) == 0) 
              {      
                return;
              }
    }
     

   
           fn.remove( fn.findRev("."), 4 );

               if ( !fn.isEmpty() ) {
                 if(!image->save((const char *)fn))
                   QMessageBox::warning(this, "FSLView",
                   QString("<h2>Save failed!</h2>") +
                           "<p>Perhaps you're out of disk space or " +
			                     "you don't have permission to write to " +
			                     "this directory." +
			                     "<p>Please see console for more details." +
			                     "<p>Click Ok and try again.",
                   QMessageBox::Ok | QMessageBox::Default | QMessageBox::Escape,
                   QMessageBox::NoButton);
               
               }
}




void ApplicationWindow::remOverlay()
{  
 ViewWidget* view = (ViewWidget*)m_ws->activeWindow();
 OverlayList::Handle ol = view->getOverlayList();
 if(!isValidOverlayList(ol))
   {
     QMessageBox::warning(this, "FSLView",
                          "Active window must be either OrthoView, Lightbox or Movie");
   }
 else
  {
    
    Image::Handle mainImg = ol->getMainImage();
    Image::Handle image = ol->inqActiveImage();
    
    if(!isValidImage(image))
      {
        QMessageBox::warning(this,
                             "FSLView",
                             "No image selected in Active window");
      }
    else if(mainImg == image)
      {
        QMessageBox::warning(this,"FSLView","Cannot remove main image");
      }
    else
      {
        QString imageMessage = QString("Removing %1")
                .arg(image->getInfo()->inqImageName().c_str());
        
        switch(QMessageBox::warning( this,
                                     "FSLView",
                                     imageMessage,"Cancel","OK","",1,0))
             {
             case 0:break;
             case 1:m_imageGroup->remOverlay(image);break;
             }
     }
  }
}

void ApplicationWindow::addOverlay()
{
  try{
    QString fn = QFileDialog::getOpenFileName( QString::null, "Analyze files (*.hdr)", this );

    if ( !fn.isEmpty() ) {
      QApplication::setOverrideCursor(Qt::waitCursor);
      fn.remove( fn.findRev("."), 4 );
      emit message( QString("Loading overlay.... %1").arg(fn), 2000 );
      Image::Handle overlay = Image::load((const char *)fn);
      QApplication::restoreOverrideCursor();

      if(m_imageGroup->getMainImage()->getInfo()->isCompatible(overlay->getInfo()))
        m_imageGroup->addOverlay( overlay );
      else
      {
        QMessageBox::warning( this, "FSLView",
			      "Cannot load incompatible overlay" );
        emit message( "Loading aborted", 2000 );
      }
    } else {
        emit message( "Loading aborted", 2000 );
    }
   }
   catch(FileError& f)
   {
     catchFileError(f);
   }  

}

void ApplicationWindow::addLookUpTable()
{
  QString fn = QFileDialog::getOpenFileName( QString::null, "LUTs (*.lut *.rgb)", this );
  if ( !fn.isEmpty() ) {
    QApplication::setOverrideCursor(Qt::waitCursor);
    emit message( QString("Loading lookup table.... %1").arg(fn), 2000 );
    LookUpTable::Handle lookUpTable = LookUpTable::load((const char *)fn);
    m_imageGroup->addLookUpTable( lookUpTable );

    QApplication::restoreOverrideCursor();

 }  else {
    emit message( "Loading aborted", 2000 );
  }
}


void ApplicationWindow::createMask()
{
 TRACKER("ApplicationWindow::createMask");      
 ViewWidget* view = (ViewWidget*)m_ws->activeWindow();
 OverlayList::Handle ol = view->getOverlayList();
 if(!isValidOverlayList(ol))
   {
     QMessageBox::warning(this,
                          "FSLView",
                          "Active window must be either OrthoView, Lightbox or Movie");
   }
 else
  {
    
    Image::Handle mainImg = ol->getMainImage();
    Image::Handle image = ol->inqActiveImage();
    
    if(!isValidImage(image))
      {
        QMessageBox::warning(this,
                             "FSLView",
                             "No image selected in Active window");
      }
    else
      {
	Image::Handle spawn = image->cloneStructure();
	
	m_imageGroup->addOverlay(spawn);
	
	emit message( QString("Created new mask: %1-mask")
		      .arg(image->getInfo()->inqImageName().c_str()), 2000 );
      }
  }

}

void ApplicationWindow::orthoView()
{  
  ViewWidget* view = new OrthoWidget(m_ws,m_imageGroup,
                                     copyActiveOverlayList(), m_cursor); 

  connect( view, SIGNAL(message(const QString&, int)), 
           this, SIGNAL(message(const QString&, int)) );
  connect( view, SIGNAL(addLookUpTable()),
           this, SLOT(addLookUpTable()));

  view->setCaption("Ortho view");

  view->resize(580,500);

  viewShow(view);
}

void ApplicationWindow::lightboxView()
{
  ViewWidget* view = new LightboxWidget(m_ws, m_imageGroup,
                                        copyActiveOverlayList(),m_cursor); 

  connect( view, SIGNAL(message(const QString&, int)), 
           this, SIGNAL(message(const QString&, int)) );   
  connect( view, SIGNAL(addLookUpTable()),
           this, SLOT(addLookUpTable()));
  view->setCaption("Lightbox view");  
  view->resize(580,500);
  viewShow(view);
}

void ApplicationWindow::movieView()
{
  ViewWidget* view = new MovieWidget(m_ws, m_imageGroup,
                                     copyActiveOverlayList()); 

  connect( view, SIGNAL(message(const QString&, int)), 
           this, SIGNAL(message(const QString&, int)) );   
  connect( view, SIGNAL(addLookUpTable()),
           this, SLOT(addLookUpTable()));
  view->setCaption("Movie view");  
  view->resize(210,345);
  viewShow(view);
}

void ApplicationWindow::openTimeSeries()
{
  if(m_imageGroup.use_count())
  {
    TimeSeriesWidget* timeseries = new TimeSeriesWidget(m_ws, m_imageGroup, m_cursor);
    timeseries->setCaption("Timeseries");    
    timeseries->resize(250,250);
    viewShow(timeseries);
  }
}

void ApplicationWindow::openMultTimeSeries()
{
  if(m_imageGroup.use_count())
  {
    MultSeriesWidget* multseries = new MultSeriesWidget(m_ws, m_imageGroup, m_cursor);    
    multseries->setCaption("Multi-Timeseries");
    multseries->resize(400,400);
    viewShow(multseries);
  }
}
void ApplicationWindow::openHistogram()
{
  if(m_imageGroup.use_count())
  {
    HistogramWidget* histogram = new HistogramWidget(m_ws, m_imageGroup);
    histogram->setCaption("Histogram");
    histogram->resize(300,200);
    viewShow(histogram);
  }
}  

void ApplicationWindow::catchFileError(FileError f)
{
  emit message( QString("Error loading %1, %2")
                        .arg(f.inqFileName().c_str())
                        .arg(f.inqMessage().c_str()),
                         3000 ); 
}

bool ApplicationWindow::checkAbsFilePath(const QString & absFilePath)
{
  QString fn(absFilePath);
  fn.remove( fn.findRev("."), 4 );
  return QFile::exists(fn + ".hdr");
}

OverlayList::Handle ApplicationWindow::activeOverlayList()
{
  OverlayList::Handle ol;
  ViewWidget* view = (ViewWidget*)m_ws->activeWindow();

  if (!windowListEmpty())ol = view->getOverlayList();
  return ol;
}

OverlayList::Handle ApplicationWindow::copyActiveOverlayList()
{
  OverlayList::Handle null;
  OverlayList::Handle ol = activeOverlayList();
  
  if(isValidOverlayList(ol)){return ol->clone();}
  else                      {return null;}
}
 
void ApplicationWindow::displayMessage(const QString & msg, int time)
{
  statusBar()->message( msg, time );
}

void ApplicationWindow::viewShow(ViewWidget* v)
{
  if(windowListEmpty()) v->showMaximized();
  else if(m_ws->windowList().count() == 1) {
    m_ws->windowList().at(0)->showNormal();
    v->showNormal();
    m_ws->tile();
  } else
    v->show();
}
