/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  Part of FSLView 1.0
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    LICENCE

FSLView v1.0 sources are released under the terms of the GPL
(http://www.gnu.org/copyleft/gpl.html). FSLView binaries use Qt
(http://www.trolltech.com) which is released under GPL for X11 and
under the "Qt non-commercial licence" for Windows.  The FSLView
binaries are released under the GPL, with the added exception that we
also give permission to link this program with the Qt non-commercial
edition, and distribute the resulting executable, without including
the source code for the Qt non-commercial edition in the source
distribution.

FSLView is distributed "AS IS" under this Licence in the hope that it
will be useful, but in order that the University as a charitable
foundation protects its assets for the benefit of its educational and
research purposes, the University makes clear that no condition is
made or to be implied, nor is any warranty given or to be implied, as
to the accuracy of the Software, or that it will be suitable for any
particular purpose or for use under any specific conditions.
Furthermore, the University disclaims all responsibility for the use
which is made of the Software. It further disclaims any liability for
the outcomes arising from using the Software.
    
 */


#if !defined(CURSOR_H)
#define CURSOR_H

#include <list>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

class CursorObserver;

/** Cursor
 * An observable cursor class @see Design Patterns p293
 */
class Cursor
{
public:
  typedef boost::shared_ptr< Cursor > Handle;
  typedef boost::weak_ptr< Cursor > WeakHandle;
  
  Handle clone();

  void setCursor(short x, short y, short z, short v);
  void setCursor(short x, short y, short z);
  void setCursorRepaint(short x, short y, short z);
  void repaint();
  bool  inqRepaint() const;
  short inqX() const;
  short inqY() const;
  short inqZ() const;
  short inqV() const;

  /** Create a new cursor
   * @param o the initial image which will follow this cursor
   */
  static Handle create(CursorObserver *o) {
    Handle c = Handle(new Cursor(o));
    c->setCountedThis(c);
      
  return c;
  }

  /** Attach a view to this cursor
   * @param o handle of a view which requires notification of any 
   *          changes to this cursor
   */
  void attach(CursorObserver *o);

  /** Detach a view from this cursor
   * @param o handle of the view to be removed from the notification list
   */
  void detach(CursorObserver *o);

  virtual ~Cursor() {}

protected:
  Handle countedThis() const { return boost::make_shared(m_countedThis); }

private:
  Cursor(CursorObserver *o);
  Cursor(short x,short y,short z, short v);
  void setCountedThis(const Handle c) { m_countedThis = WeakHandle(c); }

  /** Notify all observers of a change to this cursor
   */
  void notify() const;

  short m_x;
  short m_y;
  short m_z;
  short m_v;
  bool  m_repaint;

  WeakHandle m_countedThis;

  std::list< CursorObserver * > m_observers;
};

inline short Cursor::inqX() const { return m_x; }
inline short Cursor::inqY() const { return m_y; }
inline short Cursor::inqZ() const { return m_z; }
inline short Cursor::inqV() const { return m_v; }
inline bool  Cursor::inqRepaint() const {return m_repaint;}

class CursorObserver
{
public:
  virtual ~CursorObserver() {}

  virtual void update(const Cursor::Handle& c) = 0;

  CursorObserver() {}
};


#endif
