/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#if defined(WIN32)
#pragma warning (disable:4786)
#endif

#include "cursorwidget.h"
#include "imagedisplaysetting.h"
#include "overlaylist.h"
#include <qspinbox.h>
#include <qlineedit.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qvalidator.h>
#include <qtooltip.h>


CursorWidget::CursorWidget(QWidget *parent,const Cursor::Handle& c,OverlayList::Handle ol):QWidget(parent), m_cursor(c), m_overlayList(ol)
{
  m_cursor->attach(this);
  m_overlayList->attach(this);

  m_xVoxBox = new QSpinBox(this);
  m_yVoxBox = new QSpinBox(this);
  m_zVoxBox = new QSpinBox(this);
  m_xMmBox  = new QLineEdit(this);
  m_yMmBox  = new QLineEdit(this);
  m_zMmBox  = new QLineEdit(this);
  m_valBox  = new QLineEdit(this);

  QToolTip::add(m_xVoxBox,tr("x Voxel coordinate"));
  QToolTip::add(m_yVoxBox,tr("y Voxel coordinate"));
  QToolTip::add(m_zVoxBox,tr("z Voxel coordinate"));
         
  QToolTip::add(m_xMmBox,tr("x Coordinate in mm"));
  QToolTip::add(m_yMmBox,tr("y Coordinate in mm"));
  QToolTip::add(m_zMmBox,tr("z Coordinate in mm"));

  QToolTip::add(m_valBox, tr("Voxel value"));
                
  m_xMmBox->setAlignment(AlignLeft);
  m_yMmBox->setAlignment(AlignLeft);
  m_zMmBox->setAlignment(AlignLeft);
  m_valBox->setAlignment(AlignLeft);
  m_valBox->setReadOnly(true);

  connect(m_xVoxBox,SIGNAL(valueChanged(int)),
          this,SLOT(voxBoxChanged(int)));
  connect(m_yVoxBox,SIGNAL(valueChanged(int)),
          this,SLOT(voxBoxChanged(int)));
  connect(m_zVoxBox,SIGNAL(valueChanged(int)),
          this,SLOT(voxBoxChanged(int)));

  connect(m_xMmBox,SIGNAL(textChanged ( const QString & )), 
          this,SLOT(mmBoxChanged(const QString &)));          
  connect(m_yMmBox,SIGNAL(textChanged ( const QString & )) ,
          this,SLOT(mmBoxChanged(const QString &)));
  connect(m_zMmBox,SIGNAL(textChanged ( const QString & )) ,
          this,SLOT(mmBoxChanged(const QString &)));
  
  setInputValidators();

  m_xMmBox->setMinimumWidth( QFontMetrics(m_xMmBox->font()).width(QString("-0000.00")) );
  m_xMmBox->setMaximumWidth( QFontMetrics(m_xMmBox->font()).width(QString("-0000.00")) );
  m_yMmBox->setMinimumWidth( QFontMetrics(m_yMmBox->font()).width(QString("-0000.00")) );
  m_yMmBox->setMaximumWidth( QFontMetrics(m_yMmBox->font()).width(QString("-0000.00")) );
  m_zMmBox->setMinimumWidth( QFontMetrics(m_zMmBox->font()).width(QString("-0000.00")) );
  m_zMmBox->setMaximumWidth( QFontMetrics(m_zMmBox->font()).width(QString("-0000.00")) );

  m_valBox->setMinimumWidth( QFontMetrics(m_valBox->font()).width(QString("-0.000000e-00")) );
  m_valBox->setMaximumWidth( QFontMetrics(m_valBox->font()).width(QString("-0.000000e-00")) );
  
  QHBoxLayout* hlayout = new QHBoxLayout(this);

  hlayout->addWidget(m_xVoxBox);
  hlayout->addWidget(m_yVoxBox);
  hlayout->addWidget(m_zVoxBox);
  hlayout->addSpacing(10);
  hlayout->addWidget(m_xMmBox);
  hlayout->addWidget(m_yMmBox);
  hlayout->addWidget(m_zMmBox);
  hlayout->addWidget(new QLabel("mm", this));
  hlayout->addSpacing(10);
  hlayout->addWidget(m_valBox);

  show();

  update(m_cursor);
}

CursorWidget::~CursorWidget()
{
  m_cursor->detach(this);
  m_overlayList->attach(this);
}

void CursorWidget::update(const Cursor::Handle& c)
{
  TRACKER("CursorWidget::update(const Cursor::Handle& c");

  blockBoxSignals(true);
 
  Image::Handle image = getPairImage(m_overlayList->getMainMetaImage());

  m_xMmBox->setText(tr("%1")
  .arg((c->inqX() - image->getInfo()->inqXOrig())*image->getInfo()->inqXDim(),3,'f',2));  
  m_yMmBox->setText(tr("%1")
  .arg((c->inqY() - image->getInfo()->inqYOrig())*image->getInfo()->inqYDim(),3,'f',2));  
  m_zMmBox->setText(tr("%1")
  .arg((c->inqZ() - image->getInfo()->inqZOrig())*image->getInfo()->inqZDim(),3,'f',2));
  
  m_xVoxBox->setValue(c->inqX());
  m_yVoxBox->setValue(c->inqY());
  m_zVoxBox->setValue(c->inqZ());

  updateValBox();

  blockBoxSignals(false);
}

void CursorWidget::update(const OverlayList* ol, OverlayListMsg msg)
{
    if(msg == OverlayListMsg(Select)) updateValBox();
}

void CursorWidget::voxBoxChanged(int v)
{
  TRACKER("CursorWidget::voxBoxChanged");
  blockBoxSignals(true);
  
  Image::Handle image = getPairImage(m_overlayList->getMainMetaImage());

  m_xMmBox->setText(tr("%1")
  .arg((m_xVoxBox->value() - image->getInfo()->inqXOrig())*image->getInfo()->inqXDim(),3,'f',2));  
  m_yMmBox->setText(tr("%1")
  .arg((m_yVoxBox->value() - image->getInfo()->inqYOrig())*image->getInfo()->inqYDim(),3,'f',2));  
  m_zMmBox->setText(tr("%1")
  .arg((m_zVoxBox->value() - image->getInfo()->inqZOrig())*image->getInfo()->inqZDim(),3,'f',2));

  m_cursor->detach(this);
  m_cursor->setCursor(m_xVoxBox->value(),m_yVoxBox->value(),m_zVoxBox->value());
  m_cursor->attach(this);

  updateValBox();

  blockBoxSignals(false);
}

void CursorWidget::mmBoxChanged(const QString & s)
{
  TRACKER("CursorWidget::mmBoxChanged");
  blockBoxSignals(true);

  int pos;
  QString xText = m_xMmBox->text();
  QString yText = m_yMmBox->text();
  QString zText = m_zMmBox->text();

  if(QValidator::Intermediate == m_xBoxValidator->validate(xText,pos))
     m_xMmBox->setText(fixMmBoxVal(m_xBoxValidator,xText));
  if(QValidator::Intermediate == m_yBoxValidator->validate(yText,pos))
     m_yMmBox->setText(fixMmBoxVal(m_yBoxValidator,yText));
  if(QValidator::Intermediate == m_zBoxValidator->validate(zText,pos))
     m_zMmBox->setText(fixMmBoxVal(m_zBoxValidator,zText));

  float xMm = m_xMmBox->text().toFloat();
  float yMm = m_yMmBox->text().toFloat();
  float zMm = m_zMmBox->text().toFloat(); 

  Image::Handle image = getPairImage(m_overlayList->getMainMetaImage());

  int xCur = int(xMm/image->getInfo()->inqXDim()) + image->getInfo()->inqXOrig();
  int yCur = int(yMm/image->getInfo()->inqYDim()) + image->getInfo()->inqYOrig();
  int zCur = int(zMm/image->getInfo()->inqZDim()) + image->getInfo()->inqZOrig();

  m_xVoxBox->setValue(xCur);
  m_yVoxBox->setValue(yCur);
  m_zVoxBox->setValue(zCur);

  m_cursor->detach(this);
  m_cursor->setCursor(xCur,yCur,zCur);
  m_cursor->attach(this);

  updateValBox();

  blockBoxSignals(false);
}

void CursorWidget::blockBoxSignals(bool state)
{
  m_xVoxBox->blockSignals(state);
  m_yVoxBox->blockSignals(state);
  m_zVoxBox->blockSignals(state);  
  m_xMmBox->blockSignals(state);
  m_yMmBox->blockSignals(state);
  m_zMmBox->blockSignals(state);  
}

void CursorWidget::updateValBox()
{
  MetaImage mi = m_overlayList->getActiveMetaImage();
  Image::Handle image;
  if(isValidMetaImage(mi))
    image = getPairImage(mi);
  else
    image = getPairImage(m_overlayList->getMainMetaImage());
      
  int vol = std::min(m_cursor->inqV(), short (image->getInfo()->inqNumVolumes()-1));
  float i = image->getVolume(vol) -> value(m_cursor->inqX(),m_cursor->inqY(),m_cursor->inqZ());
  
  m_valBox->setText(tr("%1").arg(i,8));
}

void CursorWidget::setInputValidators()
{
  TRACKER("CursorWidget::setInputValidators");
  Image::Handle image = getPairImage(m_overlayList->getMainMetaImage());
  
  m_xVoxBox->setRange(0,image->getInfo()->inqX() - 1);
  m_yVoxBox->setRange(0,image->getInfo()->inqY() - 1);
  m_zVoxBox->setRange(0,image->getInfo()->inqZ() - 1);
  
  m_xBoxValidator = new QDoubleValidator(this);
  m_yBoxValidator = new QDoubleValidator(this);
  m_zBoxValidator = new QDoubleValidator(this);
  
  m_xMmBox->setValidator(m_xBoxValidator);
  m_yMmBox->setValidator(m_yBoxValidator);
  m_zMmBox->setValidator(m_zBoxValidator); 
  
  m_xBoxValidator->setRange( (- image->getInfo()->inqXOrig())*image->getInfo()->inqXDim(),
                         (image->getInfo()->inqX() - image->getInfo()->inqXOrig())*image->getInfo()->inqXDim() - 1,2);
  m_yBoxValidator->setRange( (- image->getInfo()->inqYOrig())*image->getInfo()->inqYDim(),
                         (image->getInfo()->inqY() - image->getInfo()->inqYOrig())*image->getInfo()->inqYDim() - 1,2);
  m_zBoxValidator->setRange( (- image->getInfo()->inqZOrig())*image->getInfo()->inqZDim(),
                         (image->getInfo()->inqZ() - image->getInfo()->inqZOrig())*image->getInfo()->inqZDim() - 1,2);
}

QString CursorWidget::fixMmBoxVal(QDoubleValidator* v,QString & s)
{
  TRACKER("CursorWidget::fixMmBoxVal()");
  float val = s.toFloat();

  if(val > v->top())
    {
      val = v->top();
    }
  else if(val < v->bottom())
    {
      val = v->bottom();
    }
  
  QString retStr(tr("%1").arg(val,3,'f',2));

  
  return retStr;
}
