/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include <math.h>
#include <stdio.h>
#include "graphwidget.h"
#include "graphplot.h"
#include <qpainter.h>
#include <list>
#include <qfontmetrics.h>



class GraphDim
{
public:
  GraphDim(int,int,int,int,int,int,int,int,int,int);
  ~GraphDim(){}
  
  int cLeft();
  int cRight();
  int cTop();
  int cBottom();
  int cWidth();
  int cHeight();
  int outerBorder(){return m_outerBorder;}
  int markLength(){return m_markLength;}
  int labelSpace(){return m_labelSpace;}
  int xNumSpace(){return m_xNumSpace;}
  int xNumWidth(){return m_xNumWidth;}
  int yNumWidth(){return m_yNumWidth;}
  int numHeight(){return m_numHeight;}
  int labelHeight(){return m_labelHeight;}

  void setLabelHeight(int h){m_labelHeight = h;}
  void setXNumWidth(int w){m_xNumWidth = w;}
  void setYNumWidth(int w){m_yNumWidth = w;}
  void setNumHeight(int h){m_numHeight = h;}
  void setWidgetSize(int w, int h){m_widgetWidth = w; m_widgetHeight = h;}
private: 
  int m_widgetWidth;
  int m_widgetHeight;  
  int m_labelHeight; 
  int m_numHeight;
  int m_xNumWidth;
  int m_yNumWidth;
  int m_outerBorder;  
  int m_markLength;
  int m_labelSpace;
  int m_xNumSpace;
  
  
};
  
GraphDim::GraphDim(int width, int height, int labelHeight, 
                   int numHeight, int xNumWidth,int yNumWidth,
                   int border,int mark,int space,int xNumSpace ):
                   m_widgetWidth(width), m_widgetHeight(height),m_labelHeight(labelHeight),
                   m_numHeight(numHeight),m_xNumWidth(xNumWidth),m_yNumWidth(yNumWidth),
                   m_outerBorder(border),m_markLength(mark),m_labelSpace(space),
                   m_xNumSpace(xNumSpace)
{

}
int GraphDim::cLeft()
{
  return m_outerBorder + m_labelHeight + m_labelSpace + m_yNumWidth + m_xNumSpace + m_markLength;
}
int GraphDim::cBottom()
{
  return m_widgetHeight - (m_outerBorder + m_labelHeight + m_labelSpace + m_numHeight + m_markLength);
}
int GraphDim::cRight()
{
  return m_widgetWidth - ((m_xNumWidth/2) + m_outerBorder);
}
int GraphDim::cTop()
{
  return (m_numHeight/2) + m_outerBorder;
}
int GraphDim::cWidth()
{
  return - cLeft() + cRight();
}
int GraphDim::cHeight()
{
  return - cTop() + cBottom();
}
class FindMaxMin {
public:
  FindMaxMin(){}
  void operator()(std::list<GraphPlot::Handle>::value_type g) 
  {
    m_minXValues.push_back(g->inqMinXVal());
    m_maxXValues.push_back(g->inqMaxXVal());
    m_minYValues.push_back(g->inqMinYVal());
    m_maxYValues.push_back(g->inqMaxYVal());

  }
  float inqXMinValue(){return *std::min_element(m_minXValues.begin(),m_minXValues.end());}
  float inqXMaxValue(){return *std::max_element(m_maxXValues.begin(),m_maxXValues.end());}
  float inqYMinValue(){return *std::min_element(m_minYValues.begin(),m_minYValues.end());}
  float inqYMaxValue(){return *std::max_element(m_maxYValues.begin(),m_maxYValues.end());}
private:
  std::list<float> m_minXValues;
  std::list<float> m_maxXValues;
  std::list<float> m_minYValues;
  std::list<float> m_maxYValues;

};


class DrawPlot {
public:
  DrawPlot(GraphWidget* g) : m_widget(g) {}
  void operator()(GraphPlot::Handle plot) { plot->drawPlot(m_widget); }
private:
  GraphWidget* m_widget;
};

GraphWidget::GraphWidget(QWidget* parent,QString xAxis,QString yAxis)  : 
                                             QWidget(parent),
                                             m_xAxisLabel(xAxis),m_yAxisLabel(yAxis),                                            
                                             m_penColor(green),
                                             m_xAxisMax(10.0f),m_xAxisMin(0.0f),
                                             m_yAxisMax(10.0f),m_yAxisMin(0.0f),
                                             m_fontSizeLabels(14),m_fontSizeNums(10),
                                             m_labelFont("helvetic",m_fontSizeLabels),
                                             m_numFont("arial",m_fontSizeNums)

{
 	this->setMinimumSize(50,50);
	this->resize(parent->size());               
}

GraphWidget::~GraphWidget(){}


void GraphWidget::paintEvent( QPaintEvent* )
{
  int labelHeight,numHeight,xNumWidth,yNumWidth;
  QString xStr,yStr;
  
  m_paint.begin(this);	
  findDataMaxMin(); 
  findFontSizes(labelHeight,numHeight,xNumWidth,yNumWidth);  
      
  m_gd = new GraphDim(width(),height(),labelHeight,numHeight,
                          xNumWidth,yNumWidth,5,5,2,2);      
  if(!m_plots.empty())drawChecks();
  std::for_each(m_plots.begin(),m_plots.end(),DrawPlot(this));
    

  drawAxis();
  drawLabels();

  delete(m_gd);

	m_paint.end();
}


void GraphWidget::drawAxis()
{
	m_paint.setPen( black );	
	m_paint.drawLine(m_gd->cLeft(),m_gd->cBottom(),m_gd->cRight(), m_gd->cBottom());
	m_paint.drawLine(m_gd->cLeft(),m_gd->cBottom(),m_gd->cLeft(),	 m_gd->cTop());
}


void GraphWidget::setAxisMaxMin(float xMin, float xMax, float yMin, float yMax)
{
	m_xAxisMax  = xMax;
	m_xAxisMin 	= xMin;
	m_yAxisMax  = yMax;
	m_yAxisMin 	= yMin;
}


void GraphWidget::setFontSizes(int fontSizeLabels, int fontSizeNums)
{
	m_fontSizeLabels = fontSizeLabels;
	m_fontSizeNums   = fontSizeNums;
}


void GraphWidget::drawCircle(float x, float y,int r)
{
	int 	d = 2 * r;
	m_paint.setPen( m_penColor );	
	m_paint.drawEllipse(convX(x) - r,convY(y)- r,d,d);
}

void GraphWidget::drawLine(float xA, float yA, float xB, float yB)
{
	m_paint.setPen( m_penColor );	
	m_paint.drawLine(convX(xA),convY(yA),convX(xB),convY(yB));
}

//GraphDim::GraphDim(int width, int height, int labelHeight, 
//                   int numHeight, int xNumWidth,int yNumWidth,
//                   int border,int mark,int space ):
void GraphWidget::drawChecks()
{
 	float val;
  int xTxtStep(1),yTxtStep(1);
 	QString strVal;

	m_paint.setPen( black );
  m_paint.setFont(m_numFont);
   
  int    xDivCount(int(((m_xAxisMax - m_xAxisMin)/m_xStep) + 0.00001));
  float  xDivSize(m_gd->cWidth()/float(xDivCount));
  int    yDivCount(int(((m_yAxisMax - m_yAxisMin)/m_yStep) + 0.00001));
  float  yDivSize(m_gd->cHeight()/float(yDivCount));
    

  if (xDivSize < (m_gd->xNumWidth() + 5)) xTxtStep = 2;
  int xSecFromLast = int(xDivCount/xTxtStep)*xTxtStep;

  for(int n = 0;n <= xDivCount;++n)
    {    
      val = m_xAxisMin + (n * m_xStep);
      strVal.setNum(val,'f', findDp(m_xStep));
      m_paint.drawLine(int(m_gd->cLeft() + n * xDivSize), m_gd->cBottom(),
                       int(m_gd->cLeft() + n * xDivSize), m_gd->cBottom() + m_gd->markLength());
      if(((n%xTxtStep == 0)&&(n != xSecFromLast)) || (n == 0) || (n == xDivCount))
          m_paint.drawText(int(m_gd->cLeft() + (n * xDivSize) - (m_gd->xNumWidth()/2)),
                           m_gd->cBottom() + m_gd->markLength(),	
                           m_gd->xNumWidth(), 	m_gd->numHeight(), AlignHCenter,	strVal);
    }


  if (yDivSize < (m_gd->numHeight() + 3)) yTxtStep = 2;
  int ySecFromLast = int(yDivCount/yTxtStep) * yTxtStep;

	for(int m = 0;m <= yDivCount ;m++)
	{
    val = m_yAxisMin + (m * m_yStep);
    strVal.setNum(val,'f', findDp(m_yStep));   
    m_paint.drawLine(m_gd->cLeft(),                     int(m_gd->cBottom() - int(m * yDivSize)),
                     m_gd->cLeft() - m_gd->markLength(),int(m_gd->cBottom() - int(m * yDivSize)));      

    if(((m%yTxtStep == 0)&&(m != ySecFromLast)) || (m == 0) || (m == yDivCount))
          m_paint.drawText(m_gd->outerBorder()+m_gd->labelHeight()+m_gd->labelSpace(),
                           int(m_gd->cBottom() - m * yDivSize - m_gd->numHeight()/2),
                           m_gd->yNumWidth(), m_gd->numHeight(),AlignRight,	strVal);
	}

}


void GraphWidget::drawLabels()
{

	m_paint.setPen( black );	

  if(m_yAxisLabel != "")
  {
    m_paint.setFont(m_labelFont);	
    m_paint.rotate(-90);
    m_paint.translate(-m_gd->cBottom(),0);	
    m_paint.drawText( 0,
                      m_gd->outerBorder(),
                      m_gd->cHeight(),m_gd->labelHeight(),
                      AlignHCenter,
                      m_yAxisLabel);    
    m_paint.translate(m_gd->cBottom(),0);	
    m_paint.rotate(90);
  }

  if(m_xAxisLabel != "")
  {
    m_paint.drawText( m_gd->cLeft(),
                      m_gd->cBottom() + m_gd->markLength() + 
                      m_gd->labelSpace() + m_gd->numHeight(),
                      m_gd->cWidth(),m_gd->labelHeight(),
                      AlignHCenter | AlignTop,
                      m_xAxisLabel);
  }
}

void GraphWidget::addPlot(GraphPlot::Handle plot)
{
  m_plots.push_back(plot);
  update();
}

void GraphWidget::remPlot(int n)
{
  PlotsIt pos;
  pos = std::find(m_plots.begin(),m_plots.end(),m_plots[n]);
  if(pos != m_plots.end()) m_plots.erase(pos);
  update();
}  

void GraphWidget::remAllPlots()
{
  m_plots.clear();
  update();
}

int GraphWidget::inqPlotCount()
{
  return m_plots.size();
}

void GraphWidget::findDataMaxMin()
{
  FindMaxMin f = std::for_each(m_plots.begin(),m_plots.end(),FindMaxMin());

  autoRange(f.inqYMinValue(),f.inqYMaxValue(),m_yAxisMin,m_yAxisMax,m_yStep);
  autoRange(f.inqXMinValue(),f.inqXMaxValue(),m_xAxisMin,m_xAxisMax,m_xStep);
}
  
GraphPlot::Handle GraphWidget::getPlot(int n)
{
  return m_plots[n];
}
  

float GraphWidget::findStepSize(float range)
{
  float logVal = log10(range);
  float rndVal = ceil(logVal);
  int   i      = int(rndVal) - 1;
  float ss     = pow(10,i);
  
  if((range / ss) > 10.0){ss = ss * 2.0;}
  if((range / ss) < 5.0){ss = ss/2.0;}               
  return ss;
}

void GraphWidget::autoRange(float min, float max, 
                            float &newMin, float &newMax, float &step)
{
  newMin = min;
  newMax = max;
  step = findStepSize(max - min);
  float minTop(min - step);
  float maxTop(max + step);

  float v = int(minTop/step) * step;  
  float t = int(maxTop/step) * step;

  if(!(fabs(v - minTop) < 1e-5))newMin = v;     
  if(!(fabs(t - maxTop) < 1e-5))newMax = t;      
}


int GraphWidget::findDp(float step)
{
  int result(0);
  
  if (step < 1.0f) result = int(fabs(log10(step)-1));
  
  return result;
}
int GraphWidget::convX(float x)
{
	return (int)(m_gd->cLeft() + m_gd->cWidth() * (x - m_xAxisMin)/(m_xAxisMax - m_xAxisMin));
}

int GraphWidget::convY(float y)
{
	return (int)(m_gd->cBottom() - m_gd->cHeight() * (y - m_yAxisMin)/(m_yAxisMax - m_yAxisMin));
}

void GraphWidget::findFontSizes(int & labelHeight,int & numHeight, 
                                int & xNumWidth,  int & yNumWidth)
{
  labelHeight = QFontMetrics(m_labelFont).height();
  numHeight = QFontMetrics(m_numFont).height();

  QString xStr,yStr;
  xStr.setNum(m_xAxisMax,'f', findDp(m_xStep));
  xNumWidth = QFontMetrics(m_numFont).width(xStr);
  yStr.setNum(m_yAxisMax,'f', findDp(m_yStep));
  yNumWidth = QFontMetrics(m_numFont).width(yStr);  

  if((m_xAxisLabel == "") && (m_yAxisLabel == "")) labelHeight = 0;
  if(m_plots.empty())
    {
      xNumWidth = 30;
      yNumWidth = 30;
    }
}  
