/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  Part of FSLView 1.0
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    LICENCE

FSLView v1.0 sources are released under the terms of the GPL
(http://www.gnu.org/copyleft/gpl.html). FSLView binaries use Qt
(http://www.trolltech.com) which is released under GPL for X11 and
under the "Qt non-commercial licence" for Windows.  The FSLView
binaries are released under the GPL, with the added exception that we
also give permission to link this program with the Qt non-commercial
edition, and distribute the resulting executable, without including
the source code for the Qt non-commercial edition in the source
distribution.

FSLView is distributed "AS IS" under this Licence in the hope that it
will be useful, but in order that the University as a charitable
foundation protects its assets for the benefit of its educational and
research purposes, the University makes clear that no condition is
made or to be implied, nor is any warranty given or to be implied, as
to the accuracy of the Software, or that it will be suitable for any
particular purpose or for use under any specific conditions.
Furthermore, the University disclaims all responsibility for the use
which is made of the Software. It further disclaims any liability for
the outcomes arising from using the Software.
    
 */


#if !defined(IMAGEGROUP_H)
#define IMAGEGROUP_H

#include <boost/shared_ptr.hpp>
#include <memory>
#include "storage/image.h"
#include "imagedisplaysetting.h"
#include "lookuptable.h"

typedef enum {NewLookUpTable, NewOverlay, RemOverlay, Lock} ImageGroupMsg;

class ImageGroup;

class ImageGroupObserver
{

public:
  virtual ~ImageGroupObserver() {}

  virtual void update(const ImageGroup* c, ImageGroupMsg message) = 0;

  ImageGroupObserver() {}
};

/**
 * @author David Flitney <flitney@fmrib.ox.ac.uk>
 * @author James Saunders <jim@fmrib.ox.ac.uk>
 *
 * @date   Mon Dec 23 17:25:52 2002
 * 
 * @brief ImageGroup groups a collection of images in a common
 * object. 
 *
 * A main image and several overlays with associated look up
 * tables can be stored in an ImageGroup object. It has responsibility
 * for maintaining the links between compatible and associated images.
 */
class ImageGroup
{
public:
  typedef std::list<Image::Handle> ImageList;
  typedef std::vector<LookUpTable::Handle> LutList;
  typedef boost::shared_ptr< ImageGroup > Handle;
  typedef boost::weak_ptr< ImageGroup > WeakHandle;
 
  static Handle create(Image::Handle image);
  static Handle createNullImage();

  bool addOverlay(Image::Handle& image);
  void addLookUpTable(LookUpTable::Handle lut);
  bool remOverlay(Image::Handle& image);
  ImageList::iterator begin();
  ImageList::iterator end();
  LutList::iterator beginLutList();
  LutList::iterator endLutList();
  LookUpTable::Handle getLut(int n) const;
  LookUpTable::Handle getLut(std::string const &name) const;

  int  getInitialLutCount() const;
  void attach(ImageGroupObserver* o);
  void detach(ImageGroupObserver* o);
  void notify(ImageGroupMsg message);

  int size();

  Image::Handle getMainImage();
  Image::Handle getLatestImage() const;
  LookUpTable::Handle getLatestLUT();

  int inqX();
  int inqY();
  int inqZ();
  float inqMin();
  float inqMax();
  virtual ~ImageGroup();

private:
  ImageGroup(Image::Handle& image);

  struct Implementation;
  const std::auto_ptr<Implementation> m_impl;
};


#endif
