/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "imagegroupwidget.h"
#include <qlayout.h>
#include "imagewidget.h"
#include <qcheckbox.h>
#include <qlistview.h>
#include <qslider.h>
#include <qpushbutton.h>
#include "eye.xpm"
#include <qobject.h>

class LayerListItem : public QListViewItem
{
 public:
  LayerListItem(QListView* parent, MetaImage & mi,bool main):QListViewItem(parent),m_mi(mi),m_isMainImage(main){refresh();};

  void refresh();
  MetaImage& getMetaImage()      { return m_mi;}
  bool inqIsMainImage()          { return m_isMainImage;}
  std::string inqImageName()     { return getPairImage(m_mi)->getInfo()->inqImageName();}
  float inqTransparency()        { return getPairDisplaySetting(m_mi)->inqTransparency();}  
  bool  inqVisibility()          { return getPairDisplaySetting(m_mi)->inqVisibility();}
  void setTransparency(int trans){        getPairDisplaySetting(m_mi)->setTransparency(trans/10.0f);refresh();}
  void setVisibility(bool vis)   {        getPairDisplaySetting(m_mi)->setVisibility(vis);refresh();}

 private:
  MetaImage& m_mi;
  bool m_isMainImage;
};

void LayerListItem::refresh()
{
  setText(1,inqImageName().c_str());
  setText(2,QString::number(inqTransparency()));
  if(inqVisibility()){setPixmap(0,QPixmap(eye));}
  else               {setPixmap(0,NULL);}
}


class InsertLayerItem {
public:
  InsertLayerItem(QListView* lv) : m_lv(lv) {}
  void operator()(MetaImage& mi) {new LayerListItem(m_lv, mi, false); }
private:
  QListView* m_lv;
};


ImageGroupWidget::ImageGroupWidget(ImageWidget* i): m_imageWidget(i)
{
  setCaption("Image Group");
  
  m_visibleButton = new QCheckBox("On/Off",this);
  m_layersListView = new QListView(this,"Layers");
  m_transSlider = new QSlider(QSlider::Horizontal,this,"Opacity");
  m_transSlider->setMinValue(0);
  m_transSlider->setMaxValue(10);
  
  QPushButton* upButton = new QPushButton("Up",this);
  QPushButton* downButton = new QPushButton("Down",this);
  QVBoxLayout* vlayout = new QVBoxLayout(this);
  QHBoxLayout* hlayouttop = new QHBoxLayout();
  QHBoxLayout* hlayoutbot = new QHBoxLayout();
  QVBoxLayout* vlayoutside = new QVBoxLayout();
  
  vlayout->addLayout(hlayouttop);
  vlayout->addLayout(hlayoutbot);
    
  hlayoutbot->addWidget(m_visibleButton);
  hlayoutbot->addWidget(m_transSlider);
  hlayouttop->addWidget(m_layersListView);
  hlayouttop->addLayout(vlayoutside);
  vlayoutside->addWidget(upButton); 
  vlayoutside->addWidget(downButton);
  vlayout->activate();

  m_layersListView->addColumn("On/Off");
  m_layersListView->addColumn("Name");
  m_layersListView->addColumn("Opacity");
  m_layersListView->setSorting(-1);
  m_layersListView->setAllColumnsShowFocus(true);
  updateListView();

  connect(m_layersListView,SIGNAL(selectionChanged()),
          this,SLOT(listSelectChanged()));
  connect(m_visibleButton,SIGNAL(toggled(bool)),
          this,SLOT(visibleButtonChanged(bool)));
  connect(m_layersListView,SIGNAL(doubleClicked(QListViewItem*)),
          this,SLOT(listDoubleClicked(QListViewItem*)));
  connect(m_transSlider, SIGNAL(valueChanged(int)),
          this,SLOT(transSliderChanged(int)));
  connect(upButton,SIGNAL(pressed()),
          this,SLOT(upButtonPressed())); 
  connect(downButton,SIGNAL(pressed()),
          this,SLOT(downButtonPressed()));
  show();

}



ImageGroupWidget::~ImageGroupWidget(){}

void ImageGroupWidget::updateListView()
{
  m_layersListView->clear();
  m_curItem = 0;

  new LayerListItem(m_layersListView,m_imageWidget->getMainMetaImage(),true);
 
  std::for_each(m_imageWidget->listBegin(),m_imageWidget->listEnd(),InsertLayerItem(m_layersListView));
}




void ImageGroupWidget::listSelectChanged()
{
  m_curItem = (LayerListItem*)m_layersListView->selectedItem();
  updateControls();
}

void ImageGroupWidget::visibleButtonChanged(bool state)
{
  setVisibility(state);
}

void ImageGroupWidget::updateControls()
{
  if(m_curItem)
    {
      m_visibleButton->setChecked(m_curItem->inqVisibility());
      m_transSlider->setValue((int)(m_curItem->inqTransparency() * 10));
    }
}

void ImageGroupWidget::listDoubleClicked(QListViewItem* item)
{
  if(m_curItem)
    setVisibility(!m_curItem->inqVisibility());
}

void ImageGroupWidget::setVisibility(bool state)
{
  if(m_curItem)
    {
      m_curItem->setVisibility(state);
      updateControls();
      //      m_imageGroup->notify();
    } 
}

void ImageGroupWidget::transSliderChanged(int value)
{
  if(m_curItem)
    {
      m_curItem->setTransparency(value);
      // m_imageGroup->notify();
    }
}

void ImageGroupWidget::upButtonPressed()
{
  if(m_curItem && !m_curItem->inqIsMainImage())
    { 
      LayerListItem *itemAbove = (LayerListItem*)m_curItem->itemAbove();  
     
      if(itemAbove)
        {
          std::swap( m_curItem->getMetaImage(),itemAbove->getMetaImage());
         
          m_curItem->refresh();
          itemAbove->refresh();
          m_curItem = itemAbove;
          m_layersListView->setSelected(m_curItem,true);
          //          m_imageGroup->notify();
        }
    }

}


void ImageGroupWidget::downButtonPressed()
{
   if(m_curItem && !m_curItem->inqIsMainImage())
    { 
      LayerListItem *itemBelow = (LayerListItem*)m_curItem->itemBelow();  
     
      if(!itemBelow->inqIsMainImage())
        {
          std::swap(m_curItem->getMetaImage(),itemBelow->getMetaImage());
          m_curItem->refresh();
          itemBelow->refresh();
          m_curItem = itemBelow;
          m_layersListView->setSelected(m_curItem,true);
          //m_imageGroup->notify();
        }
    }
}
