/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "lightboxwidget.h"

#include <qscrollview.h>
#include <qpixmap.h>
#include <qcheckbox.h>
#include <qtoolbutton.h>
#include <qlayout.h>
#include "overlaywidget.h"

#include "tracker.h"

#include "lightbox.xpm"

LightboxWidget::LightboxWidget(QWidget *parent,ImageGroup::Handle i,OverlayList::Handle ol, Cursor::Handle& c ) :  
  ImageWidget(parent,i,ol,c), m_image(i), m_zoom(1.0)
{
  TRACKER("LightboxWidget::LightboxWidget");
  m_sv = new QScrollView(this);
  
  setIcon( QPixmap(lightbox_xpm) );

  setCentralWidget(m_sv);
  
  m_sv->viewport()->setBackgroundColor(QColor(128, 128, 128));

  connect(m_sv->verticalScrollBar(), SIGNAL(valueChanged(int)), this, SLOT(scrolled(int)));
  connect(this, SIGNAL(zoomChanged(int)), this, SLOT(setZoom(int)));

  m_slices = SliceListHandle(new SliceList);

  for(int n = 0; n < m_image->inqZ(); ++n)
  {
    SliceWidget::Handle axial =
      SliceWidget::Handle(new AxialWidget(m_sv->viewport(), "axial", m_cursor, m_overlayList, m_briconWidget,m_drawWidget, m_undoList));
    
    m_sv->addChild(axial.get());
    m_slices->push_back(axial);

    connect(this, SIGNAL(crossHairModeChanged(bool)), axial.get(), SLOT(crossHairMode(bool)));
    connect(this, SIGNAL(modeChanged(int)),           axial.get(), SLOT(setMode(int)));
    connect(this, SIGNAL(crossHairModeChanged(bool)), axial.get(), SLOT(crossHairMode(bool)));
    connect(this, SIGNAL(resetZoom()),                axial.get(), SLOT(resetZoom()));
    
    m_cursorModeButton->setOn(true);
    emit modeChanged(SliceWidget::Cursing);

    axial->setSlice(n,m_cursor->inqV());
    axial->setSliceIsFixed(true);
  }
  
  m_crossHairCheckBox->setChecked(true);
}

LightboxWidget::~LightboxWidget()
{
  TRACKER("LightboxWidget::~LightboxWidget");
}

struct SliceHider
{
  SliceHider(const QRect &b): m_boundingBox(b) {}

  void operator() (SliceWidget::Handle s)
  {
    QRect sRect = s->geometry().normalize();

    if(m_boundingBox.intersects(sRect))
      s->enableUpdates(true);
    else
      s->enableUpdates(false);
  
	s->show();//Added to Ensure repaint at correct place
  }

  const QRect &m_boundingBox;
};

void LightboxWidget::scrolled(int v)
{
  TRACKER("LightboxWidget::scrolled");
  
  std::for_each(m_slices->begin(), m_slices->end(), SliceHider(m_sv->viewport()->geometry()));
}

struct SlicePlacer
{
  SlicePlacer(QScrollView* sv, unsigned int border, float zoom):
    m_sv(sv), m_borderPixels(border), m_zoom(zoom),
    m_x(border), m_y(border), 
    m_availableWidth(sv->visibleWidth()), m_heightIncrement(0),
    m_width(0) {}

  void operator()(SliceWidget::Handle s) 
  {
    unsigned int height = unsigned(s->inqHeight() * m_zoom);
    unsigned int width  = unsigned(s->inqWidth() * m_zoom);

    if((m_x + width) > m_availableWidth) {
	  m_width = std::max(m_width, m_x);
      m_x = m_borderPixels;
      m_y += m_heightIncrement + m_borderPixels;
      m_heightIncrement = 0;
    }

    int vx, vy;
    m_sv->contentsToViewport(m_x, m_y, vx, vy);
    s->hide(); //Added to Ensure repaint at correct place
    s->setGeometry(vx, vy, int(m_zoom * s->inqWidth()), int(m_zoom * s->inqHeight()));

    m_x += width + m_borderPixels;
    m_heightIncrement = std::max(m_heightIncrement, height);
  }

  unsigned int width() const { return m_width; }
  unsigned int height() const { return m_y + m_heightIncrement; }

  QScrollView* m_sv;
  unsigned int m_borderPixels;
  float        m_zoom;

  unsigned int m_x, m_y;
  unsigned int m_availableWidth;
  unsigned int m_heightIncrement;
  unsigned int m_width;
};

void LightboxWidget::layoutSlices() const
{
  TRACKER("LightboxWidget::layoutSlices");
  SlicePlacer sp = std::for_each(m_slices->begin(), m_slices->end(), SlicePlacer(m_sv, 6, m_zoom));
  m_sv->resizeContents(sp.width(), sp.height());
  std::for_each(m_slices->begin(), m_slices->end(), SliceHider(m_sv->viewport()->geometry()));
}

void LightboxWidget::setZoom(int factor)
{
  m_zoom = factor / 100.0;

  layoutSlices();
}

void LightboxWidget::resizeEvent(QResizeEvent *e)
{
  TRACKER("LightboxWidget::resizeEvent");
  layoutSlices();
}

void LightboxWidget::update(const Cursor::Handle& c)
{
  TRACKER("LightboxWidget::update");
  std::for_each(m_slices->begin(), m_slices->end(), SetImageCursor(c));  
}
