/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "overlaydialog.h"
#include "imagedisplaysetting.h"
#include <qlabel.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qfiledialog.h>
#include "fileopen.xpm"

class AddLutToCombo {
public:
  AddLutToCombo(QComboBox* c,LookUpTable::Handle l):m_cb(c),m_curLut(l),m_index(0)
  {
    m_cb->clear();
    m_cb->insertItem("Default");++m_index;
  }
  void operator()(LookUpTable::Handle l)
  { 
    m_cb->insertItem(l->inqLutName().c_str());    
    if(l == m_curLut){m_cb->setCurrentItem(m_index);}
    ++m_index;
  }
 
private:
  QComboBox* m_cb;
  LookUpTable::Handle m_curLut;
  int m_index;
};

OverlayDialog::OverlayDialog(QWidget* w,OverlayList::Handle l,ImageGroup::Handle i):QDialog(w,0,FALSE), m_overlayList(l),m_imageGroup(i)
{  
  setCaption("Overlay Information");  
  QFont f("Helvetica", 10);
  setFont( f );
  m_overlayList->attach(this); 
  QPixmap fileIcon = QPixmap( fileopen_xpm );
  QIconSet fileIconSet = QIconSet(fileopen_xpm);
  

  QLabel* name       = new QLabel(this);
  QLabel* voxels     = new QLabel(this);
  QLabel* dimensions = new QLabel(this);
  QLabel* volume     = new QLabel(this);
  QLabel* bits       = new QLabel(this);
  QLabel* lut        = new QLabel(this);
  
  m_grid = new QGridLayout(this, 7, 3, 10, 5);

  m_grid->addWidget(name,0,0);
  m_grid->addWidget(voxels,1,0);
  m_grid->addWidget(dimensions,2,0);
  m_grid->addWidget(volume,3,0);
  m_grid->addWidget(bits,4,0);
  m_grid->addWidget(lut,5,0);

  name->setText("Name:");
  voxels->setText("Voxels:");
  dimensions->setText("Dimensions:");
  volume->setText("Volumes:");
  bits->setText("Bits/Pixel:");
  lut->setText("Look Up Table:");

  m_overlayLabel = new QLabel(this);
  m_voxLabel = new QLabel(this); 
  m_dimLabel = new QLabel(this);
  m_volLabel = new QLabel(this);
  m_bitsLabel = new QLabel(this);
  m_lutCombo = new QComboBox(false,this);
  m_lutButton = new QPushButton(fileIcon,"",this);
  m_okButton = new QPushButton("OK",this);

  m_grid->addWidget(m_overlayLabel,0,1);
  m_grid->addWidget(m_voxLabel,1,1);
  m_grid->addWidget(m_dimLabel,2,1);
  m_grid->addWidget(m_volLabel,3,1);
  m_grid->addWidget(m_bitsLabel,4,1);
  m_grid->addWidget(m_lutCombo,5,1);

  m_grid->addWidget(m_lutButton,5,2);
  m_grid->addWidget(m_okButton,6,2);

  setMinimumSize(m_grid->minimumSize());
  setFixedSize(m_grid->minimumSize());
  
  connect(m_lutCombo, SIGNAL(activated(int)),
          this,SLOT(lutComboChanged(int)));
  connect(m_lutButton, SIGNAL(pressed()),
          this,SLOT(lutButtonPressed()));
  connect(m_okButton, SIGNAL(pressed()),
          this,SLOT(okButtonPressed()));

  updateDialog();
}

OverlayDialog::~OverlayDialog()
{  
  m_overlayList->detach(this);
}

void OverlayDialog::update(const OverlayList* l, OverlayListMsg msg)
{
  updateDialog();
}

void OverlayDialog::updateDialog()
{   
  MetaImage m_metaImage = m_overlayList->getActiveMetaImage();

  if(!isValidMetaImage(m_metaImage)) 
  {
    clearDialog();
  }
  else
  {
    ImageDisplaySetting::Handle disp = getPairDisplaySetting(m_metaImage);
    ImageInfo::Handle           info = getPairImage(m_metaImage)->getInfo();
    LookUpTable::Handle         curLut = disp->inqLookUpTable();
   
    QString voxStr(tr("%1 x %2 x %3").arg(info->inqX())
                                       .arg(info->inqY())
                                       .arg(info->inqZ()));  
    QString dimStr(tr("%1 x %2 x %3 mm").arg(info->inqXDim())
                                       .arg(info->inqYDim())
                                       .arg(info->inqZDim()));
    QString volStr(tr("%1").arg(info->inqNumVolumes()));
    QString bitsStr(tr("%1").arg(info->inqBitsPerVoxel()));
   
    m_overlayLabel->setText(info->inqImageName().c_str());
    m_voxLabel->setText(voxStr);
    m_dimLabel->setText(dimStr);
    m_volLabel->setText(volStr);
    m_bitsLabel->setText(bitsStr);
    
    std::for_each(m_imageGroup->beginLutList(),
                  m_imageGroup->endLutList(),
                  AddLutToCombo(m_lutCombo,curLut));
  }
}

void OverlayDialog::clearDialog()
{
  m_overlayLabel->setText("");
  m_lutCombo->clear();
  m_voxLabel->clear();
  m_dimLabel->clear();
  m_volLabel->clear();
  m_bitsLabel->clear();
}

void OverlayDialog::lutComboChanged(int n)
{
  LookUpTable::Handle lut;
  
  if(n){lut = m_imageGroup->getLut(n - 1);}
  m_overlayList->setLookUpTable(lut);
}

void OverlayDialog::lutButtonPressed()
{
  openLookUpTable();
}

void OverlayDialog::okButtonPressed()
{
  close();
}
