/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "overlaylist.h"
#include "tracker.h"
#include <assert.h>
#include <algorithm>
//#include <boost/format.hpp>


class CopyOverlay {
public:
  CopyOverlay(){}
  void operator()(MetaImage mi) 
  {
    MetaImage newMi;

    newMi.second = mi.second;
    newMi.first  = mi.first->clone();
    m_newList.push_back(newMi);
  }
  std::list<MetaImage> getOverlays(){return m_newList;}

private:
  std::list<MetaImage> m_newList;
};

class ImageSearch {
public:
  ImageSearch(Image::Handle i) : m_found(false),m_image(i) {}
  void operator()(MetaImage mi) 
  {
    if(mi.second ==  m_image){m_found = true;m_metaImage = mi;}
  }
  bool m_found;
  MetaImage m_metaImage;
private:
  Image::Handle m_image;
};



class AddOverlay {
public:
  AddOverlay(std::list<MetaImage>* ol, std::list<MetaImage>* nl,ImageGroup::Handle ig,int &n) 
    : m_newlist(nl), m_oldlist(ol),m_imgGrp(ig),m_curLut(n) {}
  void operator()(Image::Handle i) 
  {
    TRACKER("AddOverlay::operator()");

    ImageSearch search = std::for_each(m_oldlist->begin(),m_oldlist->end(),ImageSearch(i));

    if(search.m_found == true)
      {
	MESSAGE("Pushing an old overlay onto the new list");
        m_newlist->push_back(search.m_metaImage);
      }
    else
      {
	MESSAGE("Pushing a new overlay onto the new list");
        LookUpTable::Handle lut; 

        lut = m_imgGrp->getLut(i->getInfo()->inqLutName());
        
        if(!isValidLut(lut))lut = m_imgGrp->getLut(m_curLut++);
        
        if (m_curLut >= m_imgGrp->getInitialLutCount())m_curLut = 0;
        m_newlist->push_back(std::make_pair(ImageDisplaySetting::create(i, lut, 1.0f, true), i));    
      }
  }
  std::list<MetaImage>* m_newlist;
private:
  std::list<MetaImage>* m_oldlist;  
  ImageGroup::Handle m_imgGrp;  
  int &m_curLut;

};


OverlayList::OverlayList(ImageGroup::Handle i):m_imageGroup(i),m_currentLut(0)
{
  m_imageGroup->attach(this);
  loadOverlaysList();
}

OverlayList::~OverlayList(){m_imageGroup->detach(this);}


const MetaImage& OverlayList::getMainMetaImage() const
{
  return m_mainMetaImage;
}

OverlayList::Handle OverlayList::create(ImageGroup::Handle i)
{
  return OverlayList::Handle(new OverlayList(i));
}


const Image::Handle OverlayList::getMainImage() const 
{
  return getPairImage(m_mainMetaImage);
}

void OverlayList::loadOverlaysList()
{ 
  if (!(m_mainMetaImage.second == m_imageGroup->getMainImage()))
    {  
      LookUpTable::Handle lut =  m_imageGroup->getLut(m_imageGroup->getMainImage()->getInfo()->inqLutName()); 
      m_mainMetaImage.first = ImageDisplaySetting::create(m_imageGroup->getMainImage(), lut,1.0f,true);
      m_mainMetaImage.second = m_imageGroup->getMainImage(); 
    }
 
  m_xDim = m_mainMetaImage.second->getVolume(0)->inqX();
  m_yDim = m_mainMetaImage.second->getVolume(0)->inqY();
  m_zDim = m_mainMetaImage.second->getVolume(0)->inqZ();
    
  std::list<MetaImage> tmpOverlays;
  AddOverlay add = std::for_each(m_imageGroup->begin(),m_imageGroup->end(),AddOverlay(&m_overlays,&tmpOverlays,m_imageGroup,m_currentLut));
  m_overlays.clear();
  m_overlays = *add.m_newlist;
 
  if(m_overlays.empty())  {setActiveMetaImage(m_mainMetaImage);}
  else                    {setActiveMetaImage(m_overlays.back());}
}

struct ImageNotIn {
  ImageNotIn(ImageGroup::Handle ig): m_imageGroup(ig) {}
  
  bool operator()(const MetaImage& mi) const
  {
	Image::Handle i = mi.second;
    ImageGroup::ImageList::iterator it = std::find(m_imageGroup->begin(), m_imageGroup->end(), i);

    return (it == m_imageGroup->end());
  }

  ImageGroup::Handle m_imageGroup;
};

const MetaImageListIt OverlayList::begin()
{
  return m_overlays.begin();
}

const MetaImageListIt OverlayList::end()
{
  return m_overlays.end();
}

void OverlayList::attach(OverlayListObserver* o)
{
  m_observers.remove(o);
  m_observers.push_back(o);
}

void OverlayList::detach(OverlayListObserver* o)
{
  m_observers.remove(o);
}

struct Update
{
  Update(OverlayList* l, OverlayListMsg m): m_ol(l),m_msg(m) {}

  void operator()(OverlayListObserver* v)
  {
    v->update(m_ol,m_msg);
  }

  OverlayList*   m_ol;
  OverlayListMsg m_msg;
};


void OverlayList::notify(OverlayListMsg message)
{
  TRACKER("OverlayList::notify"); 
  std::for_each(m_observers.begin(), m_observers.end(), 
               Update(this,message));
}

const MetaImage OverlayList::getActiveMetaImage() const
{
  return m_activeMetaImage;
}

void OverlayList::assignLatestLUT()
{
  MetaImage mi = getActiveMetaImage();
  
  if(isValidMetaImage(mi))
    {
      getPairDisplaySetting(mi)->setLookUpTable(m_imageGroup->getLatestLUT());
      notify(OverlayListMsg(LookUpTable));  
    }
}

void OverlayList::update(const ImageGroup* i, ImageGroupMsg message)
{
  TRACKER("OverlayList::update(ImageGroup* i, ImageGroupMsg message");
	
  switch(message)
  {
  case ImageGroupMsg(NewLookUpTable):
    assignLatestLUT();
    notify(OverlayListMsg(LookUpTable));
    break;
  case ImageGroupMsg(NewOverlay):    
    {
      Image::Handle newImage = m_imageGroup->getLatestImage();
      LookUpTable::Handle lut = m_imageGroup->getLut(newImage->getInfo()->inqLutName());
      if(!isValidLut(lut)) lut = i->getLut(m_currentLut++);
      if (m_currentLut >= i->getInitialLutCount()) m_currentLut = 0;
      m_overlays.push_back(std::make_pair(ImageDisplaySetting::create(newImage, lut, 1.0f, true), 
					  newImage));
      if(m_overlays.empty())  {setActiveMetaImage(m_mainMetaImage);}
      else                    {setActiveMetaImage(m_overlays.back());}
      notify(OverlayListMsg(Add));
    }
    break;
  case ImageGroupMsg(RemOverlay): 

    m_overlays.erase(std::remove_if(m_overlays.begin(),m_overlays.end(),ImageNotIn(m_imageGroup)),
					 m_overlays.end());
	
	if(m_overlays.empty())  {setActiveMetaImage(m_mainMetaImage);}
    else                    {setActiveMetaImage(m_overlays.back());}
    notify(OverlayListMsg(Rem));
    break;
  case ImageGroupMsg(Lock):          notify(OverlayListMsg(Security));
                                     break;
  }

}

void OverlayList::setActiveMetaImage(const MetaImage& mi)
{    
  TRACKER("OverlayList::setActiveMetaImage");
  m_activeMetaImage = mi;
  notify(OverlayListMsg(Select));
}

void OverlayList::setTransparency(float trans)
{
  if(isValidMetaImage(m_activeMetaImage))
    {
    getPairDisplaySetting(m_activeMetaImage)->setTransparency(trans);
    notify(OverlayListMsg(Transparency));
    }
}

void OverlayList::setVisibility(bool state)
{
  if(isValidMetaImage(m_activeMetaImage))
    {    
    getPairDisplaySetting(m_activeMetaImage)->setVisibility(state);
    notify(OverlayListMsg(Visibility));
    }
}  

void OverlayList::setReadOnly(bool state)
{
  if(isValidMetaImage(m_activeMetaImage))
    {    
    getPairImage(m_activeMetaImage)->getInfo()->setReadOnly(state);
    notify(OverlayListMsg(Security));
    m_imageGroup->notify(ImageGroupMsg(Lock));
    }
}

void OverlayList::moveOverlayUp()
{
  TRACKER("OverlayList::moveOverlayUp");
  if(isValidMetaImage(m_activeMetaImage) && 
     m_activeMetaImage != m_mainMetaImage)
    {
     MetaImageListIt cur = std::find(m_overlays.begin(),
                                     m_overlays.end(),
                                     m_activeMetaImage);
     assert(cur != m_overlays.end());
  
     MetaImageListIt next(cur);
          
     if(++next != m_overlays.end())
       {
         std::swap(*cur,*next);              
         notify(OverlayListMsg(Order));
       }
    }
}

void OverlayList::moveOverlayDown()
{
  TRACKER("OverlayList::moveOverlayDown");
  if(isValidMetaImage(m_activeMetaImage) &&
     m_activeMetaImage != m_mainMetaImage)
  {
     MetaImageListIt cur = std::find(m_overlays.begin(),
                                     m_overlays.end(),
                                     m_activeMetaImage);
     assert(cur != m_overlays.end());
        
     MetaImageListIt prev(cur);
          
     if(cur != m_overlays.begin())
     {       
       --prev;
       std::swap(*cur,*prev);              
       notify(OverlayListMsg(Order));
     }
   }
}

void OverlayList::setLookUpTable(LookUpTable::Handle lut)
{
  if(isValidMetaImage(m_activeMetaImage))
  {
    getPairDisplaySetting(m_activeMetaImage)->setLookUpTable(lut);
    notify(OverlayListMsg(LookUpTable));
  }
}

Image::Handle OverlayList::inqActiveImage()
{
  return getPairImage(m_activeMetaImage);
}

OverlayList::Handle OverlayList::clone()
{    
  OverlayList::Handle hnd =  Handle(new OverlayList(m_imageGroup));
  hnd->setOverlays(m_overlays);
  hnd->setMainMetaImage(m_mainMetaImage);
  return hnd;
}

void OverlayList::setOverlays(std::list<MetaImage> list)
{    
  CopyOverlay copyOverlay = std::for_each(list.begin(),list.end(),CopyOverlay());
  m_overlays = copyOverlay.getOverlays(); 
 
 if(!m_overlays.empty()){setActiveMetaImage(m_overlays.back());}
}

void OverlayList::setMainMetaImage(MetaImage & mi)
{
  m_mainMetaImage.second = mi.second;
  m_mainMetaImage.first  = mi.first->clone(); 
  setActiveMetaImage(m_mainMetaImage);
}
