/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "overlaywidget.h"
#include <qlayout.h>
#include <qcheckbox.h>
#include <qlistview.h>
#include <qslider.h>
#include <qpushbutton.h>
#include <qobject.h>
#include <qheader.h>
#include <qlabel.h>
#include <qtooltip.h>

#include "glinfo.h"
#include "application.h"
#include "tracker.h"

#include "info.xpm"
#include "eye.xpm"
#include "padlock.xpm"
#include "uparrow.xpm"
#include "downarrow.xpm"

const char * toggleVisibiltyText = "Visibility on/off.<br><hr>Hides/shows the currently highlighted overlay.";

const char * toggleLockText = "Lock on/off.<br><hr>Images must be unlocked before they are edited.";

const char * transparencySliderEnabled = "Transparency slider.<br><hr>Controls the transparency of the currently selected overlay.";

const char * transparencySliderDisabled = "Transparency slider(Disabled).<br><hr>Controls the transparency of the currently selected overlay. This feature is not supported by the current graphics driver.";

const char * infoButtonText = "Open information dialog.<br><hr>Shows details of the currently selected image and allows the look up table to be changed.";

const char * upButtonText = "Move overlay up.<br><hr>Moves the currently selected overlay so that it appears on top of other overlays.";

const char * downButtonText = "Move overlay down.<br><hr>Moves the currently selected overlay so that it appears below other overlays.";


class LayerListItem : public QListViewItem
{
 public:
  LayerListItem(QListView* parent, const  MetaImage & mi,bool main):QListViewItem(parent),m_mi(mi),m_isMainImage(main){refresh();};

  void refresh();
  const MetaImage& getMetaImage(){ return m_mi;}
  bool inqIsMainImage()          { return m_isMainImage;}
  std::string inqImageName()     { return getPairImage(m_mi)->getInfo()->inqImageName();}
  float inqTransparency()        { return getPairDisplaySetting(m_mi)->inqTransparency();}  
  bool  inqVisibility()          { return getPairDisplaySetting(m_mi)->inqVisibility();}
  bool  inqReadOnly()            { return getPairImage(m_mi)->getInfo()->inqReadOnly();}

 private:
  const MetaImage& m_mi;
  bool m_isMainImage;
};

void LayerListItem::refresh()
{
  setText(2,inqImageName().c_str());
  setText(3,QString::number(inqTransparency()));
  if(inqVisibility()){setPixmap(0,QPixmap(eye));}
  else               {setPixmap(0,NULL);}
  if(inqReadOnly())  {setPixmap(1,QPixmap(padlock));}
  else               {setPixmap(1,NULL);}

}


class InsertLayerItem {
public:
  InsertLayerItem(QListView* lv) : m_lv(lv) {}
  void operator()(MetaImage& mi);

private:
  QListView* m_lv;
};

void InsertLayerItem::operator()(MetaImage& mi) 
{
   new LayerListItem(m_lv, mi, false);
}



OverlayWidget::OverlayWidget(QWidget* w,OverlayList::Handle l):QWidget(w,0), m_overlayList(l),m_blockEvents(false),m_blockSliderUpdate(false),m_sliderDisabled(false)
{
  m_overlayList->attach(this);

  setCaption("Image Group");
  
  m_visibleButton = new QCheckBox(this);
  m_unlockButton = new QCheckBox(this);

  QLabel*  visibleLabel = new QLabel(this);
  QLabel*  unlockLabel  = new QLabel(this);
  visibleLabel->setPixmap(QPixmap(eye));
  unlockLabel->setPixmap(QPixmap(padlock));

  m_layersListView = new QListView(this,"Layers");
  m_transSlider = new QSlider(QSlider::Horizontal,this,"Opacity");
  m_transSlider->setMinValue(0);
  m_transSlider->setMaxValue(10);
  m_transSlider->setPageStep(1);
  QPixmap upIcon = QPixmap( uparrow );
  m_upButton = new QPushButton(upIcon,"",this);
  m_upButton->setMaximumSize(25,25);
  QPixmap downIcon = QPixmap( downarrow );
  m_downButton = new QPushButton(downIcon,"",this);
  m_downButton->setMaximumSize(25,25); 
  QPixmap infoIcon = QPixmap( info );
  QPushButton* detailsButton = new QPushButton(infoIcon,"",this);
  detailsButton->setMaximumSize(25,25);

  QHBoxLayout* hlayout = new QHBoxLayout(this);
  QVBoxLayout* vlayout = new QVBoxLayout();
  QHBoxLayout* hlayouttop = new QHBoxLayout();
  QHBoxLayout* hlayoutbot = new QHBoxLayout();
  QVBoxLayout* vlayoutside = new QVBoxLayout();
  
  hlayout->addLayout(vlayout);
  vlayout->addLayout(hlayouttop);
  vlayout->addLayout(hlayoutbot);
  hlayout->addLayout(vlayoutside);


  hlayoutbot->addWidget(visibleLabel);
  hlayoutbot->addWidget(m_visibleButton);
  hlayoutbot->addWidget(unlockLabel);
  hlayoutbot->addWidget(m_unlockButton);
  hlayoutbot->addWidget(m_transSlider);
  hlayouttop->addWidget(m_layersListView);

  vlayoutside->addWidget(m_upButton,0,Qt::AlignRight); 
  vlayoutside->addWidget(m_downButton,0,Qt::AlignRight);
  vlayoutside->addWidget(detailsButton,0,Qt::AlignRight);
  vlayout->activate();

  m_layersListView->header()->hide();
  m_layersListView->setMaximumSize(225,65);
  m_layersListView->addColumn("",20);
  m_layersListView->addColumn("",20);
  m_layersListView->addColumn("",150);
  m_layersListView->addColumn("",30);
  m_layersListView->setSorting(-1);
  m_layersListView->setAllColumnsShowFocus(true);
  updateListView();

  connect(m_layersListView,SIGNAL(selectionChanged()),          
          this ,SLOT(listSelectChanged()));

  connect(m_visibleButton,SIGNAL(toggled(bool)),
        this,SLOT(visibleButtonChanged(bool)));  
  connect(m_unlockButton,SIGNAL(toggled(bool)),
        this,SLOT(unlockButtonChanged(bool)));
  connect(m_layersListView,SIGNAL(doubleClicked(QListViewItem*)),
        this,SLOT(listDoubleClicked(QListViewItem*)));
  connect(m_transSlider, SIGNAL(valueChanged(int)),
        this,SLOT(transSliderChanged(int)));
  connect(m_upButton,SIGNAL(pressed()),
        this,SLOT(upButtonPressed())); 
  connect(m_downButton,SIGNAL(pressed()),
        this,SLOT(downButtonPressed()));  
  connect(detailsButton,SIGNAL(pressed()),
        this,SLOT(detailsButtonPressed()));

  GLInfo::Handle gli = GLInfo::create();
  m_sliderDisabled   = !gli->imagingSubset();
  m_transSlider->setEnabled(!m_sliderDisabled);

  QToolTip::add(m_visibleButton,tr(toggleVisibiltyText));  
  QToolTip::add(m_unlockButton, tr(toggleLockText));
  QToolTip::add(detailsButton,  tr(infoButtonText));
  QToolTip::add(m_upButton,     tr(upButtonText));
  QToolTip::add(m_downButton,   tr(downButtonText));
  
  if (m_sliderDisabled)
    {QToolTip::add(m_transSlider,  tr(transparencySliderDisabled));}
  else
    {QToolTip::add(m_transSlider,  tr(transparencySliderEnabled));}



  updateUpDownButtons();
  updateControls();
  show();
}

OverlayWidget::~OverlayWidget()
{
  m_overlayList->detach(this);
}

void OverlayWidget::updateListView()
{
  TRACKER("OverlayWidget::updateListView()");
  QListViewItem* curItem = NULL;

  m_layersListView->clear();

  new LayerListItem(m_layersListView,
                    m_overlayList->getMainMetaImage(),
                    true);
 
  std::for_each(m_overlayList->begin(),
                m_overlayList->end(),
                InsertLayerItem(m_layersListView));

  curItem = getLayerItem(m_overlayList->getActiveMetaImage());

  if(curItem)
    m_layersListView->setSelected(curItem, true);
}


void OverlayWidget::update(const OverlayList* l, OverlayListMsg msg)
{
  m_blockEvents = true;

  switch(msg)
  {
   case OverlayListMsg(Select):       updateControls();break;
   case OverlayListMsg(Visibility):
   case OverlayListMsg(Transparency): updateListItem();updateControls();break;
   case OverlayListMsg(Security):
   case OverlayListMsg(Order):        updateListView();updateControls();break;
   case OverlayListMsg(Add):
   case OverlayListMsg(Rem):          updateListView();updateControls();updateUpDownButtons();break;
   case OverlayListMsg(LookUpTable):
   break;
  }
  m_blockEvents = false;
     
}

void OverlayWidget::listSelectChanged()
{
  TRACKER("OverlayWidget::listSelectChanged");
  if(m_blockEvents) return;

  LayerListItem* i = (LayerListItem*)m_layersListView->selectedItem();
  MetaImage mi;

  if(i){mi = i->getMetaImage();}

  updateUpDownButtons();

  m_overlayList->setActiveMetaImage(mi);
}

QListViewItem* OverlayWidget::getLayerItem(const MetaImage & mi)const
{
  bool success(false);
  QListViewItem* itemFound = NULL;
  QListViewItemIterator it( m_layersListView );
    
  while(it.current() && !success)   
  {
    if(((LayerListItem*)it.current())->getMetaImage() == mi)
      {
      itemFound = it.current(); 
      success = true;
      }

    ++it;
  }

  return itemFound;
}

void OverlayWidget::transSliderChanged(int value)
{
  if(m_blockEvents) return;
  m_blockSliderUpdate = true;
  m_overlayList->setTransparency((float)value/10.0f);
  m_blockSliderUpdate = false;
}

void OverlayWidget::visibleButtonChanged(bool state)
{  
  if(m_blockEvents) return;
  m_overlayList->setVisibility(state);
}

void OverlayWidget::unlockButtonChanged(bool state)
{  
  if(m_blockEvents) return;
  m_overlayList->setReadOnly(state);
}

void OverlayWidget::listDoubleClicked(QListViewItem* item)
{
   if(m_blockEvents) return; 
 bool state;

  state = getPairDisplaySetting(m_overlayList->getActiveMetaImage())->inqVisibility();
  m_overlayList->setVisibility(!state);
}

void OverlayWidget::upButtonPressed()
{
   m_overlayList->moveOverlayUp();
}

void OverlayWidget::downButtonPressed()
{
   m_overlayList->moveOverlayDown();
}

void OverlayWidget::updateUpDownButtons()
{   
  LayerListItem* i = (LayerListItem*)m_layersListView->selectedItem();
  MetaImage mi;
  
  if(i){mi = i->getMetaImage();} 
  if(mi == m_overlayList->getMainMetaImage() || i == 0)
    {
      m_upButton->setEnabled(false);
      m_downButton->setEnabled(false);
    }
  else
    {
      m_upButton->setEnabled(true);
      m_downButton->setEnabled(true);
    }
}

void OverlayWidget::updateListItem()
{
       LayerListItem* curOverlay = (LayerListItem*)getLayerItem(m_overlayList->getActiveMetaImage()); 
       if(curOverlay) 
         curOverlay->refresh();
}

void OverlayWidget::updateControls()
{
  TRACKER("OverlayWidget::updateControls");    
  
  LayerListItem* curOverlay = (LayerListItem*)getLayerItem(m_overlayList->getActiveMetaImage()); 

       if(curOverlay)
         {
           if(!m_blockSliderUpdate)
             m_transSlider->setValue((int)(curOverlay->inqTransparency() * 10));
           m_visibleButton->setChecked(curOverlay->inqVisibility());
           m_unlockButton->setChecked(curOverlay->inqReadOnly());
         }
}

void OverlayWidget::detailsButtonPressed()
{
  TRACKER("OverlayWidget::detailsButtonPressed");
  infoButtonAction();
}
