/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "shape.h"
#include "slicewidget.h"
#include <algorithm>
#include <qobject.h>

class DrawVertex {
public:
  DrawVertex()  {}
  void operator()(Pixel::Handle vo) 
  {
    glColor3f(1.0f,1.0f,1.0f); 
    glVertex3f(vo->inqX()+0.5f,vo->inqY()+0.5f,0.0f);
  }
private:
  LookUpTable::Handle m_lut;
};


class CommitVertex {
public:
  CommitVertex(Volume::Handle vol,int orient,int s) : m_volume(vol),
                                                      m_orient(orient),
                                                      m_slice(s)  {}
  void operator()(Pixel::Handle vo) 
  {  
    if(m_orient == SliceWidget::Axial)   m_volume->setValue(vo->inqX(),vo->inqY(),m_slice,vo->inqVal());
    if(m_orient == SliceWidget::Sagittal)m_volume->setValue(m_slice,vo->inqX(),vo->inqY(),vo->inqVal());
    if(m_orient == SliceWidget::Coronal) m_volume->setValue(vo->inqX(),m_slice,vo->inqY(),vo->inqVal());
  }
private:
  Volume::Handle m_volume;
  int m_orient;
  int m_slice;
};

class PixelRead {
public:
  PixelRead(Volume::Handle vol,int orient,int s,
            std::list<Pixel::Handle>& v) :  m_volume(vol),
                                            m_orient(orient),
                                            m_slice(s),
                                            m_volPixels(v)
                                            {}
  void operator()(Pixel::Handle pix) 
  {  
      Pixel::Handle newPixel;

      if(m_orient == SliceWidget::Axial)
        {   
  
            if(pix->inqX() >= 0 && pix->inqX() < m_volume->inqX() &&
               pix->inqY() >= 0 && pix->inqY() < m_volume->inqY() &&
               m_slice     >= 0 && m_slice     < m_volume->inqZ())
              {
               newPixel = Pixel::create(pix->inqX(),pix->inqY(),m_volume->value(pix->inqX(),pix->inqY(), m_slice));  
               m_volPixels.push_back(newPixel);
              }
        }
      if(m_orient == SliceWidget::Sagittal)
        {
         
            if(m_slice     >= 0 && m_slice     < m_volume->inqX() &&
               pix->inqX() >= 0 && pix->inqX() < m_volume->inqY() &&
               pix->inqY() >= 0 && pix->inqY() < m_volume->inqZ()) 
              {
               newPixel = Pixel::create(pix->inqX(),pix->inqY(),m_volume->value(m_slice,    pix->inqX(), pix->inqY()));
               m_volPixels.push_back(newPixel);
              }
        }
      
      if(m_orient == SliceWidget::Coronal) 
        {            
            if(pix->inqX() >= 0 && pix->inqX() < m_volume->inqX() &&
               m_slice     >= 0 && m_slice     < m_volume->inqY() &&
               pix->inqY() >= 0 && pix->inqY() < m_volume->inqZ()) 
              { 
                newPixel = Pixel::create(pix->inqX(),pix->inqY(),m_volume->value(pix->inqX(),m_slice,     pix->inqY()));
                m_volPixels.push_back(newPixel);
              }
        }

    
  }
private:
  Volume::Handle m_volume;
  int m_orient;
  int m_slice;
  std::list<Pixel::Handle>& m_volPixels;
};


Shape::Shape(Volume::Handle vol, int orient, int slice):
                                  m_volume(vol),
                                  m_orient(orient),
                                  m_slice(slice){}
Shape::~Shape(){}

Shape::Handle Shape::create(Volume::Handle vol, int orient, int slice)
{
  Handle dst(new Shape(vol,orient,slice));
  return dst;
}

void Shape::draw()
{
  glPointSize(2.0f);

  glBegin(GL_POINTS);
  std::for_each(m_pixels.begin(),m_pixels.end(),DrawVertex());
  glEnd();  

}

void Shape::commit()
{
  std::for_each(m_pixels.begin(),m_pixels.end(),CommitVertex(m_volume,m_orient,m_slice));
  m_pixels.clear();
}

void Shape::addVertex(int x, int y, int size, float val)
{
  Pixel::Handle vox = Pixel::create(x,y,val);


  for(int nx = x; nx < x + size;nx++)
    {
    for(int ny = y; ny < y + size;ny++)
      {
        Pixel::Handle vox = Pixel::create(nx - size/2,ny - size/2,val);
        m_pixels.push_back(vox);
      }
    }
}


Shape::Handle Shape::getBuffer()
{
  Shape::Handle hnd; 
  hnd = Handle(new Shape(m_volume,m_orient,m_slice));
  std::for_each(m_pixels.begin(),m_pixels.end(),PixelRead(m_volume,m_orient,m_slice,hnd->m_pixels));
  
  return hnd; 
}




std::list<Pixel::Handle>::iterator Shape::getPixelsBegin(){return m_pixels.begin();}

std::list<Pixel::Handle>::iterator Shape::getPixelsEnd(){return m_pixels.end();}


Pixel::Handle Pixel::create(int tx, int ty, float tval)
{
  Handle dst(new Pixel(tx,ty,tval));
  return dst;
}

Pixel::Pixel(int tx,int ty, float tval):m_x(tx),m_y(ty),m_val(tval)
{

}

