/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */


/****************************************************************************
** $Id: slicewidget.cpp,v 1.135 2003/07/10 14:39:35 jim Exp $
**
** Copyright (C) 2002 University of Oxford.  All rights reserved.
**
** FSLView
**
*****************************************************************************/

#if defined(WIN32)
#pragma warning (disable:4786)
#endif

#undef DEBUGGING

#include <qbitmap.h>

#include <algorithm>
#include <list>
#include "slicewidget.h"
#include "imagewidget.h"
#include "overlaylist.h"
#include "tracker.h"
#include "rect.h"
#include <assert.h>
#include <math.h>
#include "crosscursor.xpm"
#include "crossmask.xpm"
#include "zoomcursor.xpm"
#include "zoommask.xpm"
#include "pancursor.xpm"
#include "pencursor.xpm"
#include "penmask.xpm"
#include "panmask.xpm"
#include "shape.h"
#include "qtimer.h"

#ifdef WIN32
	#define WIDGET_UPDATE QWidget::repaint();
	#include <gl/glext.h>
#else
  #define WIDGET_UPDATE QWidget::repaint();//Mac 
//#define WIDGET_UPDATE QWidget::update(); //Linux

#endif

SliceWidget::OverlayPair   loadDataList(MetaImage mi)
{
  SliceWidget::OverlayPair dst;
  dst.first = mi;
  return dst;
}

QCursor* SliceWidget::m_crossCursor = 0;
QCursor* SliceWidget::m_zoomCursor = 0;
QCursor* SliceWidget::m_panCursor = 0;
QCursor* SliceWidget::m_penCursor = 0;

SliceWidget::SliceWidget(QWidget* parent, const char *name, Orientation orient, 
			 Cursor::Handle& c, OverlayList::Handle l, BriConWidget* b, DrawWidget* d, std::list<Shape::Handle>& u): 
  QGLWidget(parent, name),
  m_sliceIsFixed(false), m_updatesEnabled(true), m_crossHairsOn(true),m_imagesEnabled(true), m_slice(0), m_noSliceSet(true),m_volume(0), m_cursor(c),
  m_zoom(1.0),m_origX(0),m_origY(0),m_scaleX(1.0),m_scaleY(1.0),
  m_mainPixels(0), m_overlayList(l), m_briconWidget(b),m_drawWidget(d),m_undoList(u),
  m_orient(orient),m_mode(None),m_zooming(false),m_forceRender(false), 
  m_bufferValid(false)
{
  TRACKER("SliceWidget::SliceWidget");  

  m_cursor->attach(this);
  m_overlayList->attach(this);
  loadOverlays();
  m_briconWidget->attach(this);
  m_drawWidget->attach(this);
  setFocusPolicy(StrongFocus);
  m_prevCursor.reset();

  if(!m_crossCursor)
    {
      QPixmap pixmap = QPixmap(crosscursor_xpm);
      QBitmap mask;
      mask = QPixmap(crossmask_xpm);
      pixmap.setMask(mask);

      m_crossCursor = new QCursor(pixmap, 9, 9);
    }

  if(!m_zoomCursor)
    {
      QPixmap zoomPixmap = QPixmap(zoomcursor_xpm);
      QBitmap zoomMask;
      zoomMask = QPixmap(zoommask_xpm);
      zoomPixmap.setMask(zoomMask);

      m_zoomCursor = new QCursor(zoomPixmap, 9, 9);
    }

  if(!m_panCursor)
    {
      QPixmap panPixmap = QPixmap(pancursor_xpm);
      QBitmap panMask;
      panMask = QPixmap(panmask_xpm);
      panPixmap.setMask(panMask);

      m_panCursor = new QCursor(panPixmap, 9, 9);
    }
 
  if(!m_penCursor)
    {
      QPixmap penPixmap = QPixmap(pencursor_xpm);
      QBitmap penMask;
      penMask = QPixmap(penmask_xpm);
      penPixmap.setMask(penMask);

      m_penCursor = new QCursor(penPixmap, 1, 1);
    }

  setCursor(*m_crossCursor);
  show();

 
  m_timer = new QTimer(this);
  connect( m_timer, SIGNAL(timeout()), this, SLOT(showSlice()) );
  m_timer->start(50,false);
}

SliceWidget::~SliceWidget()
{
  TRACKER("SliceWidget::~SliceWidget");
  m_cursor->detach(this);
  m_overlayList->detach(this);
  m_briconWidget->detach(this);
  m_drawWidget->detach(this);
}

void SliceWidget::initializeGL()
{
  glClearColor(0.2f, 0.2f, 0.2f, 0.0f);
  glShadeModel(GL_FLAT);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 
  
  m_viewRect = Rect::createRect(0,0,inqWidth(),inqHeight());
  m_zoomRect = Rect::createRect(0,0,0,0);
  m_dataRect = Rect::createRect(0,0,inqWidth(),inqHeight());

#ifdef WIN32
  m_glBlendColorEXT = (PFNGLBLENDCOLORPROC) wglGetProcAddress ("glBlendColor"); 
#endif
}

void SliceWidget::resizeGL(int w, int h)
{  
  TRACKER("SliceWidget::resizeGL");

  initZoom();
  setCoorModifiers();

   glViewport((GLsizei)m_origX, (GLsizei)m_origY, 
              (GLsizei)(inqWidth()*m_zoom), 
              (GLsizei)(inqHeight()*inqRatio()*m_zoom));

}

void SliceWidget::initZoom()
{
  int l  = std::min(width(), height());
  m_zoom = std::min( (l/(float)inqWidth()), 
	  std::min( (l/((float)inqHeight()*(float)inqRatio())), 
		     (l/((float)inqDepth()*(float)depthRatio()))) );
}

void SliceWidget::enableUpdates(bool enabled)
{
  m_updatesEnabled = enabled;
}

void SliceWidget::paintGL()
{
  if(!m_updatesEnabled){ m_imagesEnabled = true; return;}
  TRACKER("SliceWidget::paintGL");

  if(!bufferValid()) 
    renderBuffer();  

  glLoadIdentity();
  
  if(m_imagesEnabled)glClear(GL_COLOR_BUFFER_BIT);  

  glOrtho((GLdouble)m_viewRect->left(),  (GLdouble)m_viewRect->right(),
          (GLdouble)m_viewRect->bottom(),(GLdouble)m_viewRect->top(), 
            -0.5f,0.5f); 

  if(m_imagesEnabled){m_prevCursor.reset();paintImages();}
  
  if(!m_zooming && (m_mode == Cursing) && m_crossHairsOn)drawCrossHairs();
  if(!m_zooming && (m_mode == Masking) && m_crossHairsOn && (m_shape.get() == NULL)) drawCrossHairs();
  if(!m_zooming && (m_mode == Masking) && m_crossHairsOn && (m_shape.get() != NULL)) {if(m_shape->empty())drawCrossHairs();}


  if(m_zooming) drawZoomRectangle();   
  if(m_mode == Masking) paintGraphics();

  glPixelZoom(1.0, 1.0);
  glFlush();

  m_forceRender = false;
  m_imagesEnabled = true;
}
void SliceWidget::paintGraphics()
{

  if(m_shape.get() != NULL)m_shape->draw();
}


void SliceWidget::paintImages()
{
  glRasterPos2f(m_dataRect->left()+0.0001,m_dataRect->bottom()+0.0001);

  float pixZoomX = inqWidth()/(float)m_viewRect->width();
  float pixZoomY = (inqHeight()/(float)m_viewRect->height());
  
  glPixelZoom(m_zoom * pixZoomX ,m_zoom * pixZoomY * inqRatio());
  glPixelStorei(GL_UNPACK_ROW_LENGTH,inqWidth());
  glPixelStorei(GL_UNPACK_SKIP_ROWS,m_dataRect->bottom());
  glPixelStorei(GL_UNPACK_SKIP_PIXELS,m_dataRect->left());

  float mainTrans = getPairDisplaySetting(m_overlayList->getMainMetaImage())->inqTransparency();
  if(mainTrans != 1.0)
    {
      glEnable(GL_BLEND);

	  glBlendFunc(GL_CONSTANT_ALPHA,GL_ONE_MINUS_CONSTANT_ALPHA);
	
	  #ifdef WIN32
		if(m_glBlendColorEXT)(*m_glBlendColorEXT)(1.0f,1.0f,1.0f,mainTrans);
	  #else	  
		 glBlendColor(1.0f,1.0f,1.0f,mainTrans);
	  #endif   
	}
  
  if (getPairDisplaySetting(m_overlayList->getMainMetaImage())->inqLookUpTable().get() != NULL)
     {   
       glEnable(GL_ALPHA_TEST);
       glAlphaFunc(GL_GREATER, 0.0);
     }

  if(m_mainPixels.get() && getPairDisplaySetting(m_overlayList->getMainMetaImage())->inqVisibility())
       glDrawPixels(m_dataRect->width(), m_dataRect->height(), GL_RGBA, GL_UNSIGNED_BYTE, m_mainPixels.get());   

  if(m_imageDataList.size() > 0)
    {
      glEnable(GL_BLEND);
      glBlendFunc(GL_CONSTANT_ALPHA,GL_ONE_MINUS_CONSTANT_ALPHA);
      glEnable(GL_ALPHA_TEST);
      glAlphaFunc(GL_GREATER, 0.0);
    }

  for(ImageDataList::iterator i = m_imageDataList.begin();i != m_imageDataList.end();i++)
    {
      if(getImageDisplaySetting(i)->inqVisibility())
        {		  
		#ifdef WIN32
			if(m_glBlendColorEXT)(*m_glBlendColorEXT)(1.0f,1.0f,1.0f,getImageDisplaySetting(i)->inqTransparency());	
		#else
			glBlendColor(1.0f,1.0f,1.0f,getImageDisplaySetting(i)->inqTransparency());
		#endif
		  
		  
		  glDrawPixels(m_dataRect->width(), m_dataRect->height(), GL_RGBA, GL_UNSIGNED_BYTE, getPixels(i).get());   
        }
    }

  glDisable(GL_BLEND);
  glDisable(GL_ALPHA_TEST);
 
}

void SliceWidget::crossHairMode(bool mode)
{
  m_crossHairsOn = mode;

  WIDGET_UPDATE;
}

void SliceWidget::setCoorModifiers()
{
  m_origX  = int(width() -  (inqWidth()            *m_zoom))/2;
  m_origY  = int(height() - (inqHeight()*inqRatio()*m_zoom))/2;  
  m_scaleX = (float)(inqWidth()*m_zoom)/(float)m_viewRect->width();
  m_scaleY = (float)(inqHeight()*inqRatio()*m_zoom)/(float)m_viewRect->height();
}

GLubyte SliceWidget::clamp(float c) const
{
  return GLubyte( std::max(0.0, std::min(255.0, c * 255.0)) );
}

/** 
 * A fixed slice will not change with the cursor. Use this if
 * you want to force the slice to stay the same regardless of the
 * cursor "depth" value.
 * 
 * @param fixed true if the slice is fixed.
 */
void SliceWidget::setSliceIsFixed(bool fixed)
{
  m_sliceIsFixed = fixed;
}

void SliceWidget::setSlice(int s, int v)
{  
  TRACKER("SliceWidget::setSlice");

  m_imagesEnabled = false;

  if((m_slice != s)||(m_volume != v) || (m_noSliceSet) || (m_forceRender))
  {
    if(!m_sliceIsFixed) 
    {
      m_slice = s;
      m_volume = v;
      m_noSliceSet = false;
      invalidateBuffer();
      m_imagesEnabled = true;
    }
  }

  WIDGET_UPDATE;
}
    
 



void SagittalWidget::setImageCursor(int x, int y, int z, int v)
{  
  TRACKER("SagittalWidget::setImageCursor");
  setSlice(x,v);
}

void CoronalWidget::setImageCursor(int x, int y, int z, int v)
{  
  TRACKER("CoronalWidget::setImageCursor");
  setSlice(y,v);
}

void AxialWidget::setImageCursor(int x, int y, int z, int v)

{  
  TRACKER("AxialWidget::setImageCursor");
  setSlice(z,v);
}

void SliceWidget::update(const Cursor::Handle& c)
{
  TRACKER("SliceWidget::update(const Cursor* c)");
  
  if(c->inqRepaint()){m_forceRender = true;invalidateBuffer();}
 
  setImageCursor(c->inqX(), c->inqY(), c->inqZ(),c->inqV());
}

void SliceWidget::update(const BriConWidget* b)
{
  TRACKER("SliceWidget::update(const BriConWidget* b)");

  invalidateBuffer();
  WIDGET_UPDATE;
}

void SliceWidget::update(const DrawWidget* b)
{

}

void SliceWidget::update(const OverlayList* i, OverlayListMsg msg)
{  
  TRACKER("SliceWidget::update(const OverlayList* i, OverlayListMsg msg)");

  switch(msg)
  {
   case OverlayListMsg(Select):      MESSAGE("Select");break;     
   case OverlayListMsg(Visibility):  MESSAGE("Visibility");
   case OverlayListMsg(Transparency):MESSAGE("Transparency");
                                     WIDGET_UPDATE; break;
   case OverlayListMsg(Order):       MESSAGE("Order");
                                     loadOverlays();invalidateBuffer(); 
                                     WIDGET_UPDATE; break;
   case OverlayListMsg(Add):         MESSAGE("Add");
   case OverlayListMsg(Rem):         MESSAGE("Rem");loadOverlays();
   case OverlayListMsg(LookUpTable): MESSAGE("LookUpTable");invalidateBuffer(); 
                                     WIDGET_UPDATE;
                                     break;
  }
}

void SliceWidget::mouseMoveEvent(QMouseEvent *e)
{ 
  QPoint w = convMouseToWorld(QPoint(e->x(),e->y()));

  bool lButton(e->state() & LeftButton);
  bool mButton(e->state() & MidButton);
  bool rButton(e->state() & RightButton);

  int dx = m_startX - w.x();
  int dy = m_startY - w.y();
  
  if(lButton) {
    switch(m_mode)
      {
      case None:
      case Cursing: 
        cursorEvent(w.x(), w.y()); 
  
        break;
      case Pan:
        {
          transEvent(dx, dy);
          QPoint v = convMouseToWorld(QPoint(e->x(),e->y()));
          setStartMove(v.x(),v.y()); 
        }
        break;
      case Zoom:
        zoomEvent(w.x(),w.y());
        break;
      case Masking:
        drawEvent(w.x(),w.y());
        break;
      default:
        break;
      }
  } else if(mButton) {
    transEvent(dx, dy);
    QPoint v = convMouseToWorld(QPoint(e->x(),e->y()));
    setStartMove(v.x(),v.y()); 
  } else if(rButton) {
    zoomEvent(w.x(),w.y());
  }
}

void SliceWidget::mousePressEvent(QMouseEvent *e)
{
  QPoint w = convMouseToWorld(QPoint(e->x(),e->y()));
  
  bool lButton(e->button() == LeftButton);
  bool mButton(e->button() == MidButton);
  bool rButton(e->button() == RightButton);

  setStartMove(w.x(),w.y());

  if(lButton) {
    switch(m_mode)
      {
      case None:
      case Cursing: 
        cursorEvent(w.x(), w.y()); break;
      case Pan: 	  
        setCursor(*m_panCursor);
        break;
      case Zoom: 
        if(e->state() & ControlButton) 
          zoomOut(w.x(), w.y());
        else {
          setCursor(*m_zoomCursor);
          m_zooming = true;
        }
        break;

      case Masking:
        if(layerValidForDrawing())
          {  
            MetaImage mi = m_overlayList->getActiveMetaImage();
            if(isValidMetaImage(mi))
              {
                m_shape = Shape::create(getPairImage(mi)->getVolume(m_volume),
                                           m_orient,
                                           m_slice); 
                drawEvent(w.x(),w.y());
              }
          }
        break;

      default:
        break;
      }
  } else if(mButton) {
  } else if(rButton) {
    if(e->state() & ControlButton) 
      zoomOut(w.x(), w.y());
    else
      m_zooming = true;
  }
}

void SliceWidget::mouseReleaseEvent(QMouseEvent *e)
{

  bool lButton(e->button() == LeftButton);
  bool mButton(e->button() == MidButton);
  bool rButton(e->button() == RightButton);

  if(lButton) {
    switch(m_mode)
      {
      case None:
      case Cursing: 
        break;

      case Pan:  
        setCursor(*m_crossCursor);
        break;

      case Zoom:    
        if(!(e->state() & ControlButton))
          setViewRect(m_zoomRect->left(),m_zoomRect->bottom(),
                      m_zoomRect->right(),m_zoomRect->top());
        setCursor(*m_crossCursor);
        m_zooming = false;
        break;

      case Masking:  
        commitGraphics();
        break;

      default:
        break;
      }
  } else if(mButton) {
  } else if(rButton && !(e->state() & ControlButton)) {
    setViewRect(m_zoomRect->left(),m_zoomRect->bottom(),
		m_zoomRect->right(),m_zoomRect->top());
    m_zooming = false;    
  }
}

void SliceWidget::commitGraphics()
{
  MetaImage mi = m_overlayList->getActiveMetaImage();
  if(isValidMetaImage(mi) && (m_shape.get() != NULL))
    {
      Shape::Handle undoBuffer = m_shape->getBuffer();
      m_undoList.push_back(undoBuffer);
      if(m_undoList.size() > 5)m_undoList.pop_front();
       
      m_shape->commit();
    }
 
  m_cursor->repaint();
}


void SliceWidget::enterEvent( QEvent *e )
{

}

void SliceWidget::leaveEvent( QEvent *e )
{
  TRACKER("SliceWidget::leaveEvent(QEvent)");
  WIDGET_UPDATE;
}

const QPoint  SliceWidget::convMouseToWorld(const QPoint & p)
{
  QPoint world;
  world.setX ((int)((float)(p.x() - m_origX)/(float)m_scaleX) + m_viewRect->left());
  world.setY ((int)((float)((height() - p.y()) - m_origY)/(float)m_scaleY) +m_viewRect->bottom());
  return world;
}



void SliceWidget::transEvent(int dx, int dy)
{
  TRACKER("SliceWidget::transEvent");
  m_viewRect->translate(dx,dy);
  setDataRect();
  WIDGET_UPDATE;
}

void SliceWidget::zoomEvent(int x, int y)
{
  m_zoomRect->setRect(m_startX,m_startY,x,y);
  m_zoomRect->setUnion(m_viewRect); 

  WIDGET_UPDATE;
}

void SliceWidget::briconEvent(int dx, int dy)
{
}

void AxialWidget::setZoom(int factor)
{  
  TRACKER("AxialWidget::setZoom(int)");
  float f = factor/100.0;
  int dx = (int)(inqWidth()/f/2);
  int dy = (int)(inqHeight()/f/2);

  setViewRect(m_cursor->inqX()-dx,m_cursor->inqY()-dy,
              m_cursor->inqX()+dx,m_cursor->inqY()+dy);
  WIDGET_UPDATE;
}

void SagittalWidget::setZoom(int factor)
{  
  TRACKER("SagittalWidget::setZoom(int)");  

  float f = factor/100.0;
  int dx = (int)(inqWidth()/f/2);
  int dy = (int)(inqHeight()/f/2);

  setViewRect(m_cursor->inqY()-dx,m_cursor->inqZ()-dy,
              m_cursor->inqY()+dx,m_cursor->inqZ()+dy);
  WIDGET_UPDATE;
}

void CoronalWidget::setZoom(int factor)
{  
  TRACKER("CoronalWidget::setZoom(int)");  

  float f = factor/100.0;
  int dx = (int)(inqWidth()/f/2);
  int dy = (int)(inqHeight()/f/2);

  setViewRect(m_cursor->inqX()-dx,m_cursor->inqZ()-dy,
              m_cursor->inqX()+dx,m_cursor->inqZ()+dy);

  WIDGET_UPDATE;
}

void SliceWidget::zoomOut(int x,int y)
{
  if(!m_zoomHistory.empty())
    {
      m_viewRect = m_zoomHistory.top();
      m_zoomHistory.pop();
    }
  setDataRect();
  setCoorModifiers();
  WIDGET_UPDATE;
}

void SliceWidget::resetZoom()
{
  TRACKER("SliceWidget::resetZoom()");

  while(!m_zoomHistory.empty()){m_zoomHistory.pop();}

  m_viewRect = Rect::createRect(0,0,inqWidth(),inqHeight());
  m_zoomRect = Rect::createRect(0,0,inqWidth(),inqHeight());
  m_dataRect = Rect::createRect(0,0,inqWidth(),inqHeight());
  
  emitZoomFactor(100);

  initZoom();
  setCoorModifiers();
  WIDGET_UPDATE;
}

void SliceWidget::setViewRect(int startX,int startY,int curX,int curY)
{  
  TRACKER("SliceWidget::setViewRect(int startX,int startY,int curX,int curY)");

  if(startX == curX && startY == curY) return;

  float dataRatio = inqHeight()/(float)inqWidth();

  m_zoomHistory.push(m_viewRect->clone());

  m_viewRect->setRect(startX,startY,curX,curY);
  
  if( m_viewRect->height() >= m_viewRect->width()*dataRatio)
      m_viewRect->setWidth((int)(m_viewRect->height()/dataRatio));
  else if( m_viewRect->height() < m_viewRect->width()*dataRatio)
      m_viewRect->setHeight((int)(m_viewRect->width()*dataRatio));

  setDataRect();
  setCoorModifiers(); 

  WIDGET_UPDATE;
}

void SliceWidget::setDataRect()
{
  TRACKER("SliceWidget::setDataRect");
 
  m_dataRect->setRect(0,0,inqWidth(),inqHeight());
  m_dataRect->setUnion(m_viewRect);
}



void SliceWidget::loadOverlays()
{  

  TRACKER("SliceWidget::loadOverlays");

  m_imageDataList.clear();
  std::transform(m_overlayList->begin(),m_overlayList->end(),std::back_inserter(m_imageDataList),loadDataList); 
}

void SliceWidget::renderBuffer()
{   
  TRACKER("SliceWidget::renderBuffer");

  Image::Handle mi = getPairImage(m_overlayList->getMainMetaImage());
  ImageDisplaySetting::Handle ds = getPairDisplaySetting(m_overlayList->getMainMetaImage());

  m_mainPixels = bufferVolume(mi->getVolume(m_volume), ds);

  for( ImageDataList::iterator i = m_imageDataList.begin();i != m_imageDataList.end();i++)
    {
      MESSAGE("Rendering overlay");
      int vol = std::min(m_volume, int(getImage(i)->getInfo()->inqNumVolumes()-1));
      assert( (vol >= 0) && (vol < getImage(i)->getInfo()->inqNumVolumes()) );

      setPixels(i, bufferVolume(getImage(i)->getVolume(vol), 
				getImageDisplaySetting(i)));
    }

  validateBuffer();
}

void SliceWidget::setMode(int m)
{  
  TRACKER("SliceWidget::setMode");
  m_mode = Mode(m);
  switch(m_mode)
    {
    case None:
    case Cursing:
      setCursor(*m_crossCursor);
      break;
    case Pan:
      setCursor(*m_panCursor);
      break;
    case Zoom:
      setCursor(*m_zoomCursor);
      break;
    case Masking:
      setCursor(*m_penCursor);
      break;
    default:
      break;
    }
  WIDGET_UPDATE;
}

void SliceWidget::drawZoomRectangle()
{  
  TRACKER("SliceWidget::drawZoomRectangle");

  glColor3f(0.0f,1.0f,0.0f);

  glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  
  glBegin(GL_QUADS);
    glVertex2f(m_zoomRect->left(),  m_zoomRect->bottom());    
    glVertex2f(m_zoomRect->left(),  m_zoomRect->top());    
    glVertex2f(m_zoomRect->right(), m_zoomRect->top()); 
    glVertex2f(m_zoomRect->right(), m_zoomRect->bottom());  
  glEnd();   

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void AxialWidget::cursorEvent(int x, int y)
{
  int maxx, maxy;
 
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());

  maxx = info->inqX() - 1;
  maxy = info->inqY() - 1;
  x = std::max(0, x); x = std::min(maxx, x);
  y = std::max(0, y); y = std::min(maxy, y);

  m_cursor->setCursor(x, y, m_slice);
}

int AxialWidget::inqWidth() const 
{ 
  return m_overlayList->inqX(); 
}

int AxialWidget::inqHeight() const 
{
  return m_overlayList->inqY(); 
}

int AxialWidget::inqDepth() const
{
  return m_overlayList->inqZ();
}

float AxialWidget::depthRatio() const
{
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());
  return fabs(info->inqZDim() / info->inqXDim());
}

float AxialWidget::inqRatio() const
{
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());
  return fabs(info->inqYDim() / info->inqXDim());
}

SliceWidget::ColorRGBAHandle AxialWidget::bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const
{
  unsigned int width  = v->inqX();
  unsigned int height = v->inqY();

  SliceWidget::ColorRGBAHandle buffer = 
    SliceWidget::ColorRGBAHandle( new ColorRGBA[width*height] );

  LookUpTable::Handle lut = ds->inqLookUpTable();
  BriCon::Handle bricon = ds->inqBriCon();

  if(lut.get() == NULL)    
    {
      for(unsigned int y = 0; y < height; ++y)
        { 
          unsigned int offset = width * y;

          for( unsigned int x = 0; x < width; ++x) 
            {
              unsigned int xloc = offset + x;
              GLubyte i = clamp(bricon->adjust(v->value(x, y, m_slice)));

              buffer[xloc][0] = i;
              buffer[xloc][1] = i;
              buffer[xloc][2] = i;
              buffer[xloc][3] = i > 0 ? 1 : 0;
            }
        }
    }
  else if(lut->isIndexLut())
    { 
      for(unsigned int y = 0; y < height; ++y)
        {
          unsigned int offset = width * y;

          for(unsigned int x = 0; x < width; ++x)  
            {
              unsigned int xloc = offset + x;

              const ColorTriplet&  col = lut->inqValueIndex(v->value(x, y, m_slice));

              buffer[xloc][0] = *col;
              buffer[xloc][1] = *(col+1);
              buffer[xloc][2] = *(col+2);
              buffer[xloc][3] = (col != lut->inqValue(0)) ? 1 : 0;
            }
        }
    }
  else
    { 
      for(unsigned int y = 0; y < height; ++y)
        {
          unsigned int offset = width * y;
	  
          for(unsigned int x = 0; x < width; ++x)  
            {
              unsigned int xloc = offset + x;
              const ColorTriplet&  col = lut->inqValue(bricon->adjust(v->value(x, y, m_slice)));
         
              buffer[xloc][0] = *col;
              buffer[xloc][1] = *(col+1);
              buffer[xloc][2] = *(col+2);
              buffer[xloc][3] = (col != lut->inqValue(0)) ? 1 : 0;
            }
        }
    }
  return buffer;
}


void SliceWidget::drawCrossHairs()
{
  if(m_prevCursor.get() != NULL)
      drawCrossHairLines(m_prevCursor,m_prevSlice);
 
  drawCrossHairLines(m_cursor,m_slice);
 
  m_prevCursor = m_cursor->clone();
  m_prevSlice = m_slice;
}

void SagittalWidget::drawCrossHairLines(const Cursor::Handle& c,int slice)
{ 
  if(c->inqX() == slice)
    glColor3f(0.0f,1.0f,0.0f);      
  else
    glColor3f(0.0f,0.5f,0.0f);
    
  glEnable(GL_COLOR_LOGIC_OP);
  glLogicOp(GL_XOR);
  
  glBegin(GL_LINES);
  glVertex2f(c->inqY()+.5,0.0f);  
  glVertex2f(c->inqY()+.5,inqHeight());
  glVertex2f(0.0f,c->inqZ()+.5); 
  glVertex2f(inqWidth(),c->inqZ()+.5);
  glEnd();

  glDisable(GL_COLOR_LOGIC_OP);
}

void CoronalWidget::drawCrossHairLines(const Cursor::Handle& c,int slice)
{ 
  if(c->inqY() == slice)   
    glColor3f(0.0f,1.0f,0.0f);
  else
    glColor3f(0.0f,0.5f,0.0f);
  
  glEnable(GL_COLOR_LOGIC_OP);
  glLogicOp(GL_XOR);

  glBegin(GL_LINES);
  glVertex3f(c->inqX()+.5,0.0f,0.0f);  
  glVertex3f(c->inqX()+.5,inqHeight(),0.0f);
  glVertex3f(0.0f,c->inqZ()+.5,0.0f); 
  glVertex3f(inqWidth(),c->inqZ()+.5,0.0f);
  glEnd();

  glDisable(GL_COLOR_LOGIC_OP);
}

void AxialWidget::drawCrossHairLines(const Cursor::Handle& c,int slice)
{  
  if(c->inqZ() == slice)
    glColor3f(0.0f,1.0f,0.0f); 
  else 
    glColor3f(0.0f,0.5f,0.0f);
  
  glEnable(GL_COLOR_LOGIC_OP);
  glLogicOp(GL_XOR);

  glBegin(GL_LINES);
  glVertex3f(c->inqX()+.5,0.0f,0.0f);  
  glVertex3f(c->inqX()+.5,inqHeight(),0.0f);
  glVertex3f(0.0f,c->inqY()+.5,0.0f); 
  glVertex3f(inqWidth(),c->inqY()+.5,0.0f);
  glEnd();
 
  glDisable(GL_COLOR_LOGIC_OP);
}
void SliceWidget::drawEvent(int x, int y)
{  
  int   size  = m_drawWidget->inqPixelSize();
  int   value = m_drawWidget->inqPixelValue();

  if(m_shape.get() != NULL && layerValidForDrawing()){m_shape->addVertex(x,y,size,(float)value);}

  WIDGET_UPDATE;
}

bool SliceWidget::layerValidForDrawing()
{  
  bool result(false);

  MetaImage mi = m_overlayList->getActiveMetaImage();
  
  if(isValidMetaImage(mi))
    { 
      if(!getPairImage(mi)->getInfo()->inqReadOnly())
        {result = true;}
      else
        {emit message("Warning: Cannot draw on locked image", 2000);}
    }
  else
    {
      emit message("Warning: No valid overlay selected",2000);
    }

  return result;
}



void CoronalWidget::cursorEvent(int x, int y)
{
  int maxx, maxy;
  
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());

  maxx = info->inqX() - 1;
  maxy = info->inqZ() - 1;
  x = std::max(0, x); x = std::min(maxx, x);
  y = std::max(0, y); y = std::min(maxy, y);

  m_cursor->setCursor(x, m_slice, y); 
}

int CoronalWidget::inqWidth() const 
{ 
  return m_overlayList->inqX();
}

int CoronalWidget::inqHeight() const 
{
  return m_overlayList->inqZ();
}

int CoronalWidget::inqDepth() const
{
  return m_overlayList->inqY();
}

float CoronalWidget::depthRatio() const
{
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());
  return fabs(info->inqYDim() / info->inqXDim());
}

float CoronalWidget::inqRatio() const
{
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());
  return fabs(info->inqZDim() / info->inqXDim());
}

SliceWidget::ColorRGBAHandle CoronalWidget::bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const
{
  unsigned int width  = v->inqX();
  unsigned int height = v->inqZ();
    
  SliceWidget::ColorRGBAHandle buffer = 
    SliceWidget::ColorRGBAHandle( new ColorRGBA[width*height] );

  LookUpTable::Handle lut = ds->inqLookUpTable();
  BriCon::Handle bricon = ds->inqBriCon();

  if(lut.get() == NULL)  
    {
      for(unsigned int z = 0; z < height; ++z)
        {
          unsigned int offset = width * z;
          
          for(unsigned int x = 0; x < width; ++x)
            {
              unsigned int xloc = offset + x;
              GLubyte i = clamp(bricon->adjust(v->value(x, m_slice, z)));
        
              buffer[xloc][0] = i;
              buffer[xloc][1] = i;
              buffer[xloc][2] = i;
              buffer[xloc][3] = i > 0 ? 1 : 0;
            }
        }
    }
  else if(lut->isIndexLut())
    {
      for(unsigned int z = 0; z < height; ++z)
        {
          unsigned int offset = width * z;
          
          for(unsigned int x = 0; x < width; ++x)
            {
              unsigned int xloc = offset + x;
              const ColorTriplet& col = lut->inqValueIndex(v->value(x, m_slice, z));

              buffer[xloc][0] = *col;
              buffer[xloc][1] = *(col+1);
              buffer[xloc][2] = *(col+2);
              buffer[xloc][3] = (col != lut->inqValue(0)) ? 1 : 0;
            }
        }
    }
  else
    {
      for(unsigned int z = 0; z < height; ++z)
        {
          unsigned int offset = width * z;
          
          for(unsigned int x = 0; x < width; ++x)
            {
              unsigned int xloc = offset + x;
              const ColorTriplet& col = lut->inqValue(bricon->adjust(v->value(x, m_slice, z)));

              buffer[xloc][0] = *col;
              buffer[xloc][1] = *(col+1);
              buffer[xloc][2] = *(col+2);
              buffer[xloc][3] = (col != lut->inqValue(0)) ? 1 : 0;
            }
        }
    }
  return buffer;
}

void SagittalWidget::cursorEvent(int x, int y)
{
  int maxx, maxy;
  
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());

  maxx = info->inqY() - 1;
  maxy = info->inqZ() - 1;
  x = std::max(0, x); x = std::min(maxx, x);
  y = std::max(0, y); y = std::min(maxy, y);

  m_cursor->setCursor(m_slice, x, y);
}

int SagittalWidget::inqWidth() const 
{
  return m_overlayList->inqY(); 
}

int SagittalWidget::inqHeight() const 
{ 
  return m_overlayList->inqZ(); 
}

int SagittalWidget::inqDepth() const
{
  return m_overlayList->inqX();
}

float SagittalWidget::depthRatio() const
{
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());
  return fabs(info->inqXDim() / info->inqYDim());
}

void SliceWidget::drawTest()
{

  glColor3f(1.0f,1.0f,0.0f); 

  glBegin(GL_LINES);
  glVertex3f(0.1f,0.1f,0.0f);  
  glVertex3f(0.9f,0.9f,0.0f);
  glEnd();

}

float SagittalWidget::inqRatio() const
{
  ImageInfo::Handle info(m_overlayList->getMainImage()->getInfo());
  return fabs(info->inqZDim() / info->inqYDim());
}

SliceWidget::ColorRGBAHandle SagittalWidget::bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const
{
  unsigned int width  = v->inqY();
  unsigned int height = v->inqZ();
    
  SliceWidget::ColorRGBAHandle buffer = 
    SliceWidget::ColorRGBAHandle( new ColorRGBA[width*height] );

  LookUpTable::Handle lut = ds->inqLookUpTable();
  BriCon::Handle bricon = ds->inqBriCon();

  if(lut.get()== NULL)
    {  
      for(unsigned int z = 0; z < height; ++z)
        {
          unsigned int offset = width * z;
      
          for(unsigned int y = 0; y < width; ++y)
            {
              unsigned int xloc = offset + y;
	      GLubyte i = clamp(bricon->adjust(v->value(m_slice, y, z)));
        
              buffer[xloc][0] = i;
              buffer[xloc][1] = i;
              buffer[xloc][2] = i; 
	      buffer[xloc][3] = i > 0 ? 1 : 0;
            }
        }
    } 
  else if(lut->isIndexLut())
    {
      for(unsigned int z = 0; z < height; ++z)
        {
          unsigned int offset = width * z;
      
          for(unsigned int y = 0; y < width; ++y)
            {
              unsigned int xloc = offset + y;
              const ColorTriplet& col = lut->inqValueIndex(v->value(m_slice, y, z));
              
              buffer[xloc][0] = *col;
              buffer[xloc][1] = *(col+1);
              buffer[xloc][2] = *(col+2);
              buffer[xloc][3] = (col != lut->inqValue(0)) ? 1 : 0;
            }
        }
    }
  else
    {
      for(unsigned int z = 0; z < height; ++z)
        {
          unsigned int offset = width * z;
      
          for(unsigned int y = 0; y < width; ++y)
            {
              unsigned int xloc = offset + y;
	      const ColorTriplet& col = lut->inqValue(bricon->adjust(v->value(m_slice, y, z)));
              
	      buffer[xloc][0] = *col;
              buffer[xloc][1] = *(col+1);
              buffer[xloc][2] = *(col+2);
	      buffer[xloc][3] = (col != lut->inqValue(0)) ? 1 : 0;
            }
        }
    }
  return buffer;
}


void SliceWidget::showSlice()
{
  m_timer->stop();
  resize(width(),height()+1);
}
