/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  Part of FSLView 1.0
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    LICENCE

FSLView v1.0 sources are released under the terms of the GPL
(http://www.gnu.org/copyleft/gpl.html). FSLView binaries use Qt
(http://www.trolltech.com) which is released under GPL for X11 and
under the "Qt non-commercial licence" for Windows.  The FSLView
binaries are released under the GPL, with the added exception that we
also give permission to link this program with the Qt non-commercial
edition, and distribute the resulting executable, without including
the source code for the Qt non-commercial edition in the source
distribution.

FSLView is distributed "AS IS" under this Licence in the hope that it
will be useful, but in order that the University as a charitable
foundation protects its assets for the benefit of its educational and
research purposes, the University makes clear that no condition is
made or to be implied, nor is any warranty given or to be implied, as
to the accuracy of the Software, or that it will be suitable for any
particular purpose or for use under any specific conditions.
Furthermore, the University disclaims all responsibility for the use
which is made of the Software. It further disclaims any liability for
the outcomes arising from using the Software.
    
 */


#if !defined (SLICEWIDGET_H)
#define SLICEWIDGET_H

#include <qgl.h>
#include <qcursor.h>

#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>
#include <vector>
#include <stack>

#include "storage/image.h"
#include "cursor.h"
#include "overlaylist.h"
#include "briconwidget.h"
#include "drawwidget.h"
#include "lookuptable.h"

#include "rect.h"
#include "shape.h"

#ifdef WIN32
	#include <gl/glext.h>
#endif

class QTimer;

class SliceWidget : public QGLWidget, public OverlayListObserver,
                                      public CursorObserver,
                                      public BriConWidgetObserver,
                                      public DrawWidgetObserver
{
  Q_OBJECT
public:
  typedef boost::shared_ptr< SliceWidget > Handle;
  typedef boost::shared_array< ColorTriplet > ColorTripletHandle;
  typedef boost::shared_array< ColorRGBA > ColorRGBAHandle;
  typedef std::pair<MetaImage,ColorRGBAHandle> OverlayPair;
  typedef std::list<OverlayPair> ImageDataList;

  typedef enum {Sagittal, Axial, Coronal} Orientation;
  typedef enum {None, Cursing, Pan, Zoom, TimeSeries, Masking} Mode;

  SliceWidget(QWidget* parent, const char* name, 
              Orientation orient, Cursor::Handle& c, OverlayList::Handle l, BriConWidget* b,DrawWidget* d,std::list<Shape::Handle>& u);
  virtual ~SliceWidget();
  
  virtual void initializeGL();
  virtual void resizeGL(int w, int h);
  virtual void paintGL();

  void renderBuffer();

  virtual void update(const Cursor::Handle& c);  
  virtual void update(const BriConWidget* b);
  virtual void update(const OverlayList* l, OverlayListMsg msg);
  virtual void update(const DrawWidget* d);
 
signals:
  void imageCursorChanged(int, int, int);
  void volumeChanged(int);
  void emitZoomFactor(int);
  void message(const QString&, int );


public slots:
  virtual void setImageCursor(int, int, int, int) = 0;
  virtual void setZoom(int) = 0;
  void zoomOut(int,int);  
  void setViewRect(int,int,int,int);
  void resetZoom();
  void setMode(int);
  void setSlice(int,int);
  void setSliceIsFixed(bool);
  virtual int inqWidth() const = 0;
  virtual int inqHeight() const = 0;
  void enableUpdates(bool);
  void crossHairMode(bool);

protected:
  virtual void mousePressEvent(QMouseEvent*);
  virtual void mouseMoveEvent(QMouseEvent*);
  virtual void mouseReleaseEvent(QMouseEvent*);
  virtual void enterEvent(QEvent*);
  virtual void leaveEvent(QEvent*);
  inline GLubyte briConAdjust(float) const;
  inline GLubyte clamp(float) const;  


  bool               m_sliceIsFixed;
  bool               m_updatesEnabled;
  bool               m_crossHairsOn;
  bool               m_imagesEnabled;
  int                m_slice;  
  bool               m_noSliceSet;
  int                m_volume;  
  int                m_startX, m_startY;
  Cursor::Handle     m_cursor;
  Cursor::Handle     m_prevCursor;
  int                m_prevSlice; 
  float              m_zoom; 
  int                m_origX,m_origY;
  float              m_scaleX,m_scaleY;
  Rect::Handle       m_viewRect, m_dataRect, m_zoomRect;
  ColorRGBAHandle    m_mainPixels;
  ImageDataList      m_imageDataList;
  OverlayList::Handle  m_overlayList;
  BriConWidget*        m_briconWidget;
  DrawWidget*          m_drawWidget;
  
  #ifdef WIN32
  PFNGLBLENDCOLORPROC  m_glBlendColorEXT; 
  #endif

  

  std::stack<Rect::Handle> m_zoomHistory;

private slots:

 void showSlice();

private:
  virtual int inqDepth() const = 0;
  virtual float depthRatio() const = 0;
  virtual float inqRatio() const = 0;
  void drawZoomRectangle();
  void drawTest();
  void paintImages();
  void paintGraphics();  
  void drawCrossHairs();
  virtual void drawCrossHairLines(const Cursor::Handle& c,int slice) = 0;
  virtual ColorRGBAHandle bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const = 0;
  virtual void cursorEvent(int x, int y) = 0; 
  virtual void drawEvent(int x, int y);
  virtual void commitGraphics();
  virtual void loadOverlays();
 
  void undoGraphics();
  void briconEvent(int x, int y);
  void zoomEvent(int x, int y);
  void transEvent(int dx,int dy);
  void setStartMove(int x, int y) { m_startX = x; m_startY = y; }
  void setCoorModifiers();
  void setDataRect();
  void initZoom();
  bool layerValidForDrawing();
  const QPoint  convMouseToWorld(const QPoint &);
  const QPoint  convWorldToViewport(const QPoint &);
  inline ColorRGBAHandle getPixels(ImageDataList::iterator i){return i->second;}
  inline ImageDisplaySetting::Handle getImageDisplaySetting(ImageDataList::iterator i){return getPairDisplaySetting(i->first);}
  inline LookUpTable::Handle getLookUpTable(ImageDataList::iterator i){return getPairDisplaySetting(i->first)->inqLookUpTable();}
  inline Image::Handle getImage(ImageDataList::iterator i){return getPairImage(i->first);}
  inline void setPixels(ImageDataList::iterator i, ColorRGBAHandle pixels){i->second = pixels;} 
  
  inline void invalidateBuffer() const { m_bufferValid = false; }
  inline bool bufferValid() const { return m_bufferValid; }
  inline void validateBuffer() const { m_bufferValid = true; }

  static QCursor    *m_crossCursor;
  static QCursor    *m_zoomCursor;
  static QCursor    *m_panCursor;
  static QCursor    *m_penCursor;  
  std::list<Shape::Handle>     &m_undoList;
  Orientation        m_orient;  
  int                m_mode;
  bool               m_zooming;
  bool               m_forceRender;
  mutable bool       m_bufferValid;
  Shape::Handle      m_shape;
  QTimer            *m_timer;
};

class SagittalWidget : public SliceWidget
{
public:
  SagittalWidget(QWidget* parent, const char* name, Cursor::Handle c,OverlayList::Handle l, BriConWidget* b, DrawWidget* d,std::list<Shape::Handle>& u):
    SliceWidget(parent, name, SliceWidget::Sagittal, c,l,b,d,u) {};

  virtual void setImageCursor(int, int, int, int);
  virtual void setZoom(int);
private:
  virtual int inqWidth() const;
  virtual int inqHeight() const;
  virtual int inqDepth() const;
  virtual float depthRatio() const;
  virtual float inqRatio() const;
  virtual void drawCrossHairLines(const Cursor::Handle& c, int slice);
  virtual ColorRGBAHandle bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const;
  virtual void cursorEvent(int x, int y);  
};

class CoronalWidget : public SliceWidget
{
public:
  CoronalWidget(QWidget* parent, const char* name, Cursor::Handle& c, OverlayList::Handle l, BriConWidget* b,DrawWidget* d,std::list<Shape::Handle>& u):
    SliceWidget(parent, name, SliceWidget::Coronal, c, l, b,d,u) {};
  virtual void setImageCursor(int, int, int, int);  
  virtual void setZoom(int);
private:
  virtual int inqWidth() const;
  virtual int inqHeight() const;
  virtual int inqDepth() const;
  virtual float depthRatio() const;
  virtual float inqRatio() const;
  virtual void drawCrossHairLines(const Cursor::Handle& c,int slice);
  virtual ColorRGBAHandle bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const;
  virtual void cursorEvent(int x, int y);  
};

class AxialWidget : public SliceWidget
{
public:
  AxialWidget(QWidget* parent, const char* name, Cursor::Handle& c, OverlayList::Handle l, BriConWidget* b, DrawWidget* d,std::list<Shape::Handle>& u):
    SliceWidget(parent, name, SliceWidget::Axial, c, l, b, d, u) {};
  virtual void setImageCursor(int, int, int, int);  
  virtual void setZoom(int);
private:
  virtual int inqWidth() const;
  virtual int inqHeight() const;
  virtual int inqDepth() const;
  virtual float depthRatio() const;
  virtual float inqRatio() const;
  virtual void drawCrossHairLines(const Cursor::Handle& c, int slice);
  virtual ColorRGBAHandle bufferVolume(Volume::Handle v, ImageDisplaySetting::Handle ds) const;
  virtual void cursorEvent(int x, int y);  

};

typedef std::vector<SliceWidget::Handle> SliceList;
typedef boost::shared_ptr<SliceList> SliceListHandle;


class SetImageCursor {
public:
  SetImageCursor(const Cursor::Handle& c) : m_c(c) {}
  void operator()(SliceList::value_type& e) const
  { 
    e->setImageCursor(m_c->inqX(), m_c->inqY(), m_c->inqZ(), m_c->inqV()); 
  }
private:
  const Cursor::Handle m_c;
};

#endif
