/* {{{ Copyright etc. */

/*  avwmaths.c - image maths

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2003 University of Oxford  */

/*  CCOPYRIGHT */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "libss/libss.h"
#include "libss/libavw.h"

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("\nUsage: avwmaths <first_input> [operations and inputs] <output>\n\n");

  printf("\nBinary operations:\n");
  printf("  some inputs can be either an image or a number\n");
  printf("  \"current image\" may be n_volumes>1 and the second n_volumes=1\n");
  printf("-add  : add following input to current image\n");
  printf("-sub  : subtract following input from current image\n");
  printf("-mul  : multiply current image by following input\n");
  printf("-div  : divide current image by following input\n");
  printf("-mas  : use (following image>0) to mask current image\n");
  printf("-thr  : use following number to threshold current image (zero anything below the number)\n");
  printf("-uthr : use following number to upper-threshold current image (zero anything above the number)\n");
  printf("-max  : take maximum of following input and current image\n");
  printf("-min  : take minimum of following input and current image\n");

  printf("\nUnary operations:\n");
  printf("-exp  : exponential\n");
  printf("-log  : natural logarithm\n");
  printf("-sqr  : square\n");
  printf("-sqrt : square root\n");
  printf("-abs  : absolute value\n");
  printf("-bin  : use (current image>0) to binarise\n");
  printf("-dil  : use (current image>0) to dilate using 3x3x3 neighourhood (new value is average of non-zero neighbours)\n");
  printf("-dils : use (current image>0) to dilate using 3x3x3 neighourhood (simple - new value is last non-zero neighbour\n");
  printf("-dil2 : use (current image>0) to dilate using 3x3 (2D) neighourhood (new value is average of non-zero neighbours)\n");
  printf("-ero  : use (current image>0) to erode using 3x3x3 neighourhood\n");
  printf("-ero2 : use (current image>0) to erode using 3x3 (2D) neighourhood\n");
  printf("-edge : edge strength\n");
  printf("-nan  : replace NaNs (improper numbers) with 0\n");
  printf("-nanm : make NaN (improper number) mask with 1 for NaN voxels, 0 otherwise\n");
  printf("-roi <xmin> <xsize> <ymin> <ysize> <zmin> <zsize> <tmin> <tsize> : zero outside roi\n");

  printf("\nDimensionality reduction operations:\n");
  printf("  The \"T\" can be replaced by X, Y or Z to collapse across a different dimension\n");
  printf("-Tmean : mean across time\n");
  printf("-Tstd  : standard deviation across time\n");
  printf("-Tmax  : max across time\n");
  printf("-Tmaxn : time index of max across time\n");
  printf("-Tmin  : min across time\n");

  printf("\ne.g. avwmaths grotin -add grotin2 grotout\n");
  printf("     avwmaths grotin -add 2.5 grotout\n");
  printf("     avwmaths grotin -add 2.5 -mul grotin2 grotout\n\n");

  exit(1);
}

/* }}} */
/* {{{ isavw */

int isavw (char *filename, double *tmpd)
{
  char filename2[10000];
  FILE *fp;

  strcpy(filename2,filename);
  if ( (strcmp(".hdr",filename2+strlen(filename2)-4)==0) ||
       (strcmp(".img",filename2+strlen(filename2)-4)==0) )
    filename2[strlen(filename2)-4]=0;

  strcat(filename2,".hdr");

  if((fp=fopen(filename2,"rb"))==NULL)
    {
      *tmpd=atof(filename);
      return 0;
    }
  else
    {
      fclose(fp);
      return 1;
    }
}

/* }}} */
/* {{{ main */

int main(argc, argv)
  int   argc;
  char  *argv [];
{
  /* {{{ vars */

image_struct im, im2;
FDT  *in[1000];
int  i, j, x, y, z, t, xysize, xyzsize, size;
double tmpd;

/* }}} */

  if (argc<3)
    usage();

  avw_read(argv[1],&im);
  in[0]=im.i;
  xysize=im.x*im.y;
  xyzsize=xysize*im.z;
  size=xyzsize*im.t;

  for (i = 2; i < argc-1; i++) {
    if (isupper((int)argv[i][1]))
      /* {{{ group operations (dimensionality reduction) */

#define CI int nmax=0; double sum=0,sumsquares=0,min=im.dtmax,max=im.dtmin

#define CU { sum        += in[0][t*im.z*im.y*im.x+z*im.y*im.x+y*im.x+x];                                                    \
             sumsquares += in[0][t*im.z*im.y*im.x+z*im.y*im.x+y*im.x+x]*in[0][t*im.z*im.y*im.x+z*im.y*im.x+y*im.x+x];       \
             if (in[0][t*im.z*im.y*im.x+z*im.y*im.x+y*im.x+x]>max) {                                                        \
               max=in[0][t*im.z*im.y*im.x+z*im.y*im.x+y*im.x+x];                                                            \
                    if (argv[i][1] == 'T')  nmax=t;                                                                         \
               else if (argv[i][1] == 'Z')  nmax=z;                                                                         \
               else if (argv[i][1] == 'Y')  nmax=y;                                                                         \
               else if (argv[i][1] == 'X')  nmax=x; }                                                                       \
             min        =  MIN(min,in[0][t*im.z*im.y*im.x+z*im.y*im.x+y*im.x+x]); }

#define CO      if (!strncmp(argv[i]+2, "mean", 4)) tmpim.i[t*tmpim.z*tmpim.y*tmpim.x+z*tmpim.y*tmpim.x+y*tmpim.x+x]=sum/n;  \
           else if (!strncmp(argv[i]+2, "std",  3)) tmpim.i[t*tmpim.z*tmpim.y*tmpim.x+z*tmpim.y*tmpim.x+y*tmpim.x+x]=        \
                                                                   sqrt((sumsquares/MAX(1,n-1))-((sum*sum)/(n*MAX(1,n-1)))); \
	   else if (!strncmp(argv[i]+2, "maxn", 4)) tmpim.i[t*tmpim.z*tmpim.y*tmpim.x+z*tmpim.y*tmpim.x+y*tmpim.x+x]=nmax;   \
	   else if (!strncmp(argv[i]+2, "max",  3)) tmpim.i[t*tmpim.z*tmpim.y*tmpim.x+z*tmpim.y*tmpim.x+y*tmpim.x+x]=max;    \
	   else if (!strncmp(argv[i]+2, "min",  3)) tmpim.i[t*tmpim.z*tmpim.y*tmpim.x+z*tmpim.y*tmpim.x+y*tmpim.x+x]=min;

{
  image_struct tmpim=im;
  tmpim.i=malloc(size*sizeof(FDT));

  if (argv[i][1] == 'T')
    {
      tmpim.t=1;
      for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
	{ CI, n=im.t; for(t=0;t<im.t;t++) CU; t=0; CO; }
    }
  else if (argv[i][1] == 'Z')
    {
      tmpim.z=1;
      for(t=0;t<im.t;t++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
	{ CI, n=im.z; for(z=0;z<im.z;z++) CU; z=0; CO; }
    }
  else if (argv[i][1] == 'Y')
    {
      tmpim.y=1;
      for(z=0;z<im.z;z++) for(t=0;t<im.t;t++) for(x=0;x<im.x;x++)
	{ CI, n=im.y; for(y=0;y<im.y;y++) CU; y=0; CO; }
    }
  else if (argv[i][1] == 'X')
    {
      tmpim.x=1;
      for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(t=0;t<im.t;t++)
	{ CI, n=im.x; for(x=0;x<im.x;x++) CU; x=0; CO; }
    }

  free(in[0]);
  im=tmpim;
  in[0]=im.i;
  xysize=im.x*im.y;
  xyzsize=xysize*im.z;
  size=xyzsize*im.t;
}

/* }}} */
    else if (!strncmp(argv[i], "-mas", 4))
      /* {{{ mask */

{
  i++;
  avw_read(argv[i],&im2);
  in[1]=im2.i;
  for(j=0;j<size;j++)
    if (in[1][j%(im2.x*im2.y*im2.z*im2.t)]<=0)
      in[0][j]=0;
  free(in[1]);
}

/* }}} */
    else if (!strncmp(argv[i], "-add", 4))
      /* {{{ add */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      in[0][j]+=tmpd;
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]+=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
      free(in[1]);
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-sub", 4))
      /* {{{ subtract */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      in[0][j]-=tmpd;
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	  in[0][j]-=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
      free(in[1]);
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-mul", 4))
      /* {{{ multiply */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      in[0][j]*=tmpd;
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]*=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
      free(in[1]);
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-div", 4))
      /* {{{ divide */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    {
      if (tmpd!=0)
	for(j=0;j<size;j++)
	  in[0][j]/=tmpd;
      else
	for(j=0;j<size;j++)
	  in[0][j]=0;
    }
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	if (in[1][j%(im2.x*im2.y*im2.z*im2.t)]!=0)
	  in[0][j]/=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
	else
	  in[0][j]=0;
      free(in[1]);
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-thr", 4))
      /* {{{ threshold */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      if (in[0][j]<tmpd)
	in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-uthr", 5))
      /* {{{ upper-threshold */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      if (in[0][j]>tmpd)
	in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-exp", 4))
      /* {{{ exp */

{
  for(j=0;j<size;j++)
    in[0][j]=exp(in[0][j]);
}

/* }}} */
    else if (!strncmp(argv[i], "-log", 4))
      /* {{{ log */

{
  for(j=0;j<size;j++)
    if (in[0][j]>0)
      in[0][j]=log(in[0][j]);
    else
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-sqrt", 5))
      /* {{{ square root */

{
  for(j=0;j<size;j++)
    if (in[0][j]>0)
      in[0][j]=sqrt(in[0][j]);
    else
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-sqr", 4))
      /* {{{ square */

{
  for(j=0;j<size;j++)
    in[0][j]*=in[0][j];
}

/* }}} */
    else if (!strncmp(argv[i], "-abs", 4))
      /* {{{ abs */

{
  for(j=0;j<size;j++)
    if (in[0][j]<0)
      in[0][j]=-in[0][j];
}

/* }}} */
    else if (!strncmp(argv[i], "-bin", 4))
      /* {{{ binarise */

{
  for(j=0;j<size;j++)
    if (in[0][j]>0)
      in[0][j]=1;
    else
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-max", 4))
      /* {{{ max */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      in[0][j]=MAX(tmpd,in[0][j]);
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]=MAX(in[1][j%(im2.x*im2.y*im2.z*im2.t)],in[0][j]);
      free(in[1]);
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-min", 4))
      /* {{{ min */

{
  i++;
  if ( ! isavw(argv[i],&tmpd) )
    for(j=0;j<size;j++)
      in[0][j]=MIN(tmpd,in[0][j]);
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]=MIN(in[1][j%(im2.x*im2.y*im2.z*im2.t)],in[0][j]);
      free(in[1]);
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-dils", 5))
      /* {{{ dilate simple */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]==0)
      {
	int xx, yy, zz;
	for(zz=MAX(z-1,0);zz<=MIN(z+1,im.z-1);zz++)
	  for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	    for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	      if (in[0][t*xyzsize+zz*xysize+yy*im.x+xx]!=0)
		tmpim[t*xyzsize+z*xysize+y*im.x+x]=in[0][t*xyzsize+zz*xysize+yy*im.x+xx];
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-dil2", 5))
      /* {{{ dilate 2D */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]==0)
      {
	int xx, yy, count=0;
	double total=0;
	for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	  for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	    if (in[0][t*xyzsize+z*xysize+yy*im.x+xx]!=0)
	      {
		total+=in[0][t*xyzsize+z*xysize+yy*im.x+xx];
		count++;
	      }
	if (count>0)
	  tmpim[t*xyzsize+z*xysize+y*im.x+x]=total/count;
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-dil", 4))
      /* {{{ dilate */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]==0)
      {
	int xx, yy, zz, count=0;
	double total=0;
	for(zz=MAX(z-1,0);zz<=MIN(z+1,im.z-1);zz++)
	  for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	    for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	      if (in[0][t*xyzsize+zz*xysize+yy*im.x+xx]!=0)
		{
		  total+=in[0][t*xyzsize+zz*xysize+yy*im.x+xx];
		  count++;
		}
	if (count>0)
	  tmpim[t*xyzsize+z*xysize+y*im.x+x]=total/count;
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-ero2", 5))
      /* {{{ erode */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]!=0)
      {
	int xx, yy;
	for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	  for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	    if (in[0][t*xyzsize+z*xysize+yy*im.x+xx]==0)
	      tmpim[t*xyzsize+z*xysize+y*im.x+x]=0;
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-ero", 4))
      /* {{{ erode */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]!=0)
      {
	int xx, yy, zz;
	for(zz=MAX(z-1,0);zz<=MIN(z+1,im.z-1);zz++)
	  for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	    for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	      if (in[0][t*xyzsize+zz*xysize+yy*im.x+xx]==0)
		tmpim[t*xyzsize+z*xysize+y*im.x+x]=0;
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-edge", 5))
      /* {{{ edge */

{
  float tmpf=2 * sqrt(1/(im.xv*im.xv) + 1/(im.yv*im.yv) + 1/(im.zv*im.zv));
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=1;z<im.z-1;z++) for(y=1;y<im.y-1;y++) for(x=1;x<im.x-1;x++)
    tmpim[t*xyzsize+z*xysize+y*im.x+x] = sqrt (

	( ((double)in[0][t*xyzsize+(z+1)*xysize+y*im.x+x]-(double)in[0][t*xyzsize+(z-1)*xysize+y*im.x+x]) *
	  ((double)in[0][t*xyzsize+(z+1)*xysize+y*im.x+x]-(double)in[0][t*xyzsize+(z-1)*xysize+y*im.x+x]) ) 
	/ (im.zv*im.zv)

	+

	( ((double)in[0][t*xyzsize+z*xysize+(y+1)*im.x+x]-(double)in[0][t*xyzsize+z*xysize+(y-1)*im.x+x]) *
	  ((double)in[0][t*xyzsize+z*xysize+(y+1)*im.x+x]-(double)in[0][t*xyzsize+z*xysize+(y-1)*im.x+x]) )
	/ (im.yv*im.yv)

	+

	( ((double)in[0][t*xyzsize+z*xysize+y*im.x+(x+1)]-(double)in[0][t*xyzsize+z*xysize+y*im.x+(x-1)]) *
	  ((double)in[0][t*xyzsize+z*xysize+y*im.x+(x+1)]-(double)in[0][t*xyzsize+z*xysize+y*im.x+(x-1)]) )
	/ (im.xv*im.xv)

	) / tmpf;

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-nanm", 5))
      /* {{{ NaN mask */

{
  for(j=0;j<size;j++)
    if ( finite((double)in[0][j]) )
      in[0][j]=0;
    else
      in[0][j]=1;
}

/* }}} */
    else if (!strncmp(argv[i], "-nan", 4))
      /* {{{ delete NaNs */

{
  for(j=0;j<size;j++)
    if ( ! finite((double)in[0][j]) )
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-roi", 4))
      /* {{{ ROI masking */

{
  int xmin,xsize,ymin,ysize,zmin,zsize,tmin,tsize;

  i++; xmin  = atoi(argv[i]);
  i++; xsize = atoi(argv[i]);
  i++; ymin  = atoi(argv[i]);
  i++; ysize = atoi(argv[i]);
  i++; zmin  = atoi(argv[i]);
  i++; zsize = atoi(argv[i]);
  i++; tmin  = atoi(argv[i]);
  i++; tsize = atoi(argv[i]);

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if ( (t<tmin) || (t>=tmin+tsize) ||
	 (z<zmin) || (z>=zmin+zsize) ||
	 (y<ymin) || (y>=ymin+ysize) ||
	 (x<xmin) || (x>=xmin+xsize) )
      in[0][t*xyzsize+z*xysize+y*im.x+x]=0;
}

/* }}} */
    else
      usage();
  }

  im.i=in[0];
  find_thresholds(&im,0.1);
  avw_write(argv[i],im);
  return(0);
}

/* }}} */
