/*  convertwarp.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 2001 University of Oxford  */

/*  CCOPYRIGHT  */

#include "utils/options.h"
#include "warpfns.h"

#define _GNU_SOURCE 1
#define POSIX_SOURCE 1

using namespace Utilities;

////////////////////////////////////////////////////////////////////////////

// COMMAND LINE OPTIONS

string title="convertwarp (Version 1.0)\nCopyright(c) 2001, University of Oxford (Mark Jenkinson)";
string examples="convertwarp -m affine_matrix_file -r refvol -o output_warp\nconvertwarp --premat=mat1 --warp1=vol1 --warp2=vol2 --postmat=mat2 -o output_warp\nconvertwarp -s shiftmapvol -o output_warp";

Option<bool> verbose(string("-v,--verbose"), false, 
		     string("switch on diagnostic messages"), 
		     false, no_argument);
Option<bool> help(string("-h,--help"), false,
		  string("display this message"),
		  false, no_argument);
Option<string> prematname(string("-m,--premat"), string(""),
			  string("filename of pre-affine transform (not MEDx)"),
			  false, requires_argument);
Option<string> midmatname(string("--midmat"), string(""),
			  string("filename of mid-warp-affine transform (not MEDx)"),
			  false, requires_argument);
Option<string> postmatname(string("--postmat"), string(""),
			  string("filename of post-affine transform (not MEDx)"),
			  false, requires_argument);
Option<string> shiftmapname(string("-s,--shiftmap"), string(""),
		       string("filename for shiftmap"),
		       false, requires_argument);
Option<string> warp1name(string("-w,--warp1"), string(""),
		       string("filename for initial warp (follows pre-affine)"),
		       false, requires_argument);
Option<string> warp2name(string("--warp2"), string(""),
		       string("filename for secondary warp (after initial warp, before post-affine)"),
		       false, requires_argument);
Option<string> refname(string("-r,--ref"), string(""),
		       string("filename for reference image"),
		       false, requires_argument);
Option<string> outname(string("-o,--out"), string(""),
		       string("filename for output (warp) image"),
		       true, requires_argument);
Option<string> shiftdir(string("-d,--shiftdir"), string("y"),
		       string("direction to apply shiftmap {x,y,z,x-,y-,z-}"),
		       false, requires_argument);

////////////////////////////////////////////////////////////////////////////

void update_warp(volume4D<float>& totalwarp,  const volume4D<float>& newwarp,
		 bool& warpset)
{
  if (warpset) { 
    volume4D<float> tmpwarp = totalwarp;  // previous warp (prewarp)
    concat_warps(tmpwarp,newwarp,totalwarp);
  } else {
    totalwarp = newwarp; 
    warpset = true;
  }
}
		 


int convert_warp()
{
  volume<float> refvol;

  if (refname.set()) {
    read_volume(refvol,refname.value());
  } else if (warp2name.set()) {
    read_volume(refvol,warp2name.value());  // only reads first volume
  } else if (warp1name.set()) {
    read_volume(refvol,warp1name.value());  // only reads first volume
  } else if (shiftmapname.set()) {
    read_volume(refvol,shiftmapname.value());  // only reads first volume
  } else {
    cerr << "Cannot determine size of output warp!" << endl;
    cerr << "Must specify a reference volume (using -r)" << endl;
    exit(EXIT_FAILURE);
  }

  volume4D<float> nextwarp, finalwarp;
  bool warpset = false;

  Matrix premat, midmat, postmat;
  premat = Identity(4);
  midmat = Identity(4);
  postmat = Identity(4);

  if (prematname.set()) {
    read_ascii_matrix(premat,prematname.value());
    affine2warp(premat,nextwarp,refvol);
    update_warp(finalwarp,nextwarp,warpset);
  }

  if (warp1name.set()) {
    read_volume4D(nextwarp,warp1name.value());
    update_warp(finalwarp,nextwarp,warpset);
  }

  if (midmatname.set()) {
    read_ascii_matrix(midmat,midmatname.value());
    affine2warp(midmat,nextwarp,refvol);
    update_warp(finalwarp,nextwarp,warpset);
  }

  if (warp2name.set()) {
    read_volume4D(nextwarp,warp2name.value());
    update_warp(finalwarp,nextwarp,warpset);
  }

  if (postmatname.set()) {
    read_ascii_matrix(postmat,postmatname.value());
    affine2warp(postmat,nextwarp,refvol);
    update_warp(finalwarp,nextwarp,warpset);
  }

  // apply shiftmap last (if it exists)
  if (shiftmapname.set()) {
    volume<float> shiftmap;
    read_volume(shiftmap,shiftmapname.value());
    shift2warp(shiftmap,nextwarp,shiftdir.value());
    update_warp(finalwarp,nextwarp,warpset);
  }

  // force the final warp to be the same size as refvol (if it isn't already)
  if (!samesize(finalwarp[0],refvol)) {
    affine2warp(Identity(4),nextwarp,refvol);
    update_warp(finalwarp,nextwarp,warpset);
  }
    
  save_volume4D(finalwarp,outname.value());
  
  return 0;
}





int main(int argc, char* argv[])
{
  Tracer tr("main");

  OptionParser options(title, examples);

  try {
    options.add(outname);
    options.add(prematname);
    options.add(warp1name);
    options.add(midmatname);
    options.add(warp2name);
    options.add(postmatname);
    options.add(refname);
    options.add(shiftmapname);
    options.add(shiftdir);
    options.add(verbose);
    options.add(help);
    
    options.parse_command_line(argc, argv);

    if ( (help.value()) || (!options.check_compulsory_arguments(true)) )
      {
	options.usage();
	exit(EXIT_FAILURE);
      }
  }  catch(X_OptionError& e) {
    options.usage();
    cerr << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  } catch(std::exception &e) {
    cerr << e.what() << endl;
  } 

  return convert_warp();
}




