/*
  irva - fMRI IRVA analysis 

  Stuart Clare, FMRIB Physics Group

  Copyright (c) 1998-2000 University of Oxford
   
  Clare, S. et.al. (1999) Magnetic Resonance in Medicine. 42,1117-1122.
*/

/* CCOPYRIGHT */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "avwio/avwio.h"
#include "ftoz.h"

/* usage: irva infile Ffile Zfile %thold offset points [-av file] */

#define MAXLEN 128

int main(int argc,char *argv[])
{
  char infile[MAXLEN],foutfile[MAXLEN],zoutfile[MAXLEN],avfile[MAXLEN];
  short *image,thold;
  short max=0;
  double *F,*Z,thresh;
  int h,i,j,k,l,points,cycles,offset;
  short ppl,lpi,ipv,vols,type,av=0;
  long volsize,c;
  float *fdata;
  double *data,T,C,SST,SSTr,Ti,SSE;
  AVW *ip,*op;

  if(argc<7){
    printf("usage: irva infile Ffile Zfile %%thold offset points [-av file]\n");
    return(1);
  }

  sprintf(infile,"%s",argv[1]);
  sprintf(foutfile,"%s",argv[2]);
  sprintf(zoutfile,"%s",argv[3]);
  sscanf(argv[4],"%lf",&thresh);
  sscanf(argv[5],"%d",&offset);
  sscanf(argv[6],"%d",&points);
  if(argc>7){
    if(!strcmp(argv[7],"-av")){
      av=1;
      sprintf(avfile,"%s",argv[8]);
    }
  }

  /* Open input AVW file */
  if((ip=AvwOpen(infile,"r"))==NULL){
    printf("Failed to open AVW file: %s\n",infile);
    return(1);
  }
  AvwGetDim(ip,&ppl,&lpi,&ipv,&vols);
  AvwGetDataType(ip,&type);
  
  if(type!=DT_SIGNED_SHORT){
    printf("Data type must be signed short");
    return(1);
  }
  volsize=ppl*lpi*ipv;
  cycles=(vols-offset)/points;
  vols=(cycles*points);

  /* Malloc */
  if(volsize>ppl*vols){
    if((image=(short *)malloc(volsize*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  } else {
    if((image=(short *)malloc(ppl*vols*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  }
  if((data=(double *)malloc(vols*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((F=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((Z=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((fdata=(float *)malloc(volsize*sizeof(float)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }

  /* Read in first image and find max */
  AvwSeekVolume(ip,0);
  if(AvwReadVolumes(ip,image,1)!=1){
    printf("Read error: %s\n",infile);
    return(1);
  }
  for(i=0;i<volsize;i++)if(image[i]>max)max=image[i];
  thold=(short)(max*thresh/100);

  /* For each line in image */
  for(h=0;h<ipv;h++){
    for(i=0;i<lpi;i++){	
      
      /* Read in data */ 
      AvwSeekVolume(ip,0);
      if((short)AvwReadRowSeries(ip,image,i,h,vols)!=vols){
	printf("Read error: %s\n",infile);
	return(1);
      }
      
      /* For each point in line */
      for(j=0;j<ppl;j++){		
	if(image[j]>thold){
	  /* Unaveraged data set */
	  for(k=0;k<vols;k++)data[k]=(double)image[(k*ppl)+j];
	  
	  T=SST=0.0;
	  for(k=0;k<vols;k++){
	    T+=data[k];
	    SST+=(data[k]*data[k]);
	  }
	  C=(T*T)/vols;
	  SST-=C;
	  SSTr=0.0;
	  for(k=0;k<points;k++){
	    for(l=0,Ti=0.0;l<cycles;l++){
	      Ti+=data[(l*points)+k];
	    }
	    SSTr+=(Ti*Ti);
	  }
	  SSTr/=cycles;
	  SSTr-=C;
	  SSE=SST-SSTr;
	  F[(h*lpi*ppl)+(i*ppl)+j]=(SSTr/(points-1))/(SSE/(points*(cycles-1)));
	}
	else F[(h*lpi*ppl)+(i*ppl)+j]=0;
      }
    }
  }

  /* Create average image */
  if(av){
    if((op=AvwOpen(avfile,"w"))==NULL){
      printf("Unable to create AVW file: %s\n",avfile);
    }
    AvwCloneHeader(op,ip);
    AvwSetDim(op,ppl,lpi,ipv,points);
    AvwSetDataType(op,DT_FLOAT);
    AvwSetMinMax(op,0,max);

    for(k=0;k<points;k++){
      for(c=0;c<volsize;c++)fdata[c]=0.0;
      for(j=0;j<cycles;j++){
	AvwSeekVolume(ip,j*points+k+offset);
	if(AvwReadVolumes(ip,image,1)!=1){
	  printf("Read error: %s\n",infile);
	  return(1);	  
	}
	for(c=0;c<volsize;c++)fdata[c]+=((float)image[c]/(float)cycles);
      }
      AvwSeekVolume(op,k);
      if(AvwWriteVolumes(op,fdata,1)!=1){
	printf("Write error: %s\n",avfile);
	return(1);
      }
    }
    AvwClose(op);
  }

  /* Transform F to Z */
  ftoz(F,Z,volsize,points-1,points*(cycles-1));

  /* Write output */
  if((op=AvwOpen(foutfile,"w"))==NULL){
    printf("Unable to create AVW file: %s\n",foutfile);
    return(1);
  }
  AvwCloneHeader(op,ip);
  AvwSetDim(op,ppl,lpi,ipv,1);
  AvwSetDataType(op,DT_FLOAT);
  AvwSetMinMax(op,0,10);
  for(c=0;c<volsize;c++)fdata[c]=(float)F[c];
  AvwSeekVolume(op,0);
  if(AvwWriteVolumes(op,fdata,1)!=1){
    printf("Write error: %s\n",foutfile);
    return(1);
  }
  AvwClose(op);
  
  if((op=AvwOpen(zoutfile,"w"))==NULL){
    printf("Unable to create AVW file: %s\n",zoutfile);
    return(1);
  }
  AvwCloneHeader(op,ip);
  AvwSetDim(op,ppl,lpi,ipv,1);
  AvwSetDataType(op,DT_FLOAT);
  AvwSetMinMax(op,0,10);
  for(c=0;c<volsize;c++)fdata[c]=(float)Z[c];
  AvwSeekVolume(op,0);
  if(AvwWriteVolumes(op,fdata,1)!=1){
    printf("Write error: %s\n",foutfile);
    return(1);
  }
  AvwClose(op);

  AvwClose(ip);
  return(0);
}
