// Templated image storage class
//  Written by Mark Jenkinson  March 2000

#if !defined(__newimage_h)
#define __newimage_h

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include "newmatap.h"
#include "lazy.h"
#include "lazyiterators.h"
#include "positerators.h"
#include "miscmaths/miscmaths.h"
#include "miscmaths/kernel.h"
 
using namespace NEWMAT;
using namespace LAZY;
using namespace MISCMATHS;
using namespace std;

namespace NEWIMAGE {


  const bool USEASSERT=false;  // used inside imthrow to determine type of action

  void imthrow(const string& msg, int nierrnum);

  enum extrapolation { zeropad, constpad, extraslice, mirror, periodic,
                       boundsassert, boundsexception, userextrapolation };

  enum interpolation { nearestneighbour, trilinear, sinc, userkernel, 
		       userinterpolation };

#pragma interface  

  template <class T>
  class volume : public lazymanager {
  private:

    T* Data;
    bool data_owner;
    int SizeBound;
    int SliceOffset;

    int SlicesZ;
    int RowsY;
    int ColumnsX;

    float Xdim;
    float Ydim;
    float Zdim;
    int Xsign;
    int Ysign;
    int Zsign;
    mutable ColumnVector origin;  // always in VOXEL coordinates

    mutable std::vector<int> ROIbox;
    mutable bool activeROI;
    mutable std::vector<int> Limits;

    mutable int no_voxels;
    lazy<std::vector<T>, volume<T> > minmax;
    lazy<std::vector<float>, volume<T> > sums;
    lazy<std::vector<T>, volume<T> > robustlimits;
    lazy<Matrix, volume<T> > principleaxes;
    lazy<std::vector<T>, volume<T> > percentiles;
    mutable std::vector<float> percentilepvals;
    lazy<ColumnVector, volume<T> > l_histogram;
    mutable int HISTbins;
    mutable T HISTmin;
    mutable T HISTmax;

    mutable kernel interpkernel;
    mutable extrapolation p_extrapmethod;
    mutable interpolation p_interpmethod;
    mutable T (*p_userextrap)(const volume<T>& , int, int, int);
    mutable float (*p_userinterp)(const volume<T>& , float, float, float);
    mutable T padvalue;
    mutable T extrapval;  // the reference target for all extrapolations



    // Internal functions
    inline T* basicptr(int x, int y, int z)
      { return (Data + (z*RowsY + y)*ColumnsX + x); }
    inline T* basicptr(int x, int y, int z) const
      { return (Data + (z*RowsY + y)*ColumnsX + x); }
    
    int initialize(int xsize, int ysize, int zsize, T *d, bool d_owner);
    void setdefaultproperties();
    void setupsizeproperties() const;
    void setdefaultlimits() const;
    void enforcelimits(std::vector<int>& lims) const;
    void calc_no_voxels() const;
    const T& extrapolate(int x, int y, int z) const;
    float kernelinterpolation(const float x, const float y, 
			      const float z) const;

    typedef T* nonsafe_fast_iterator;
    inline nonsafe_fast_iterator nsfbegin()
      { set_whole_cache_validity(false); return Data; }
    inline nonsafe_fast_iterator nsfend()
      { return Data+SizeBound; }
    template <class S, class D> friend
    void copybasicproperties(const volume<S>& source, volume<D>& dest);

  public:

    // CONSTRUCTORS AND DESTRUCTORS (including copy, = and reinitialize)
    volume();
    volume(const volume<T>& source);
    volume(int xsize, int ysize, int zsize);
    volume(int xsize, int ysize, int zsize, T *d, bool d_owner);
    ~volume();
    void destroy();
    const volume<T>& operator=(const volume<T>& source); 
    int reinitialize(const volume<T>& source);
    int reinitialize(int xsize, int ysize, int zsize);
    int reinitialize(int xsize, int ysize, int zsize, T *d, bool d_owner);
    int copyproperties(const volume<T>& source);
    int copydata(const volume<T>& source);

    // BASIC PROPERTIES
    inline int xsize() const { return ColumnsX; }
    inline int ysize() const { return RowsY; }
    inline int zsize() const { return SlicesZ; }
    inline float xdim() const { return Xdim; }
    inline float ydim() const { return Ydim; }
    inline float zdim() const { return Zdim; }
    inline const int xsign(void) const { return Xsign; }
    inline const int ysign(void) const { return Ysign; }
    inline const int zsign(void) const { return Zsign; }

    void setxdim(float x) { Xdim = fabs(x);  if (x<0) Xsign=-1; else Xsign=1; }
    void setydim(float y) { Ydim = fabs(y);  if (y<0) Ysign=-1; else Ysign=1; }
    void setzdim(float z) { Zdim = fabs(z);  if (z<0) Zsign=-1; else Zsign=1; }
    void setdims(float x, float y, float z) 
      { setxdim(x); setydim(y); setzdim(z); }
    int nvoxels() const { return no_voxels; }

    // ROI FUNCTIONS
    inline const std::vector<int>& limits() const { return Limits; }
    inline int limits(int n) const { return Limits[n]; }
    inline int minx() const { return Limits[0]; }
    inline int maxx() const { return Limits[3]; }
    inline int miny() const { return Limits[1]; }
    inline int maxy() const { return Limits[4]; }
    inline int minz() const { return Limits[2]; }
    inline int maxz() const { return Limits[5]; }
    inline const std::vector<int>& ROIlimits() const { return ROIbox; }
    inline int ROIlimits(int n) const { return ROIbox[n]; }
    inline bool usingROI() const { return activeROI; }
    void setROIlimits(int x0, int y0, int z0, int x1, int y1, int z1) const;
    void setROIlimits(const std::vector<int>& lims) const;
    void activateROI() const; 
    void deactivateROI() const;
    volume<T> ROI() const;  // returns a new volume = ROI
    int copyROIonly(const volume<T>& source);

    // SECONDARY PROPERTIES
    Matrix sampling_mat() const;
    const ColumnVector& getorigin() const { return origin; }
    void setorigin(const ColumnVector& orig) const { origin = orig; }
    void setorigin(float x, float y, float z) const
      { origin(1) = x; origin(2) = y; origin(3) = z; }
    T min() const { return minmax()[0]; }
    T max() const { return minmax()[1]; }
    float sum() const { return sums()[0]; }
    float sumsquares() const { return sums()[1]; }
    float mean() const { return sum()/((float) no_voxels); }
    float variance() const { return sumsquares()/((float) no_voxels-1) -
					   mean()*mean(); }
    float stddev() const { return sqrt(variance()); }
    T robustmin() const { return robustlimits()[0]; }
    T robustmax() const { return robustlimits()[1]; }
    ColumnVector principleaxis(int n) const;
    Matrix principleaxes_mat() const;
    T percentile(float pvalue) const;  // argument in range [0.0 , 1.0]
    std::vector<float> percentilepvalues() const { return percentilepvals; }
    ColumnVector histogram(int nbins) const;
    ColumnVector histogram(int nbins, T minval, T maxval) const;
    int histbins() const { return HISTbins; }
    T histmin() const { return HISTmin; }
    T histmax() const { return HISTmax; }

    lazy<T, volume<T> > backgroundval;
    lazy<ColumnVector, volume<T> > cog;


    // DATA ACCESS FUNCTIONS (iterators)
//      typedef poslazyiterator<T> iterator;
//      typedef posconstiterator<T> const_iterator;
//      typedef rlazyiterator<T*,T> fast_iterator;
    typedef const T* fast_const_iterator;
    
//      inline iterator begin();
//      inline iterator end();
//      inline const_iterator begin() const; 
//      inline const_iterator end() const;
//      inline fast_iterator fbegin()
//        { return fast_iterator(Data,(lazymanager*) this); }
//      inline fast_iterator fend()
//        { return fast_iterator(Data+SizeBound,(lazymanager*) this); }
    inline fast_const_iterator fbegin() const 
      { return Data; }
    inline fast_const_iterator fend() const 
      { return Data+SizeBound; }


    // BASIC DATA ACCESS FUNCTIONS
    inline bool in_bounds(int x, int y, int z) const
      { return ( (x>=0) && (y>=0) && (z>=0) 
		 && (x<ColumnsX) && (y<RowsY) && (z<SlicesZ) ); }
    inline T& operator()(int x, int y, int z)
      { set_whole_cache_validity(false); 
        if (in_bounds(x,y,z)) return *(basicptr(x,y,z)); 
	else                  return const_cast<T& > (extrapolate(x,y,z)); }
    inline const T& operator()(int x, int y, int z)
      const { 	if (in_bounds(x,y,z)) return *(basicptr(x,y,z)); 
	        else                  return extrapolate(x,y,z); }
    float interpolate(float x, float y, float z) const;

    inline T& value(int x, int y, int z)
      { set_whole_cache_validity(false); 
        return *(basicptr(x,y,z)); }
    inline const T& value(int x, int y, int z) const
      { return *(basicptr(x,y,z)); }
    float interpolatevalue(float x, float y, float z) const;


    // SECONDARY FUNCTIONS
    void setextrapolationmethod(extrapolation extrapmethod) const 
      { p_extrapmethod = extrapmethod;}; 
    extrapolation getextrapolationmethod() const { return p_extrapmethod; }
    void setpadvalue(T padval) const { padvalue = padval; }
    T getpadvalue() const { return padvalue; }
    void defineuserextrapolation(T (*extrap)(
             const volume<T>& , int, int, int)) const;

    void setinterpolationmethod(interpolation interpmethod) const;
    interpolation getinterpolationmethod() const { return p_interpmethod; }
    void defineuserinterpolation(float (*interp)(
             const volume<T>& , float, float, float)) const;
    void definekernelinterpolation(const ColumnVector& kx, 
				   const ColumnVector& ky,
				   const ColumnVector& kz, 
				   int wx, int wy, int wz) const;  // full-width
    void definekernelinterpolation(const volume<T>& vol) const;
    void definesincinterpolation(const string& sincwindowtype,
				 int w, int nstore=1201) const;  // full-width
    void definesincinterpolation(const string& sincwindowtype,
				 int wx, int wy, int wz, int nstore=1201) const;
                                  // full-width

    inline void getneighbours(int x, int y, int z, 
			      T &v000, T &v001, T &v010,
			      T &v011, T &v100, T &v101,
			      T &v110, T &v111) const;
    inline void getneighbours(int x, int y, int z, 
			      T &v000, T &v010,
			      T &v100, T &v110) const;
    
    

    // ARITHMETIC FUNCTIONS
    T operator=(T val); 
    const volume<T>& operator+=(T val); 
    const volume<T>& operator-=(T val); 
    const volume<T>& operator*=(T val); 
    const volume<T>& operator/=(T val); 
    const volume<T>& operator+=(const volume<T>& source); 
    const volume<T>& operator-=(const volume<T>& source); 
    const volume<T>& operator*=(const volume<T>& source); 
    const volume<T>& operator/=(const volume<T>& source); 

    volume<T> operator+(T num) const;
    volume<T> operator-(T num) const;
    volume<T> operator*(T num) const;
    volume<T> operator/(T num) const;
    volume<T> operator+(const volume<T>& vol2) const;
    volume<T> operator-(const volume<T>& vol2) const;
    volume<T> operator*(const volume<T>& vol2) const;
    volume<T> operator/(const volume<T>& vol2) const;

    template <class S>
    friend volume<S> operator+(S num, const volume<S>& vol);
    template <class S>
    friend volume<S> operator-(S num, const volume<S>& vol);
    template <class S>
    friend volume<S> operator*(S num, const volume<S>& vol);
    template <class S>
    friend volume<S> operator/(S num, const volume<S>& vol);

    // GENERAL MANIPULATION

    void binarise(T lowerth, T upperth);
    void binarise(T thresh) { this->binarise(thresh,this->max()); }
    void threshold(T lowerth, T upperth);
    void threshold(T thresh) { this->threshold(thresh,this->max()); }
    void swapdimensions(int dim1, int dim2, int dim3);
     
    // CONVERSION FUNCTIONS
    template <class S, class D> friend
    void copyconvert(const volume<S>& source, volume<D>& dest);
      
  };


  // HELPER FUNCTIONS

  template <class S, class D>
  void convertbuffer(const S* source, D* dest, int len);

  template <class S1, class S2>
  bool samesize(const volume<S1>& vol1, const volume<S2>& vol2);

  template <class S1, class S2>
  bool sameabssize(const volume<S1>& vol1, const volume<S2>& vol2);



  //////////////////////// VOLUME4D CLASS /////////////////////////////


  template <class T>
  class volume4D : public lazymanager {
  private:
    std::vector<volume<T> > vols;
    float p_TR;

    mutable std::vector<int> ROIbox;
    mutable bool activeROI;
    mutable std::vector<int> Limits;

    mutable extrapolation p_extrapmethod;
    mutable interpolation p_interpmethod;
    mutable T (*p_userextrap)(const volume<T>& , int, int, int);
    mutable float (*p_userinterp)(const volume<T>& , float, float, float);
    mutable T p_padval;

    lazy<std::vector<T>, volume4D<T> > tsminmax;

    // Internal functions
    int initialize(int xsize, int ysize, int zsize, int tsize, T *d=0);
    void setdefaultproperties();
    void setdefaultlimits() const;
    void enforcelimits(std::vector<int>& lims) const;
    template <class S, class D> friend
    void copybasicproperties(const volume4D<S>& source, volume4D<D>& dest);
    int copyproperties(const volume<T>& source);

  public:
    // CONSTRUCTORS AND DESTRUCTORS (including copy, = and reinitialize)
    volume4D();
    volume4D(const volume4D<T>& source);
    volume4D(int xsize, int ysize, int zsize, int tsize, T *d=0);
    ~volume4D();
    void destroy();
    const volume4D<T>& operator=(const volume4D<T>& source); 
    int reinitialize(const volume4D<T>& source);
    int reinitialize(int xsize, int ysize, int zsize, int tsize, T *d=0);
    int copyproperties(const volume4D<T>& source);
    int copyvolumes(const volume4D<T>& source);
    
    // DATA ACCESS
    inline bool in_bounds(int x, int y, int z) const
      { return ( (vols.size()>0) && vols[0].in_bounds(x,y,z) ); }
    inline bool in_bounds(int t) const
      { return ( (t>=0) && (t<this->tsize()) ); }
    inline bool in_bounds(int x, int y, int z, int t) const
      { return ( (t>=0) && (t<this->tsize()) && 
		 vols[Limits[3]].in_bounds(x,y,z) ); }

    inline T& operator()(int x, int y, int z, int t) 
      { set_whole_cache_validity(false);  
        if (!in_bounds(t)) imthrow("Out of Bounds (time index)",5); 
	return vols[t](x,y,z); }
    inline const T& operator()(int x, int y, int z, int t) const 
      { if (!in_bounds(t)) imthrow("Out of Bounds (time index)",5); 
        return vols[t](x,y,z); }

    inline T& value(int x, int y, int z, int t) 
      { set_whole_cache_validity(false);  return vols[t].value(x,y,z); }
    inline const T& value(int x, int y, int z, int t) const 
      { return vols[t].value(x,y,z); }

    inline volume<T>& operator[](int t) {
      set_whole_cache_validity(false);  
      if (!in_bounds(t)) imthrow("Out of Bounds (time index)",5);   
      return vols[t]; }
    inline const volume<T>& operator[](int t) const 
      { if (!in_bounds(t)) imthrow("Out of Bounds (time index)",5);  
        return vols[t]; }
    
    // SINGLE VOXEL TIME-SERIES ACCESS
    ReturnMatrix voxelts(int x, int y, int z) const;
    void setvoxelts(const ColumnVector& ts, int x, int y, int z);
    
    // WHOLE VOLUME MODIFIERS
    void addvolume(const volume<T>& source);
    void insertvolume(const volume<T>& source, int t);
    void deletevolume(int t);

    // MATRIX <-> VOLUME4D CONVERSIONS
    Matrix matrix(const volume<T>& mask) const;
    Matrix matrix() const;
    void setmatrix(const Matrix& newmatrix, const volume<T>& mask, 
		   const T pad=0);
    void setmatrix(const Matrix& newmatrix); 

    // SIZE AND DIMENSIONS
    inline int xsize() const 
      { if (vols.size()>0) return vols[0].xsize(); else return 0; }
    inline int ysize() const
      { if (vols.size()>0) return vols[0].ysize(); else return 0; }
    inline int zsize() const
      { if (vols.size()>0) return vols[0].zsize(); else return 0; }
    inline int tsize() const { return (int) vols.size(); }
    inline float xdim() const
      { if (vols.size()>0) return vols[0].xdim(); else return 1.0; }
    inline float ydim() const
      { if (vols.size()>0) return vols[0].ydim(); else return 1.0; }
    inline float zdim() const
      { if (vols.size()>0) return vols[0].zdim(); else return 1.0; }
    inline float tdim() const { return p_TR; }
    inline float TR() const { return p_TR; }
    inline const int xsign(void) const
      { if (vols.size()>0) return vols[0].xsign(); else return 1; }
    inline const int ysign(void) const
      { if (vols.size()>0) return vols[0].ysign(); else return 1; }
    inline const int zsign(void) const
      { if (vols.size()>0) return vols[0].zsign(); else return 1; }

    void setxdim(float x);
    void setydim(float y);
    void setzdim(float z);
    void settdim(float tr) { p_TR = fabs(tr); }
    void setTR(float tr)   { settdim(tr); }
    void setdims(float x, float y, float z, float tr) 
      { setxdim(x); setydim(y); setzdim(z); settdim(tr); }
    int nvoxels() const 
      { if (vols.size()>0) return vols[0].nvoxels(); else return 0; }
    int ntimepoints() const { return this->tsize(); }

    void setinterpolationmethod(interpolation interpmethod) const;
    interpolation getinterpolationmethod() const;
    void setextrapolationmethod(extrapolation extrapmethod) const;
    extrapolation getextrapolationmethod() const;
    void setpadvalue(T padval) const;
    T getpadvalue() const;
    void defineuserinterpolation(float (*interp)(
		       const volume<T>& , float, float, float)) const;
    void defineuserextrapolation(T (*extrap)(
		       const volume<T>& , int, int, int)) const;
    void definekernelinterpolation(const ColumnVector& kx, 
				   const ColumnVector& ky,
				   const ColumnVector& kz, 
				   int wx, int wy, int wz) const;
    void definekernelinterpolation(const volume4D<T>& vol) const;
    void definekernelinterpolation(const volume<T>& vol) const;
    void definesincinterpolation(const string& sincwindowtype,
				 int w, int nstore=1201) const;
    void definesincinterpolation(const string& sincwindowtype,
				 int wx, int wy, int wz, int nstore=1201) const;

    // ROI FUNCTIONS
    inline const std::vector<int>& limits() const { return Limits; }
    inline int limits(int n) const { return Limits[n]; }
    inline int minx() const { return Limits[0]; }
    inline int maxx() const { return Limits[4]; }
    inline int miny() const { return Limits[1]; }
    inline int maxy() const { return Limits[5]; }
    inline int minz() const { return Limits[2]; }
    inline int maxz() const { return Limits[6]; }
    inline int mint() const { return Limits[3]; }
    inline int maxt() const { return Limits[7]; }
    inline const std::vector<int>& ROIlimits() const { return ROIbox; }
    inline int ROIlimits(int n) const { return ROIbox[n]; }
    inline bool usingROI() const { return activeROI; }
    void setROIlimits(int x0, int y0, int z0, int t0,
                      int x1, int y1, int z1, int t1) const;
    void setROIlimits(int t0, int t1) const;
    void setROIlimits(int x0, int y0, int z0, int x1, int y1, int z1) const; 
    void setROIlimits(const std::vector<int>& lims) const;
    void activateROI() const; 
    void deactivateROI() const;
    volume4D<T> ROI() const;  // returns a new volume = ROI
    int copyROIonly(const volume4D<T>& source);


    T min() const { return tsminmax()[0]; }
    T max() const { return tsminmax()[1]; }

    // GENERAL MANIPULATION

    void binarise(T lowerth, T upperth);
    void binarise(T thresh) { this->binarise(thresh,this->max()); }
    void threshold(T lowerth, T upperth);
    void threshold(T thresh) { this->threshold(thresh,this->max()); }
    void swapdimensions(int dim1, int dim2, int dim3);

    // ARITHMETIC FUNCTIONS
    T operator=(T val); 
    const volume4D<T>& operator+=(T val); 
    const volume4D<T>& operator-=(T val); 
    const volume4D<T>& operator*=(T val); 
    const volume4D<T>& operator/=(T val); 
    const volume4D<T>& operator+=(const volume<T>& source); 
    const volume4D<T>& operator-=(const volume<T>& source); 
    const volume4D<T>& operator*=(const volume<T>& source); 
    const volume4D<T>& operator/=(const volume<T>& source); 
    const volume4D<T>& operator+=(const volume4D<T>& source); 
    const volume4D<T>& operator-=(const volume4D<T>& source); 
    const volume4D<T>& operator*=(const volume4D<T>& source); 
    const volume4D<T>& operator/=(const volume4D<T>& source); 

    volume4D<T> operator+(T num) const;
    volume4D<T> operator-(T num) const;
    volume4D<T> operator*(T num) const;
    volume4D<T> operator/(T num) const;
    volume4D<T> operator+(const volume<T>& vol2) const;
    volume4D<T> operator-(const volume<T>& vol2) const;
    volume4D<T> operator*(const volume<T>& vol2) const;
    volume4D<T> operator/(const volume<T>& vol2) const;
    volume4D<T> operator+(const volume4D<T>& vol2) const;
    volume4D<T> operator-(const volume4D<T>& vol2) const;
    volume4D<T> operator*(const volume4D<T>& vol2) const;
    volume4D<T> operator/(const volume4D<T>& vol2) const;

    template <class S>
    friend volume4D<S> operator+(S num, const volume4D<S>& vol);
    template <class S>
    friend volume4D<S> operator-(S num, const volume4D<S>& vol);
    template <class S>
    friend volume4D<S> operator*(S num, const volume4D<S>& vol);
    template <class S>
    friend volume4D<S> operator/(S num, const volume4D<S>& vol);
    template <class S>
    friend volume4D<S> operator+(const volume<S>& v1, const volume4D<S>& v2);
    template <class S>
    friend volume4D<S> operator-(const volume<S>& v1, const volume4D<S>& v2);
    template <class S>
    friend volume4D<S> operator*(const volume<S>& v1, const volume4D<S>& v2);
     
    // CONVERSION FUNCTIONS
    template <class S, class D> friend
    void copyconvert(const volume4D<S>& source, volume4D<D>& dest);
  };


  // HELPER FUNCTIONS

  template <class S1, class S2>
  bool samesize(const volume4D<S1>& vol1, const volume4D<S2>& vol2);

  template <class S1, class S2>
  bool sameabssize(const volume4D<S1>& vol1, const volume4D<S2>& vol2);

////////////////////////////////////////////////////////////////////////
///////////////////////// DEBUGGING FUNCTIONS //////////////////////////
////////////////////////////////////////////////////////////////////////

  template <class T>
  void print_info(const volume<T>& source, const string& name)
    {
      cout << name << ":: Size = (" << source.xsize() << ","
	   << source.ysize() << "," << source.zsize() << ")" << endl;
      cout << name << ":: ROI Size = (" << source.maxx() - source.minx() + 1 
	   << "," << source.maxy() - source.miny() + 1
	   << "," << source.maxz() - source.minz() + 1 << ")" << endl;
      cout << name << ":: Dims = (" << source.xdim() << ","
	   << source.ydim() << "," << source.zdim() << ")" << endl;
      cout << name << ":: Maximum and minimum intensities are: " 
	   << source.max() << " and " << source.min() << endl;
    }  

  template <class T>
  void print_info(const volume4D<T>& source, const string& name)
    {
      cout << name << ":: Size = (" << source.xsize() << ","
	   << source.ysize() << "," << source.zsize() << ","
	   << source.tsize() << ")" << endl;
      cout << name << ":: ROI Size = (" << source.maxx() - source.minx() + 1 
	   << "," << source.maxy() - source.miny() + 1
	   << "," << source.maxz() - source.minz() + 1 
	   << "," << source.maxt() - source.mint() + 1 << ")" << endl;
      cout << name << ":: Dims = (" << source.xdim() << ","
	   << source.ydim() << "," << source.zdim() << "," 
	   << source.tdim() << ")" << endl;
      cout << name << ":: Maximum and minimum intensities are: " 
	   << source.max() << " and " << source.min() << endl;
    }  


////////////////////////////////////////////////////////////////////////
///////////////////////// INLINE DEFINITIONS ///////////////////////////
////////////////////////////////////////////////////////////////////////

  template <class T>
  inline void volume<T>::getneighbours(int x, int y, int z, 
					    T &v000, T &v001, T &v010,
					    T &v011, T &v100, T &v101,
					    T &v110, T &v111) const {
    T *ptr = basicptr(x,y,z);
    v000 = *ptr;
    ptr++; 
    v100 = *ptr;
    ptr+= ColumnsX;  
    v110 = *ptr;
    ptr--; 
    v010 = *ptr;
    ptr += SliceOffset;
    v011 = *ptr;
    ptr++;
    v111 = *ptr;
    ptr-= ColumnsX;
    v101 = *ptr;
    ptr--;
    v001 = *ptr;
  }

	
  template <class T>
  inline void volume<T>::getneighbours(int x, int y, int z, 
					    T &v000, T &v010, 
					    T &v100, T &v110) const {
    T *ptr = basicptr(x,y,z);
    v000 = *ptr;
    ptr++; 
    v100 = *ptr;
    ptr+= ColumnsX;  
    v110 = *ptr;
    ptr--; 
    v010 = *ptr;
  }
	

//    template <class T>
//    inline volume<T>::iterator volume<T>::begin()
//    { return volume<T>::iterator(basicptr(Limits[0],Limits[1],Limits[2]),
//  		    (lazymanager*) this, 
//  		    Limits[0],Limits[1],Limits[2],
//  		    Limits[0],Limits[1],Limits[2],
//  		    Limits[3],Limits[4],Limits[5],
//  		    ColumnsX,SliceOffset); }

//    template <class T>
//    inline volume<T>::iterator volume<T>::end()
//    { volume<T>::iterator tmp(basicptr(Limits[3],Limits[4],Limits[5]),
//  		 (lazymanager*) this,
//  		 Limits[3],Limits[4],Limits[5],
//  		 Limits[0],Limits[1],Limits[2],
//  		 Limits[3],Limits[4],Limits[5],
//  		 ColumnsX,SliceOffset); 
//                   return ++tmp; }

//    template <class T>
//    inline volume<T>::const_iterator volume<T>::begin() const 
//    { return 
//        volume<T>::const_iterator(basicptr(Limits[0],Limits[1],Limits[2]),
//  			  Limits[0],Limits[1],Limits[2],
//  			  Limits[0],Limits[1],Limits[2],
//  			  Limits[3],Limits[4],Limits[5],
//  			  ColumnsX,SliceOffset); }

//    template <class T>
//    inline volume<T>::const_iterator volume<T>::end() const 
//    { volume<T>::const_iterator tmp(basicptr(Limits[3],Limits[4],Limits[5]),
//  		       Limits[3],Limits[4],Limits[5],
//  		       Limits[0],Limits[1],Limits[2],
//  		       Limits[3],Limits[4],Limits[5],
//  		       ColumnsX,SliceOffset); 
//                         return ++tmp; }
  

	     
////////////////////////////////////////////////////////////////////////
/////////////////////////// HELPER FUNCTIONS ///////////////////////////
////////////////////////////////////////////////////////////////////////

  
  template <class S, class D>
  void convertbuffer(const S* source, D* dest, int len)
  {
    D* dptr=dest;
    for (const S* sptr=source; sptr<(source+len); sptr++) {
      *dptr = (D) *sptr;
      dptr++;
    }
  }


  template <class S1, class S2>
  bool samesize(const volume<S1>& vol1, const volume<S2>& vol2)
  {  
    return ( ( (vol1.maxx()-vol1.minx())==(vol2.maxx()-vol2.minx()) ) && 
	     ( (vol1.maxy()-vol1.miny())==(vol2.maxy()-vol2.miny()) ) && 
	     ( (vol1.maxz()-vol1.minz())==(vol2.maxz()-vol2.minz()) ) ); 
  }
  
  template <class S1, class S2>
  bool samesize(const volume4D<S1>& vol1, const volume4D<S2>& vol2)
  {  
    return ( ( (vol1.maxt()-vol1.mint())==(vol2.maxt()-vol2.mint()) ) && 
	     ((vol1.tsize()<=0) || (vol2.tsize()<=0) 
	        || samesize(vol1[0],vol2[0]) ) );
  }
  
  template <class S1, class S2>
  bool sameabssize(const volume<S1>& vol1, const volume<S2>& vol2)
  {  
    return ( ( (vol1.xsize())==(vol2.xsize()) ) && 
	     ( (vol1.ysize())==(vol2.ysize()) ) && 
	     ( (vol1.zsize())==(vol2.zsize()) ) ); 
  }
  
  template <class S1, class S2>
  bool sameabssize(const volume4D<S1>& vol1, const volume4D<S2>& vol2)
  {  
    return ( ( (vol1.tsize())==(vol2.tsize()) ) && 
	     ((vol1.tsize()==0) || samesize(vol1[0],vol2[0]) ) );
  }
  


////////////////////////////////////////////////////////////////////////
/////////////////////////// FRIEND FUNCTIONS ///////////////////////////
////////////////////////////////////////////////////////////////////////

  template <class S, class D>
  void copybasicproperties(const volume<S>& source, volume<D>& dest)
  {
    // set up properties (except lazy ones)
    dest.Xdim = source.Xdim;
    dest.Ydim = source.Ydim;
    dest.Zdim = source.Zdim;
    dest.Xsign = source.Xsign;
    dest.Ysign = source.Ysign;
    dest.Zsign = source.Zsign;
    dest.origin = source.origin;
    
    dest.ROIbox = source.ROIbox;
    dest.enforcelimits(dest.ROIbox);
    dest.activeROI = source.activeROI;
    if (dest.activeROI) {
      dest.Limits = source.Limits;  
      dest.enforcelimits(dest.Limits);
    } else {
      dest.setdefaultlimits();
    }
    dest.calc_no_voxels();
    
    dest.interpkernel = source.interpkernel;
    dest.p_interpmethod = source.p_interpmethod;
    dest.p_extrapmethod = source.p_extrapmethod;
    dest.padvalue = (D) source.padvalue;
  }


  template <class S, class D>
  void copyconvert(const volume<S>& source, volume<D>& dest)
  {
    // set up basic size and data storage
    dest.reinitialize(source.xsize(),source.ysize(),source.zsize());
    // set up properties (except lazy ones)
    copybasicproperties(source,dest);
    // now copy across the data
    convertbuffer(source.Data,dest.Data,source.SizeBound);
    // doubly ensure that lazy properties are not valid
    dest.set_whole_cache_validity(false);
  }
  

  template <class S, class D>
  void copybasicproperties(const volume4D<S>& source, volume4D<D>& dest)
  {
    // set up properties (except lazy ones)
    dest.p_TR = source.p_TR;
    dest.ROIbox = source.ROIbox;
    dest.enforcelimits(dest.ROIbox);
    dest.activeROI = source.activeROI;
    if ((dest.activeROI) && (sameabssize(source,dest))) {
      dest.Limits = source.Limits;  
      dest.enforcelimits(dest.Limits);
    } else {
      dest.setdefaultlimits();
    }

    dest.p_interpmethod = source.p_interpmethod;
    dest.p_extrapmethod = source.p_extrapmethod;
    // Cannot convert the following between different types
    //dest.p_userextrap = source.p_userextrap;
    //dest.p_userinterp = source.p_userinterp;
    dest.p_padval = (D) source.p_padval;

    int toffset = dest.mint() - source.mint();
    for (int t=source.mint(); t<=source.maxt(); t++) {
      copybasicproperties(source[t],dest[t + toffset]);
    }
  }


  template <class S, class D>
  void copyconvert(const volume4D<S>& source, volume4D<D>& dest)
  {
    // set up basic size and data storage
    dest.reinitialize(source.xsize(),source.ysize(),source.zsize(),
		      source.tsize());

    copybasicproperties(source,dest);
    for (int t=0; t<source.tsize(); t++) {
      copyconvert(source[t],dest[t]);
    }

    // doubly ensure that lazy properties are not valid
    dest.set_whole_cache_validity(false);
  }
  


  template <class S>
  volume<S> operator+(S num, const volume<S>& vol)
    { return (vol + num); }

  template <class S>
  volume<S> operator-(S num, const volume<S>& vol)
  {
    volume<S> tmp = vol;
    tmp=num;
    tmp-=vol;
    return tmp;
  }

  template <class S>
  volume<S> operator*(S num, const volume<S>& vol)
    { return (vol * num); }

  template <class S>
  volume<S> operator/(S num, const volume<S>& vol)
  {
    volume<S> tmp = vol;
    tmp=num;
    tmp/=vol;
    return tmp;
  }


  template <class S>
  volume4D<S> operator+(S num, const volume4D<S>& vol)
    { return (vol + num); }

  template <class S>
  volume4D<S> operator-(S num, const volume4D<S>& vol)
  {
    volume4D<S> tmp = vol;
    tmp=num;
    tmp-=vol;
    return tmp;
  }

  template <class S>
  volume4D<S> operator*(S num, const volume4D<S>& vol)
    { return (vol * num); }

  template <class S>
  volume4D<S> operator/(S num, const volume4D<S>& vol)
  {
    volume4D<S> tmp = vol;
    tmp=num;
    tmp/=vol;
    return tmp;
  }

  template <class S>
  volume4D<S> operator+(const volume<S>& v1, const volume4D<S>& v2)
    { return (v2 + v1); }

  template <class S>
  volume4D<S> operator-(const volume<S>& v1, const volume4D<S>& v2)
    { return (v2*(-1.0) + v1); }

  template <class S>
  volume4D<S> operator*(const volume<S>& v1, const volume4D<S>& v2)
    { return (v2 * v1); }

}  // end namespace

#endif








